#ifndef WIMESH_CAALGOS
#define WIMESH_CAALGOS


#include "WMtypes.h"



/*
 * Hierarchy of classes that can be used to run
 * different channel (re-)assignment algorithms
 */


/** Base class (does nothing) **/
class CAalgo
{
protected:
  
  /* A list of common operations performed by channel assignment algorithms */

  void SetChannel (RadioIfaceIterator ri, unsigned int channel) {  RadioIfaceFriend::SetChannel (ri, channel); }
  void SetChannel (RadioIface* ri, unsigned int channel) { RadioIfaceFriend::SetChannel (RadioIfaceIterator(ri), channel); }

  void SetRadio (TGraphNeighborIterator ni, RadioIface* radio) { TGraphNeighborFriend::SetRadio (ni, radio); }
  void SetFlow (TGraphNeighborIterator ni, double f) { TGraphNeighborFriend::SetFlow (ni, f); }
  void SetUtot (TGraphNeighborIterator ni, double utot) { TGraphNeighborFriend::SetUtot (ni, utot); }
  void SetRate (TGraphNeighborIterator ni, float rate) { TGraphNeighborFriend::SetRate (ni, rate); }
  void SetTxPower (TGraphNeighborIterator ni, unsigned int txpower) { TGraphNeighborFriend::SetTxPower (ni, txpower); }
  void SetAvailBW (TGraphNeighborIterator ni, double bw) { TGraphNeighborFriend::SetAvailBW (ni, bw); }
  void ClearFlowDemands (TGraphNeighborIterator ni) { TGraphNeighborFriend::ClearFlowDemands (ni); }
  void SetFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::SetFlowDemand (ni, d, f); }
  void IncreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::IncreaseFlowDemand (ni, d, f); }
  void DecreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::DecreaseFlowDemand (ni, d, f); }

  TGraphNeighborIterator AddNeighbor (TGraphNodeIterator nodei, const TGraphNeighbor* ni) { return TGraphNodeFriend::AddNeighbor (nodei, ni); }
  void EraseNeighbor (TGraphNodeIterator nodei, TGraphNeighborIterator ni) { TGraphNodeFriend::EraseNeighbor (nodei, ni); }

public:
  /* static method to create a new object of a derived class.
   * NOTE This method (defined in 1channel.cpp) must be updated every time a new derived
   * class is added
   */
  static CAalgo* Create (string algo, const list<pair<string,string> >& params);

  // should return a non-zero value in case of errors, 0 otherwise
  virtual double Run(Topology* T)  { return 0; }
  // returns the name of the CA algorithm
  string name()  { return m_name; }
  // returns the power constraints
  virtual vector<float>* getPowerConstraints () { return NULL; }
  // gives a CA algorithm the possibility to print further info
  virtual void printFurtherInfo (ostream& ofs) {}
  // destructor
  virtual ~CAalgo() {}

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params) {}

  virtual void SetName () { m_name = "N"; }

  string m_name;
};


/*** Channel Assignment algorithms ***/

/** OneChannel **/
class OneChannel : public CAalgo
{
public:
  virtual double Run(Topology* T);

protected:
  virtual void SetName ()  { m_name = "ONE"; }
};


/** IATC **/
class IATC: public CAalgo
{
public:
  virtual double Run(Topology* T);

protected:
  virtual void SetName ()  { m_name = "IATC"; }

private:
  void replace(Topology *T, TGraphNeighborIterator u, TGraphNeighborIterator v, unsigned int k, unsigned int k1);
};


/** LACA **/
class LACA: public CAalgo
{
public:
  virtual double Run(Topology* T);

  /*
   * Implementes the (iterative) channel assignment and routing algoritm proposed by Raniwala,
   * Gopalan and Chiueh in ACM Mobile Comput. Commun. Rev. 2004.
   * Call this method (instead of Run) after the PFR_Raniwala pre-computed flow rate method
   * Bwdr and seed should be the same as those passed to the PFR_Raniwala pre-computed flow rate method
   */
  void RanGopChi(Topology *T, int NumFlows, RndVar *Source, RndVar *Dest, RndVar *Bwdr, double seed);

protected:
  virtual void SetName ()  { m_name = "LACA"; }

private:
  void replace(Topology *T, TLinkItrs l, int j, unsigned int k, unsigned int k1);
  double EstimateCapacity(Topology *T, TLinkItrs l);
};


/** FCPRA **/
class FCPRA : public CAalgo
{
public:

  virtual double Run(Topology* T);

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params);

  virtual void SetName ();

  /* This function is invoked when we extract i<->j and want to optimize the radios of i (i.e., remove channel 1)*/
  /* Returns true if channel 1 is removed */
  bool fcra_OptimizeRadios (Topology *T, const TLinkItrs& l, int i, set<TLinkItrs> Q);

  /* Compute the new Utot for every link involved (except l itself) in case c is assigned to l */
  void fcra_computeNewUtot(Topology *T, const TLinkItrs& l, unsigned int c);

private:
  bool m_optimize;
  bool m_turbo;
  bool m_ratepoweradapt;
};

namespace MVCRA_NS
{
  class ChannelCostFunc;
  class LoopDetector;
}

/** MVCRA **/
class MVCRA : public CAalgo
{
public:

  virtual double Run(Topology* _T);

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params);

  virtual void SetName ();

private:
  /*
  * Determines the highest rate for l (if any) that makes Utot(l)-f(l)/c(l) smaller than Umax.
  * If such rate is not found, the highest rate that minimizes Utot(l) is returned.
  * Note that the rate of l is not changed
  */
  float Adjust_Rate (TLinkItrs& l, double Umax);

  /*
  * Determines the channel and (in case RateAdapt is true) the rate for link l that minimizes the maximum among the Utot of l and 
  * of the links including l in their collision domain. The selected channel is returned, while the selected rate is assigned to
  * new_rate. Note that this function does not change anything
  */
  unsigned int Min_Utot (TLinkItrs l, const std::set< unsigned int >& S, float& new_rate);

  /*
   * Invokes Min_Utot to select a channel, but checks whether the returned choice causes a loop. If so,
   * invokes Min_Utot again to select a different channel
   */
  unsigned int Min_Utot_safe (TLinkItrs l, const std::set< unsigned int >& S, float& new_rate);

  /*
  * Returns the channel that would disrupt the minimum number of links if it were replaced
  * by c. If Ldisrupt is a non-null pointer, it points to such set of links (must be freed after use)
  */
  unsigned int Min_Disrupt (int u, unsigned int c, MVCRA_NS::ChannelCostFunc* ccf, set<TLinkItrs>** Ldisrupt = NULL);

  /*
  * Assigns channel c to one of the radios of g. The channel to be replaced is selected by
  * calling Min_Disrupt. Some links may be added to the queue Qp of pending links.
  */
  void Change_If (MeshNode* g, unsigned int c, MVCRA_NS::ChannelCostFunc* ccf);

  /*
   * Given a TLinkItrs l having no channel (i.e., l.second->radio == NULL) which must be assigned channel c,
   * checks whether there already exists a link between l.first and l.second using channel c. If not, l
   * is assigned channel c and rate r. Otherwise, the flow on l is moved onto the previous link, l is
   * removed from the topology and from the queue Q. Requires that l.first and l.second have a radio on c.
   * The goal is to avoid duplicate parallel links on the same channel.
   */
  void AssignChannelRate (TLinkItrs& l, unsigned int c, float r);

  Topology* T;
  bool m_rateadapt;
  unsigned int m_maxnumchanges, m_numchanges;
  string m_ccf, m_lp, m_lpp;
  vector<set<unsigned int> > origChannels;
  set<TLinkItrs> Q, Qp;
  MVCRA_NS::LoopDetector* m_ld;
  vector<vector<int> > changesPerChannel;      // for each node, stores the number of times each channel is assigned to that node

  friend class MVCRA_NS::ChannelCostFunc;
};


/** MP-CARA **/
class MPCARA : public CAalgo
{
public:
    
  virtual double Run(Topology* T);
  virtual vector<float>* getPowerConstraints () { return m_nodepc; }
  virtual ~MPCARA() { if (m_nodepc) delete m_nodepc; };

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params);

  virtual void SetName ();

private:
  double doRun(Topology* T);

  RndVar* m_npc;                     // random variable to generate node power constraints
  double m_seed;                     // seed for the npc random variable
  bool m_reset;                      // true to reset the channel of all the radios
  double m_u0;                       // Utot threshold
  bool m_skipchannelcheck;           // true to skip the check for links without a channel
  bool m_old;                        // true to activate the old version of MPCARA
  vector<float>* m_nodepc;           // the vector of node power constraint (generated by npc)
  CAalgo* m_initalgo;                // the CA algorithm to run before MPCARA but after resetting the channels, if required
};


/** MPCARA-OPT **/
class MPCARA_OPT : public CAalgo
{
public:

  virtual double Run(Topology* T);
  virtual vector<float>* getPowerConstraints () { return m_nodepc; }
  virtual void printFurtherInfo (ostream& ofs) { ofs << endl << "# Highest lower bound found: " << m_lbound << endl; }
  virtual ~MPCARA_OPT() { if (m_nodepc) delete m_nodepc; };

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params);

  virtual void SetName ();

private:
  RndVar* m_npc;
  double m_seed;
  double m_u0;
  long m_to1;
  long m_to2;
  vector<float>* m_nodepc;
  double m_lbound;
};


/** E2CARA-TD **/
class E2CARA_TD : public CAalgo
{
public:

  virtual double Run(Topology* T);

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params);

  virtual void SetName ();

private:
  bool m_reset;                      // true to reset the channel of all the radios
  double m_u0;                       // Utot threshold
  CAalgo* m_initalgo;                // the CA algorithm to run before MPCARA but after resetting the channels, if required
};


/** PACA **/
class PACA : public CAalgo
{
public:

  virtual double Run(Topology* T);

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params);

  virtual void SetName ()  { m_name = "E2CARA"; }

private:
  /*
   * Attempts to reroute up to f_uv flow from link u->v subject
   * to the constraint that the Utot should be less than u0
   * Returns false if there is no path to reroute flow
   * otherLinksToPrune is a set of additional links to prune
   */
  bool Reroute(Topology* T, TLinkItrs uv, double f_uv, double u0, set<TLinkItrs>* otherLinksToPrune = 0);

  /* DEPRECATED
   * Attempts to move flow from u->v to u->w 
   * It makes sense when u->v and u->w are two "parallel" links between the same pair of nodes 
   * Returns true if it has been able to move some flow
   */
  bool MoveFlowBetweenParallelLinks(Topology *T, TGraphNodeIterator u, TGraphNeighborIterator v, TGraphNeighborIterator w, double u0);

  double m_u0;
};


/** E2CAR-OPT **/
class E2CAR_OPT : public CAalgo
{
public:

  virtual double Run(Topology* T);

protected:
  // configures all the parameters of a CA algorithm
  virtual void Configure (const list<pair<string,string> >& params);

  virtual void SetName ()  { m_name = "E2CAR-OPT"; }

private:
  bool m_relax;
  double m_u0;
};




#endif
