#include "WMtypes.h"
#include "CAalgos.h"
#include "WMutils.h"
#include "lp.h"
#include "milp.h"
#include <sstream>
#include <queue>
#include <map>
#include <algorithm>
#include <ext/functional>
#include <numeric>
#include <cassert>
#include <memory>



namespace E2CARA_TD_NS
{

  double calcPriority(Topology* T, double u0, const pair<int, set<TLinkItrs> >& p) {
    double ret=0;
    for (int x=0; x<T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++) 
	if (y->UTot() >= u0)
	  for (set<TLinkItrs>::iterator pp=p.second.begin(); pp!=p.second.end(); pp++)
	    if (pp->IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y))) 
	      ret += pp->second->Flow() / pp->second->Rate();

    return ret;
  } 

  /* Binary predicate used to compare the priorities of nodes
  * The priority of a node is the sum of the costs of the links incident on it
  * The cost of a link is its f/c ratio multiplied by the number of links
  * with Utot >= u0 having that link in their collision domain
  */
  class SumUtotPriority {
      Topology* T;
      double u0;
    public:
      SumUtotPriority(Topology* _T, double _u0) : T(_T), u0(_u0) {}
      bool operator()(const pair<int, set<TLinkItrs> >& p, const pair<int, set<TLinkItrs> >& q) {
	return calcPriority(T,u0,p) < calcPriority(T,u0,q);
      }
  };

  /* computes the sum of the flows of all the links in L */
  double sumFlows(set<TLinkItrs> L) {
    vector<double> F(L.size());
    
    transform(L.begin(), L.end(), F.begin(),
	      __gnu_cxx::compose1(mem_fun_ref (&TGraphNeighbor::Flow),
				  __gnu_cxx::compose1(mem_fun_ref (&TGraphNeighborIterator::operator*),
					              __gnu_cxx::select2nd<pair<TGraphNodeIterator,TGraphNeighborIterator> >() )  ) );
// 								mem_fun_ref(&TLinkItrs::getNeighb_ptr) )  );
    return accumulate(F.begin(), F.end(), 0.);
  }
}  /* E2CARA_TD_NS namespace */


using namespace E2CARA_TD_NS;


void E2CARA_TD::Configure (const list<pair<string,string> >& params)
{
  // default values
  m_reset = false;
  m_u0 = 0.5;

  string initalgo;
  list<pair<string,string> > initparams;

  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("reset", m_reset);
    RP.read ("init_algo", initalgo);
    RP.read ("u0", m_u0);

    if (RP.notFound())
    {
      // if the attribute name starts with "i_", it is an attribute of the init algo
      if (p->first.compare(0, 2, "i_") == 0)
      {
	initparams.push_back(pair<string,string>(p->first.substr(2),p->second));
      }
      else
	cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the E2CARA-TD algorithm" << endl << endl;
    }
  }
  m_initalgo = CAalgo::Create (initalgo, initparams);
}

void E2CARA_TD::SetName ()
{
  stringstream ss;

  ss << "E2CARA-TD";

  if (m_reset)
    ss << "R";

  if (m_initalgo)
    ss << m_initalgo->name();

  ss >> m_name;
}


/* Given a network topology, attempts to switch off as many radios as possible
   while keeping the maximum Utot below u0
   It should be always true that if a link u->v exists then a link v->u exists
   (not necessarily using the same channel)
*/
double E2CARA_TD::Run (Topology *T)
{
  if (m_reset)
    T->ResetChannels();

  if (m_initalgo)
    m_initalgo->Run(T);

  int r;
  double maxUtot=numeric_limits<double>::max(), prevUtot;
  
  cout << "Starting E2CARA-TD" << endl << endl;

#if 0
  set<unsigned int> Ne,U,V,Ccom,Ctemp;
  
  /* FIRST STAGE: attempt to add parallel links */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++) {
    /* find the set n of neighbors of u such that id(u) > id(v) (so that the pair is considered just once) */
    Ne.clear();
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      if (v->info->id < u->info->id)
	Ne.insert(v->info->id);
      
    /* for each neighbor */
    for (set<unsigned int>::iterator n=Ne.begin(); n!=Ne.end(); n++) {
      findChannels(u,U);
      findChannels((T->node.begin()+(*n)),V);
      Ccom.clear();
      set_intersection(U.begin(), U.end(), V.begin(), V.end(), inserter(Ccom, Ccom.begin()));
      
      /* find common channels that are not already used for a link u->n */
      Ctemp = Ccom;
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
	if (v->info->id == *n)
	  Ctemp.erase(v->radio->channel);

      if (canBeNeighbors(u,T->node.begin()+(*n),r,p))
	for (set<unsigned int>::iterator ch=Ctemp.begin(); ch!=Ctemp.end(); ch++) {
	  /* establish a parallel link u->n on ch */
	  u->neighbor.push_front(TGraphNeighbor((T->node.begin()+(*n))->info, r, p));
	  TLinkItrs lnew(u, u->neighbor.begin());
	  T->AssignChannel2Link(lnew,*ch);
	  lnew.second->Utot = T->computeUtot(lnew);
	  cout << "Aggiunto link (" << lnew.first->info->id << "," << lnew.second->info->id << ")[" << lnew.second->radio->channel << "]" << endl;
	}

      /* find common channels that are not already used for a link n->u */
      Ctemp = Ccom;
      for (list<TGraphNeighbor>::iterator v=T->node[*n].neighbor.begin(); v!=T->node[*n].neighbor.end(); v++)
	if (v->info->id == u->info->id)
	  Ctemp.erase(v->radio->channel);

      if (canBeNeighbors(T->node.begin()+(*n),u,r,p))
	for (set<unsigned int>::iterator ch=Ctemp.begin(); ch!=Ctemp.end(); ch++) {
	  /* establish a parallel link n->u on ch */
	  T->node[*n].neighbor.push_front(TGraphNeighbor(u->info, r, p));
	  TLinkItrs lnew(T->node.begin()+(*n), T->node[*n].neighbor.begin());
	  T->AssignChannel2Link(lnew,*ch);
	  lnew.second->Utot = T->computeUtot(lnew);
	  cout << "Aggiunto link (" << lnew.first->info->id << "," << lnew.second->info->id << ")[" << lnew.second->radio->channel << "]" << endl;
	}
    }
  }
#endif  
  
  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands;
  
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: E2CARA-TD: Checking E2E flows before beginning FAILED" << endl << endl;
  cout << "Demands:" << endl;
  for(map<pair<int,int>,double>::iterator m=demands.begin(); m!=demands.end(); m++)
    cout << "d[" << m->first.first << "," << m->first.second << "] = " << m->second << endl;

  /* FIRST STAGE: rearrange e2e flows to minimize the maximum Utot */
  MinMaxUtot mmu (T,&demands,&maxUtot);
  mmu.lp_solve();

  cout << "Back to E2CARA-TD after MinMaxUtot" << endl;

  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: E2CARA-TD: Checking E2E flows after MinMaxUtot FAILED" << endl << endl;

  
  /* FIRST STAGE: attempt to turn on radios to minimize Utot */

  
  /* compute a map that associates each node having radios that can be turned on with the set of its links */
  map<int, set<TLinkItrs> > nodeLinks;

  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
      if (T->Node(u)->Info()->countRadiosWithChannel(0))
	nodeLinks[T->Node(u)->Id()].insert(T->LinkItrs(T->Node(u),v));
      if (v->Info()->countRadiosWithChannel(0))
	nodeLinks[v->Id()].insert(T->LinkItrs(T->Node(u),v));
    }


  while (!nodeLinks.empty() && maxUtot > m_u0) {
    
    map<int,set<TLinkItrs> >::iterator r0 = max_element(nodeLinks.begin(), nodeLinks.end(), SumUtotPriority(T,m_u0));
    
    /* if the priority of the extracted node is null, it means that none of its links
     * is in the collision domain of a link with Utot > u0, hence turning on its radio is useless
     * Exit the loop because all the other nodes will have a zero priority 
     */
//     if (!calcPriority(T,u0,*r0))
//       break;

    
    map<unsigned int, double> S;            // pairs channel, maximum Utot
    map<unsigned int, double>::iterator maxS;
    TGraphNeighborIterator v;
    map<unsigned int,set<pair<int,TGraphNeighbor> > > addedNeighbors;       // set of neighbors added for each channel
    map<int,set<int> > neighbOn;        // for each channel stores the neighbor where an additional radio has been turned on
    int ptr_radio;

    /* find the channels of the extracted node and avoid to put them in S */
    set<unsigned int> r0Channels = T->Node(r0->first)->Info()->findChannels();
//     findChannels((T->node.begin()+r0->first), r0Channels);

    /* for each available channel not in use by the extracted node */
    for (unsigned int c=1; c<=WMS.numAvailChannels(); c++)
      if (r0Channels.find(c) == r0Channels.end()) {

	double meanrate = 0.;

	S[c] = 0.;

	cout << "Trying to turn on a radio of node " << r0->first << " on channel " << c << endl;

	/* find all the links that can be added on such channel */
	for (int u=0; u<T->GetNumNodes(); u++)
	  /* for each node u in the network (other than the extracted node) that can be a neighbor */
	  //if ((u->info->id != r0->first) && T->canBeNeighbors(T->node.begin()+r0->first,u,r,p,T->maxP+1) ) {
	  if ((u != r0->first) && (r = T->MaxAllowLinkRate(r0->first,u,WMS.defaultTxPowerLevel(),WMS.min_rate(),WMS.max_rate())) >= WMS.min_rate() ) {
	    /* if u has a radio on channel c  or an available radio*/
	    if (T->Node(u)->Info()->countRadiosWithChannel(c) || T->Node(u)->Info()->countRadiosWithChannel(0)) {
	      /* a link on channel c can be established between the extracted node and u,*/
	      addedNeighbors[c].insert(pair<int,TGraphNeighbor>(r0->first,TGraphNeighbor(T->Node(u)->Info(),WMS.availTXrates(r),WMS.defaultTxPowerLevel())));
	      /* a link on channel c can be established between u and the extracted node */
	      addedNeighbors[c].insert(pair<int,TGraphNeighbor>(u,TGraphNeighbor(T->Node(r0->first)->Info(),WMS.availTXrates(r),WMS.defaultTxPowerLevel())));

	      if (!T->Node(u)->Info()->countRadiosWithChannel(c))
		neighbOn[c].insert(u);
	    }
	  }

	/* add the neighbors computed above for the selected channel to the topology */
	for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[c].begin(); n!=addedNeighbors[c].end(); n++) {
	  v = AddNeighbor (T->Node(n->first), &n->second);
	  T->AssignChannel2Link(T->LinkItrs(T->Node(n->first),v), c);

	  S[c] = max(S[c], T->computeUtot(T->LinkItrs(T->Node(n->first),v)));
	  meanrate += v->Rate();
	}
	meanrate = meanrate / addedNeighbors[c].size();

	/* estimate the amount of flow that can be routed on the new links */
	if (S[c])
	  S[c] = (m_u0-S[c])*meanrate;

	/* remove the added links */
	for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[c].begin(); n!=addedNeighbors[c].end(); n++) {
	  v = T->Node(n->first)->FindNeighbor (n->second.Id(), c);

	  if (v != T->Node(n->first)->NeighborEnd())
	    EraseNeighbor (T->Node(n->first), v);
	}

	/* remove channel c (if it was assigned to) from the channels of the extracted node */
// 	ptr_radio = find_if(T->node[r0->first].info->radio, T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios,
// 		__gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	
	ptr_radio = T->Node(r0->first)->Info()->findRadioWithChannel(c);

// 	if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
	if (ptr_radio >= 0)
	  SetChannel (T->Node(r0->first)->Info()->Radio(ptr_radio), 0);
	
	for (set<int>::iterator n=neighbOn[c].begin(); n!=neighbOn[c].end(); n++) {
	  /* remove channel c (if it was assigned to) from the channels of the selected neighbors */
// 	  ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	  ptr_radio = T->Node(*n)->Info()->findRadioWithChannel(c);

// 	  if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	  if (ptr_radio >= 0)
	    SetChannel (T->Node(*n)->Info()->Radio(ptr_radio), 0);
	}

      }

    if (addedNeighbors.empty()) {
      nodeLinks.erase(r0);
      continue;
    }

    /* select the channel maximizing S */
    maxS = max_element(S.begin(), S.end(), compareSecond<unsigned int,double>());

    cout << "il canale selezionato e' " <<  maxS->first << endl;

    /* add the neighbors computed earlier for the selected channel to the topology before calling MinMaxUtot */
    for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++) {
      v = AddNeighbor (T->Node(n->first), &n->second);

      T->AssignChannel2Link(T->LinkItrs(T->Node(n->first),v),maxS->first);

      SetUtot (v, T->computeUtot(T->LinkItrs(T->Node(n->first),v)));
    }

    bool extrRadioOn;

    prevUtot = maxUtot;
    MinMaxUtot mmu2 (T,&demands,&maxUtot);
    mmu2.lp_solve();

    if (maxUtot < prevUtot) {
      extrRadioOn = true;

      cout << "** Turning on the radio decreased the max Utot to " << maxUtot << " **" << endl << endl;

      /* add the new links to the nodeLinks map to correctly compute the priority */
      for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++)
	if ( (v = T->Node(n->first)->FindNeighbor(n->second.Id(),maxS->first)) != T->Node(n->first)->NeighborEnd() )
	{
	  if ( nodeLinks.find(n->first) != nodeLinks.end() )
	    nodeLinks[n->first].insert(T->LinkItrs(T->Node(n->first), v));

	  if ( nodeLinks.find(n->second.Id()) != nodeLinks.end() )
	    nodeLinks[n->second.Id()].insert(T->LinkItrs(T->Node(n->first), v));
	}
    }
    else {
      extrRadioOn = false;
      /* the extracted radio can be turned off and the new links removed */
      for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++) {
	/* find the added link and remove it */
	v = T->Node(n->first)->FindNeighbor(n->second.Id(), maxS->first);

	if (v != T->Node(n->first)->NeighborEnd())
	  EraseNeighbor (T->Node(n->first), v);
	else
	  cerr << "** WARNING: an added link has not been found!" << endl << endl;
      }
      /* remove channel c from the channels of the extracted node */
//       ptr_radio = find_if(T->node[r0->first].info->radio.begin(), T->node[r0->first].info->radio.end(),
// 	      __gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
      ptr_radio = T->Node(r0->first)->Info()->findRadioWithChannel(maxS->first);

//       if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
      if (ptr_radio >= 0)
	SetChannel (T->Node(r0->first)->Info()->Radio(ptr_radio), 0);

      /* if turnOnNeighb, remove channel c from the channels of the neighbor */
	for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++) {
// 	  ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
	  ptr_radio = T->Node(*n)->Info()->findRadioWithChannel(maxS->first);

// 	  if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	  if (ptr_radio >= 0)
	    SetChannel (T->Node(*n)->Info()->Radio(ptr_radio), 0);
	}
    }

    /* if the radio has been turned on and there are other radios on the extracted node that can be turned on,
     * leave the node in the queue, so that the other radios can be turned on */
    if ( !extrRadioOn || !(T->Node(r0->first)->Info()->countRadiosWithChannel(0)) )
      nodeLinks.erase(r0);

    /* if turnOnNeighb, if the neighbor has no more available radios, remove it from the queue */
    if (extrRadioOn)
      for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++)
	if (!T->Node(*n)->Info()->countRadiosWithChannel(0) && (nodeLinks.find(*n)!=nodeLinks.end()))
	  nodeLinks.erase(*n);
  }


  /* Ok, we tried our best to keep Utot <= u0 for every link 
     Now, let's try to switch some radio off, while not 
     introducing any other violation to the constraint Utot <= u0 */

  cout << endl << "** STARTING SECOND STAGE **" << endl << endl;


  map<TRadio, set<TLinkItrs> > Lr;  // maps each radio to the set of links using it
  map<TRadio, set<TLinkItrs> >::iterator minLr;
  int ri;


  /* compute Lr for all the radios */
  for (unsigned int i=0; i<T->GetNumNodes(); i++)
    for (TGraphNeighborIterator v=T->Node(i)->NeighborBegin(); v!=T->Node(i)->NeighborEnd(); v++)
    {
//       ch = SearchChannel(v->info, v->radio->channel);
      ri = v->Info()->findRadioWithChannel(v->RadioChannel());
      /* insert the link in the sets of the two radios at the endpoints of the link */
      Lr[TRadio(i,distance(T->Node(i)->Info()->Radio(0),RadioIfaceIterator(v->Radio())))].insert(T->LinkItrs(T->Node(i),v));
      if (ri >= 0)
	Lr[TRadio(v->Id(), ri)].insert(T->LinkItrs(T->Node(i),v));
    }


  while (!Lr.empty()) {
    
    /* DEBUG */
    for (map<TRadio, set<TLinkItrs> >::iterator r=Lr.begin(); r!=Lr.end(); r++) {
      cout << "Links using the radio of node " << r->first.first << " set to channel " << T->Node(r->first.first)->Info()->Radio(r->first.second)->Channel() << endl;
      for (set<TLinkItrs>::iterator l=r->second.begin(); l!=r->second.end(); l++)
	cout << l->first->Id() << "->" << l->second->Id() << "[" << l->second->RadioChannel() << "]" << endl;
      cout << endl;
    }

    /* extract the pair (node,radio) with the minimum flow */
    minLr = Lr.begin();
    double minP = sumFlows(minLr->second), temp;

    for (map<TRadio, set<TLinkItrs> >::iterator i=Lr.begin(); i!=Lr.end(); i++)
      if ( (temp=sumFlows(i->second)) < minP ) {
	minP = temp;
	minLr = i;
      }

    cout << "** Extracting node " << minLr->first.first << " radio " << minLr->first.second << "(channel " << T->Node(minLr->first.first)->Info()->Radio(minLr->first.second)->Channel() << ") flow " << minP << " **" << endl;


    /* if TurnOffRadio returns true, the radio can be turned off and all the links in Lr can be removed from the topology */
    TurnOffRadio tor (T, &demands, &(minLr->second), m_u0);
    tor.lp_solve();
    if (!minP || tor.Succeeded()) {
//    if (TurnOffRadio(T, &demands, &(Lr[minQr->first]), u0)) {
      cout << "** EUREKA ** turning off radio " << minLr->first.second << " of node " << minLr->first.first << "(channel " << T->Node(minLr->first.first)->Info()->Radio(minLr->first.second)->Channel() << ")  **" << endl;
      /* for all the links in minLr */
      for (set<TLinkItrs>::iterator l=minLr->second.begin(); l!=minLr->second.end(); l++) {
	if (l->second->Flow())
	  cerr << "WARNING: TurnOffRadio is removing a link with non-null flow!" << endl;
	/* first, remove the link from the Lr set of the other radio (not the extracted one), if it is still in Lr */
// 	ch = SearchChannel(l->second->info, l->second->radio->channel);
	ri = l->second->Info()->findRadioWithChannel(l->second->RadioChannel());
	if (minLr->first.first == l->second->Id()) {
	  if (Lr.find(TRadio(l->first->Id(), distance(l->first->Info()->Radio(0),RadioIfaceIterator(l->second->Radio())))) != Lr.end())
	    Lr[TRadio(l->first->Id(), distance(l->first->Info()->Radio(0),RadioIfaceIterator(l->second->Radio())))].erase(*l);
	}
	else if (ri >= 0) {
	  if (Lr.find(TRadio(l->second->Id(), ri)) != Lr.end())
	    Lr[TRadio(l->second->Id(), ri)].erase(*l);
	}
	/* then, remove the link from the topology*/
	EraseNeighbor (T->Node(l->first->Id()), l->second);
      }
      /* switch off the radio (set channel to 0) */
      SetChannel (T->Node(minLr->first.first)->Info()->Radio(minLr->first.second), 0);
    }

    /* remove the whole Lr set of the extracted radio */
    Lr.erase(minLr);

    // T->printTopology();  // DEBUG
  }

  /* erase links with null flow */
  //T->eraseLinksWithNullFlow();

  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: E2CARA-TD: Checking E2E flows at the end FAILED" << endl << endl;

  return 0;
}





void E2CAR_OPT::Configure (const list<pair<string,string> >& params)
{
  // default values
  m_relax = false;
  m_u0 = 0.5;

  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("relax", m_relax);
    RP.read ("u0", m_u0);

    if (RP.notFound())
      cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the E2CAR_OPT algorithm" << endl << endl;
  }
}


double E2CAR_OPT::Run (Topology* T)
{
  /* clone the topology because MinMaxUtot_MILP removes links with null flow */
  Topology* U = new Topology(*T);

  /* find the minimum maxUtot */
  double maxUtot;

  //MinMaxUtot_MILP(U, maxUtot, relax, NULL, u0);
  MinMaxUtot_MILP mmu(U, NULL, 7200);  // max 2 hours

  mmu.set_breakValue(m_u0);

  mmu.lp_solve();

  maxUtot = mmu.getObjValue();

  U->printStats("MinMaxUtot_MILP",true);

  MinRadios_MILP mr(T,max(maxUtot,m_u0));

  if (mr.lp_solve())
    return -1;

  delete U;

  return 0;
}
