#include "CAalgos.h"
#include "WMutils.h"

#include <math.h>
#include <iostream>
#include <algorithm>
#include <numeric>
#include <functional>
#include <cassert>





void FCPRA::Configure (const list<pair<string,string> >& params)
{
  // default values
  m_optimize = true;
  m_turbo = true;
  m_ratepoweradapt = false;

  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("optimize", m_optimize);
    RP.read ("turbo", m_turbo);
    RP.read ("rate_power_adapt", m_ratepoweradapt);

    if (RP.notFound())
    {
      cerr << endl << "Ignoring unknown attribute (" << p->first << ") for the FCPRA algorithm" << endl << endl;
    }
  }
}


void FCPRA::SetName ()
{
  stringstream ss;

  ss << "F" << (m_ratepoweradapt ? "T" : "F");

  ss >> m_name;
}


/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
double FCPRA::Run (Topology *T)
{
	/* INITIALIZATION */
	T->ResetChannels();

	//for each node, one of the radios is assigned channel 1
	for (int u = 0; u < T->GetNumNodes(); u++)
		SetChannel (T->Node(u)->Info()->Radio(0), 1);

	// each node can reach its neighbors through that radio
	// (this implies that the channel of each link is set to 1)
	for (int u = 0; u < T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v = T->Node(u)->NeighborBegin(); v != T->Node(u)->NeighborEnd(); v++)
			SetRadio (v, &(*T->Node(u)->Info()->Radio(0)));

	// compute Utot for each link (initially the collision domain equals the neighborhood)
	// for every directed link u->v
	for (int u = 0; u < T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v = T->Node(u)->NeighborBegin(); v != T->Node(u)->NeighborEnd(); v++)
			SetUtot (v, T->computeUtot(T->LinkItrs(T->Node(u),v)));

	// Set containing a pointer to all the directed links
	set<TLinkItrs> Q;
	for (int u = 0; u < T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v = T->Node(u)->NeighborBegin(); v != T->Node(u)->NeighborEnd(); v++)
				Q.insert(T->LinkItrs(T->Node(u),v));

	/* END INITIALIZATION */

	set<TLinkItrs>::iterator ptr_l;
	double max_Utot;
	bool ch1removed = false;
	unsigned int l_channel;
	TLinkItrs *invLink;

	while (!Q.empty())
	{
		// Extract the link with the maximum Utot from Q
		max_Utot = -1.;
		for (set<TLinkItrs>::iterator p=Q.begin(); p!=Q.end(); p++)
			if (p->second->UTot() > max_Utot)
			{
				max_Utot = p->second->UTot();
				ptr_l = p;
			}

		// Erase the extracted link from Q
		TLinkItrs l (*ptr_l);
		Q.erase(ptr_l);
		l_channel = l.second->Radio()->Channel();
		if (m_optimize)
			ch1removed = fcra_OptimizeRadios(T, l, l.first->Id(), Q) || fcra_OptimizeRadios(T, l, l.second->Id(), Q);

		/* if the end nodes only share channel 1 and the optimization step removes channel 1 from one of the nodes
		   remove the inverse link (by assigning channel 0) and updates the Utot of the links involved */
		if (ch1removed)
		{
			invLink = new TLinkItrs (T->FindInverseLink(l));
			RadioIface *fakeradio = new RadioIface();
			SetRadio (invLink->second, fakeradio);
			SetChannel (invLink->second->Radio(), (unsigned int)1);
			fcra_computeNewUtot(T, *invLink, 0);
			SetChannel (invLink->second->Radio(), 0);
		}

		set<unsigned int> uChannels, vChannels, S, Sc;
		S.clear();
		Sc.clear();
		// Find the set of channels of the end nodes of the link to be extracted
		uChannels = l.first->Info()->findChannels();
		vChannels = l.second->Info()->findChannels();

		// Consider all the cases
		// if both end nodes have available radios
		if ( uChannels.size()<l.first->Info()->numAvailRadios() && vChannels.size()<l.second->Info()->numAvailRadios() )
			// put all the channels in S
			for (unsigned int i=1; i<=WMS.numAvailChannels(); S.insert(i++));
		// if u has all radios assigned a channel
		else if ( uChannels.size()==l.first->Info()->numAvailRadios() && vChannels.size()<l.second->Info()->numAvailRadios() )
			// put the channels of u in S
			copy(uChannels.begin(), uChannels.end(), inserter(S, S.begin()));
		// if v has all radios assigned a channel
		else if ( uChannels.size()<l.first->Info()->numAvailRadios() && vChannels.size()==l.second->Info()->numAvailRadios() )
			// put the channels of v in S
			copy(vChannels.begin(), vChannels.end(), inserter(S, S.begin()));
		// if all the radios of both end nodes have been assigned a channel
		else
			// put the common channels of u and v in S
			set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

		// S is now the set of candidate channels to be assigned to the link to be extracted
		assert( !S.empty() );

		// put the common channels of u and v in Sc
		set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(Sc, Sc.begin()));

		map<unsigned int, double> Umax, Umax2;
		map<unsigned int, set<TLinkItrs> > D1;

		// for each channel c in S
		for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
			Umax[*c] = (Umax2[*c] = 0.);

		// compute Umax[c] for all c in S
		// for every directed link x->y
		for (int x = 0; x < T->GetNumNodes(); x++)
			for (TGraphNeighborIterator y = T->Node(x)->NeighborBegin(); y != T->Node(x)->NeighborEnd(); y++)
				// whose potential collision domain includes the extracted link and having a channel in S
				if ((T->LinkItrs(T->Node(x),y) != l) && (l.PotInterferesWith(T->LinkItrs(T->Node(x),y))) && (S.find(y->RadioChannel())!=S.end()))
				{
					/* remove the contribution of the extracted link from the Utot of the links having the same channel */
					if (y->RadioChannel() == l_channel)
						SetUtot (y, y->UTot() - l.second->Flow()/l.second->Rate());
					if (y->UTot() > Umax[y->RadioChannel()])
						Umax[y->RadioChannel()] = y->UTot();
					/* add the link to D1 */
					D1[y->RadioChannel()].insert(T->LinkItrs(T->Node(x),y));
				}

		map<unsigned int, float> rate;
		map<unsigned int, unsigned int> power;
		int m, mmin, w, wmin;
		float save_rate;
		unsigned int save_power;
		double min,save;

		/* we use a fake radio to compute Utot(u->v) for every channel */
		RadioIface *fakeradio = new RadioIface();
		SetRadio (l.second, fakeradio);

		/* we need to save the power and rate of the original link, in order to restore them before evaluating a new channel */
		save_rate = l.second->Rate();
		save_power = l.second->TxPower();

		// for each channel c in S
		for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
		{
			// we overwrite the channel and the rate of the extracted link to compute its Utot correctly
			SetChannel (l.second->Radio(), *c);
			SetRate (l.second, save_rate);
			SetTxPower (l.second, save_power);

// 			for (w=0; w<NUMTXPWR && availTxPwr[w]!=l.second->power; w++);
// 			assert(w<NUMTXPWR);
			w = save_power;

			wmin = w;
			min = T->computeUtot(l) - l.second->Flow()/l.second->Rate();

			while (m_ratepoweradapt  &&  w<WMS.numTxPowerLevels()-1  &&  min>Umax[*c]  &&  Umax2[*c]<Umax[*c] )
			{
				SetTxPower (l.second, ++w);

				// compute Umax2[c]
				// for every directed link x->y
				for (int x = 0; x < T->GetNumNodes(); x++)
					for (TGraphNeighborIterator y = T->Node(x)->NeighborBegin(); y != T->Node(x)->NeighborEnd(); y++)
						// not in D1 and whose potential collision domain includes the extracted link
						if ( (T->LinkItrs(T->Node(x),y) != l) && (l.PotInterferesWith(T->LinkItrs(T->Node(x),y)))
						  && (D1[*c].find(T->LinkItrs(T->Node(x),y)) == D1[*c].end()) && (y->RadioChannel() == *c) )
						{
							if (y->UTot() > Umax2[*c])
								Umax2[*c] = y->UTot();
						}

				if ( (save = T->computeUtot(l) - l.second->Flow()/l.second->Rate()) < min  && Umax2[*c] < Umax[*c] )
				{
					min = save;
					wmin = w;
				}
			}
			// this power will be used in case a link u-> v is established on this channel
			SetTxPower (l.second, (power[*c] = wmin) );
			
			if (save_power != wmin)
				cout << endl << "** Varying power on link " << l.first->Id() << "->" << l.second->Id() 
				     << " ** from "  << WMS.TxPowerLevel(save_power) << " to " << WMS.TxPowerLevel(wmin) << endl;

			for (m=0; m<WMS.numTXrates() && WMS.availTXrates(m)!=l.second->Rate(); m++);
			assert(m<WMS.numTXrates());

			mmin = m;

			while (m_ratepoweradapt && (m>0) && (min>Umax[*c]))
			{
				SetRate (l.second, WMS.availTXrates(--m));
				if ( (save = T->computeUtot(l) - l.second->Flow()/l.second->Rate()) < min)
				{
					min = save;
					mmin = m;
				}
			}
			// this rate will be used in case a link u-> v is established on this channel
			rate[*c] = WMS.availTXrates(mmin);
			
			if (save_rate != WMS.availTXrates(mmin))
				cout << endl << "** Varying rate on link " << l.first->Id() << "->" << l.second->Id() 
				     << " ** from "  << save_rate << " to " << WMS.availTXrates(mmin) << endl;

			if (min > Umax[*c])
				Umax[*c] = min;
		}

		delete fakeradio;

		set<unsigned int> C;
		set_difference(S.begin(), S.end(), Sc.begin(), Sc.end(), inserter(C, C.begin()));

		/* leave in Umax only the channels in Sc and the channel in S-Sc that minimizes Umax */
		if (!C.empty())
		{
			set<unsigned int>::iterator c = C.begin();
			min = Umax[*c];
			for (set<unsigned int>::iterator c0=++C.begin(); c0!=C.end(); c0++)
				if (Umax[*c0]<min)
				{
					Umax.erase(*c);
					c = c0;
					min = Umax[*c0];
				} 
				else
					Umax.erase(*c0);
		}

		double F = l.second->Flow();
		TGraphNeighbor *newlink;
		list<TGraphNeighbor>::iterator currlink;

		MeshNode* backup_info = l.second->Info();
		double backup_flow = l.second->Flow();
		/* save the list of end-to-end flows crossing the extracted link */
		map<pair<int,int>,double> backup_e2e;

		for (FlowDemandsIterator fd = l.second->F_e2eBegin(); fd != l.second->F_e2eEnd(); fd++)
			backup_e2e.insert (*fd);

		/* remove v from among u's neighbors, i.e., remove the extracted link */
		EraseNeighbor (l.first, l.second);

		unsigned int c;
		float sumrates;
		double base, I;
		bool done = false;


		do
		{
			/* add a new link u-> v on the channel that minimizes Umax */
			newlink = new TGraphNeighbor(backup_info, 0, 0);  // rate and power are set below

			currlink = AddNeighbor (l.first, newlink);
// 			AddNeighbor (l.first, newlink);
// 			currlink = --l.first->NeighborEnd();

			c = min_element(Umax.begin(), Umax.end(), compareSecond<unsigned int, double>())->first;

			T->AssignChannel2Link( T->LinkItrs(l.first,currlink), c );
			SetRate (currlink, rate[c]);            /* set the previously computed rate */
			SetTxPower (currlink, power[c]);        /* set the previously computed power */
// 			currlink->f_e2e = backup_e2e;		/* restore the set of end-to-end flows and reset the values*/
// 			for (map<pair<int,int>,double>::iterator mp=currlink->f_e2e.begin(); mp!=currlink->f_e2e.end(); mp++)
// 			  mp->second = 0.;

			ClearFlowDemands (currlink);            /* restore the set of end-to-end flows and reset the values*/
			for (FlowDemandsIterator fd = backup_e2e.begin(); fd != backup_e2e.end(); fd++)
				SetFlowDemand (currlink, fd->first, 0.);

			sumrates = 0.;
			for (TGraphNeighborIterator v = l.first->NeighborBegin(); v != l.first->NeighborEnd(); v++)
				if (v->Id() == currlink->Id())
					sumrates+=v->Rate();

			base = Umax[c];
			Umax.erase(c);

			if ( (!Umax.empty()) && (F>=(min_element(Umax.begin(),Umax.end(),compareSecond<unsigned int, double>())->second - base)*sumrates) )
				I = min_element(Umax.begin(),Umax.end(),compareSecond<unsigned int, double>())->second - base;
			else
			{
				I = F / sumrates;
				done = true;
			}

			for (TGraphNeighborIterator k = l.first->NeighborBegin(); k != l.first->NeighborEnd(); k++)
				if (k->Id() == currlink->Id())
				{
					SetFlow (k, k->Flow() + I * k->Rate());
					for (FlowDemandsIterator mp = k->F_e2eBegin(); mp != k->F_e2eEnd(); mp++)
// 					  mp->second += (I * k->Rate()) * backup_e2e[mp->first] / backup_flow;
					  SetFlowDemand (k, mp->first, mp->second + (I * k->Rate()) * backup_e2e[mp->first] / backup_flow);
					F -= I * k->Rate();
				}

		} while (!done);

		if (ch1removed)
		// assign the same channel as the extracted link to the inverse link
		{
			fcra_computeNewUtot(T, *invLink, c);
			T->AssignChannel2Link( *invLink, c );
			SetUtot (invLink->second, T->computeUtot(*invLink));
		}

		// compute the Utot for the newly established links
		for (TGraphNeighborIterator k = l.first->NeighborBegin(); k != l.first->NeighborEnd(); k++)
			if (k->Id() == currlink->Id())
				SetUtot (k, T->computeUtot(T->LinkItrs(l.first,k)));

		// Update the Utot for all the other links involved
		// for every directed link x->y
		for (int x = 0; x < T->GetNumNodes(); x++)
			for (TGraphNeighborIterator y = T->Node(x)->NeighborBegin(); y != T->Node(x)->NeighborEnd(); y++)
				// for every newly establisthed link u->v
				for (TGraphNeighborIterator k = l.first->NeighborBegin(); k != l.first->NeighborEnd(); k++)
					if ( (k->Id() == currlink->Id()) && ( !ch1removed || T->LinkItrs(T->Node(x),y) != *invLink ) &&
					// such that the potential collision domain of x->y includes the new link and have the same channel
					     (T->Node(x) != l.first || y != k) && (T->LinkItrs(l.first,k).PotInterferesWith(T->LinkItrs(T->Node(x),y))) && (y->RadioChannel()==k->RadioChannel()) )
						/* add the contribution of the new link to the Utot of x->y */
						SetUtot (y, y->UTot() + k->Flow()/k->Rate());

		if (ch1removed)
			delete invLink;
	}
	return 0;
}



/* This function is invoked when we extract link l and want to optimize the radios of i (i.e., remove channel 1)
   */
bool FCPRA::fcra_OptimizeRadios(Topology *T, const TLinkItrs& l, int i, set<TLinkItrs> Q)
{
	bool invremoved = false;
	set<TLinkItrs> L, extracted, nonextracted;
	set<pair<TLinkItrs,unsigned int> > tomodify;

	set<unsigned int> iChannels, lChannels, S;
	iChannels = T->Node(i)->Info()->findChannels();

	/* if i has available radios or no radio is assigned to 1, do nothing */
	if ( (iChannels.size()<T->Node(i)->Info()->numAvailRadios()) || (find(iChannels.begin(),iChannels.end(),(unsigned int)1)==iChannels.end()) )
		return false;

	/* insert all the links leaving or entering i into L */
	for (int x = 0; x < T->GetNumNodes(); x++)
		for (TGraphNeighborIterator y = T->Node(x)->NeighborBegin(); y != T->Node(x)->NeighborEnd(); y++)
			if ( T->Node(x)->Id() == i || y->Id() == i )
				L.insert(T->LinkItrs(T->Node(x),y));

	L.erase(l);

	/* compute L-Q, i.e., the links that have already been extracted */
	set_difference(L.begin(), L.end(), Q.begin(), Q.end(), inserter(extracted, extracted.begin()));

	/* if an extracted link has been assigned channel 1, we can do nothing */
	for (set<TLinkItrs>::iterator l0=extracted.begin(); l0!=extracted.end(); l0++)
		if ((*l0).second->RadioChannel() == 1)
			return false;

	/* compute L\cap Q, i.e., the links that have not been extracted yet */
	set_difference(L.begin(), L.end(), extracted.begin(), extracted.end(), inserter(nonextracted, nonextracted.begin()));

	list<TGraphNeighbor>::iterator invL;
	if (m_turbo && (invL = T->Node(l.second->Id())->FindNeighbor(l.first->Id())) != T->Node(l.second->Id())->NeighborEnd() )
	{
		bool inverse = ( find(nonextracted.begin(), nonextracted.end(), T->LinkItrs(T->Node(l.second->Id()), invL)) != nonextracted.end() );
		if (inverse)
		{
			iChannels = T->Node(l.first->Id())->Info()->findChannels();
			lChannels = T->Node(l.second->Id())->Info()->findChannels();
			S.clear();
			set_intersection(iChannels.begin(), iChannels.end(), lChannels.begin(), lChannels.end(), inserter(S, S.begin()));
			S.erase(1);
			if (S.empty())
			{
				nonextracted.erase(T->LinkItrs(T->Node(l.second->Id()), invL));
				invremoved = true;
			}
		}
	}

	/* if a nonextracted link has no channel but 1 in common with i, we can do nothing */
	for (set<TLinkItrs>::iterator l0=nonextracted.begin(); l0!=nonextracted.end(); l0++)
	{
		S.clear();
		// Find the set of channels of the end nodes of the link to be extracted
		iChannels = l0->first->Info()->findChannels();
		lChannels = l0->second->Info()->findChannels();

		set_intersection(iChannels.begin(), iChannels.end(), lChannels.begin(), lChannels.end(), inserter(S, S.begin()));
		S.erase(1);
		if (!S.empty())
			tomodify.insert(pair<TLinkItrs,unsigned int>(*l0,*S.begin()));
		else
			return false;
	}

	// if we reach this point, it means that we can remove channel 1 from the radios of i
	/* for each link to be modified */
	for (set<pair<TLinkItrs,unsigned int> >::iterator l0=tomodify.begin(); l0!=tomodify.end(); l0++)
	{
		if (l0->first.second->RadioChannel() != l0->second)
		{
			/* update Utot for neighbors*/
			fcra_computeNewUtot(T, l0->first, l0->second);
			/* assign the new channel to l0 */
			T->AssignChannel2Link(l0->first,l0->second);
			/* compute Utot for the modified link*/
			SetUtot (l0->first.second, T->computeUtot(l0->first));
		}
	}

	/* set to channel 0 the radio of i that is set to channel 1, i.e.,make it available */
	int ri = T->Node(i)->Info()->findRadioWithChannel (1);

	assert ( ri >= 0 );

	SetChannel (T->Node(i)->Info()->Radio (ri), 0);

	return invremoved;
}


/* Compute the new Utot for every link involved (except l itself) in case c is assigned to l */
void FCPRA::fcra_computeNewUtot(Topology *T, const TLinkItrs& l, unsigned int c)
{
	/* the new Utot for the other links is obtained from the current Utot value */
	// for every directed link x->y
	for (int x = 0; x < T->GetNumNodes(); x++)
		for (TGraphNeighborIterator y = T->Node(x)->NeighborBegin(); y != T->Node(x)->NeighborEnd(); y++)
			// such that the potential collision domain of x->y includes l
			if ( (l!=T->LinkItrs(T->Node(x),y)) && (l.PotInterferesWith(T->LinkItrs(T->Node(x),y))) )
			{
				/* by subtracting the F/C ratio of l if l escapes the collision domain of l0 */
				if ((y->RadioChannel() == l.second->RadioChannel()) && (y->RadioChannel() != c))
					SetUtot (y, y->UTot() - l.second->Flow() / l.second->Rate());
				/* by adding the F/C ratio of l if l joins the collision domain of l0 */
				else if ((y->RadioChannel() != l.second->RadioChannel()) && (y->RadioChannel() == c))
					SetUtot (y, y->UTot() + l.second->Flow() / l.second->Rate());
			}
}

