#ifndef WIMESH_MILP
#define WIMESH_MILP

#include "WMtypes.h"
#include <OsiClpSolverInterface.hpp>
#include <CbcModel.hpp>



/* node id-radio id pair to represent a radio */
typedef pair<int,int> TRadio;


namespace WMMILP {
/* counter function objects
 * it is initialized at init and increased by incr each time it is invoked
 * Can be used with generate to produce sequences like 1,2,3,4,5... (default behavior)
 */
template <class T>
class counter {
private:
  T curr;
  T incr;
public:
  counter(T _init = 0, T _incr = 1) : curr(_init), incr(_incr) {}
  T operator()() { return (curr += incr); }
};
}


class sub_grad_method;


/* Abstract base class for MILPs */

class MILP
{
public:
  // constructor
  // if _timeout is negative, set_timeout is not invoked
  MILP (Topology* _T, long _timeout, bool _apply = true)
    : T(_T),
      timeout(_timeout),
      apply(_apply),
      hasBreakValue(false),
      constrGroupSize(vector<int>(1,0))
  {}

  // returns the result of solve()
  virtual int lp_solve ();

  // returns true if a feasible solution has been found
//   virtual int lag_solve (const set<int>& torelax, REAL mu_0, int max_iter, sub_grad_method* sgm = 0, MILP_SOLVER ms = LpSolve);
  virtual int lag_solve (const set<int>& torelax, double mu_0, int max_iter, sub_grad_method* sgm = 0);

  //virtual void print () { print_solution(lp,4); }

  // save the program in MPS format
//   virtual void saveMPS (char* filename);

  virtual void set_breakValue (double bvalue) { breakValue = bvalue; hasBreakValue = true; }
  
  // makes sense if a feasible solution has been found
  virtual double getObjValue () { return obj_value; }

  virtual double getBound () { return bound; }
  
  // returns the number of constraint groups (useful for lagrangian relaxation)
  int getConstrGroupCount () { return constrGroupSize.size()-1; }

  // destructor
  virtual ~MILP ();

protected:
//   virtual void set_params ();   // can be overridden to set lp parameters such as precision

  // apply the solution of the program
  virtual void updateFlows (const double* ptr_var, const double* ptr_constr = 0) = 0;

  /* A list of common operations performed by linear programs */

//   void SetRadio (TGraphNeighborIterator ni, RadioIface* radio) { TGraphNeighborFriend::SetRadio (ni, radio); }
  void SetFlow (TGraphNeighborIterator ni, double f) { TGraphNeighborFriend::SetFlow (ni, f); }
  void SetUtot (TGraphNeighborIterator ni, double utot) { TGraphNeighborFriend::SetUtot (ni, utot); }
//   void SetRate (TGraphNeighborIterator ni, float rate) { TGraphNeighborFriend::SetRate (ni, rate); }
//   void SetTxPower (TGraphNeighborIterator ni, unsigned int txpower) { TGraphNeighborFriend::SetTxPower (ni, txpower); }
//   void SetAvailBW (TGraphNeighborIterator ni, double bw) { TGraphNeighborFriend::SetAvailBW (ni, bw); }
  void ClearFlowDemands (TGraphNeighborIterator ni) { TGraphNeighborFriend::ClearFlowDemands (ni); }
  void SetFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::SetFlowDemand (ni, d, f); }
  void IncreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::IncreaseFlowDemand (ni, d, f); }
  void DecreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::DecreaseFlowDemand (ni, d, f); }

  TGraphNeighborIterator AddNeighbor (TGraphNodeIterator nodei, const TGraphNeighbor* ni) { return TGraphNodeFriend::AddNeighbor (nodei, ni); }
  void EraseNeighbor (TGraphNodeIterator nodei, TGraphNeighborIterator ni) { TGraphNodeFriend::EraseNeighbor (nodei, ni); }

  void SetChannel (RadioIfaceIterator ri, unsigned int channel) {  RadioIfaceFriend::SetChannel (ri, channel); }
  void SetChannel (RadioIface* ri, unsigned int channel) { RadioIfaceFriend::SetChannel (RadioIfaceIterator(ri), channel); }

  OsiClpSolverInterface* solver;

  double obj_value;
  double bound;

  Topology* T;
  map<pair<int,int>,double> demands;
  unsigned int numDemands;
  double breakValue;

  long timeout;
  bool apply;        // true means that the changes will be applied to the topology
  bool hasBreakValue;
  vector<int> constrGroupSize;
  set<int> relaxedConstrGroups;   // set of the groups of constraints relaxed by lag_solve
};


/*
 * Assign channels and rearrange e2e flows of given demands to minimize the maximum Utot
 * based on a Mixed Integer Linear Program
 * Topology::ResetChannels() is called to remove duplicate neighbors in TGraphNeighbor lists
 * and set the channel of every radio to zero
 * The search stops if the maximum total utilization goes below u0 (this is currently
 * disabled because it is dangerous with lagrangian relaxation - can be enabled in set_params)
 * If apply is true and a feasible solution has been found, channels, links, flows and Utots
 * are updated
 */
class MinMaxUtot_MILP : public MILP
{
public:
  // constructor, which builds the model
  MinMaxUtot_MILP (Topology* _T, vector<float> *nodePowerConstr, long _timeout = -1, bool _apply = true);

private:
  // apply the solution of the program (if ptr_var is null, get_ptr_variables is called)
  virtual void updateFlows (const double* ptr_var, const double* ptr_constr = 0);
};

/*
 * Assign channels and rearrange e2e flows of given demands to minimize the total power consumption
 * based on a Mixed Integer Linear Program
 * Topology::ResetChannels() is called to remove duplicate neighbors in TGraphNeighbor lists
 * and set the channel of every radio to zero
 * If apply is true and a feasible solution has been found, channels, links, flows and Utots
 * are updated
 */
class MinPower_MILP : public MILP
{
public:
  // constructor, which builds the model
  MinPower_MILP (Topology* _T, double _maxUtot, vector<float>* nodePowerConstr, long _timeout = -1, bool _apply = true);

private:
  // apply the solution of the program (if ptr_var is null, get_ptr_variables is called)
  virtual void updateFlows (const double* ptr_var, const double* ptr_constr = 0);

  double maxUtot;
};


/*
 * Assign channels and rearrange e2e flows of given demands to minimize the number
 * of radios turned on given a constraint on the maximum total utilization
 * Based on a Mixed Integer Linear Program
 * ResetChannels is called to avoid duplicate neighbors in TGraphNeighbor lists and
 * reset the channel of every radio to zero (they are set after solving the MILP)
 * Returns true if the model is infeasible, false otherwise
 * The current configuration of T is used as initial basis
 */
class MinRadios_MILP : public MILP
{
public:
  MinRadios_MILP (Topology* _T, double u0, long _timeout = -1, bool _apply = true);

private:
  // apply the solution of the program (if ptr_var is null, get_ptr_variables is called)
  virtual void updateFlows (const double* ptr_var, const double* ptr_constr = 0);

};


/*
 * Rearrange e2e flows of given demands to minimize the maximum Utot
 * while maximizing the number of radios turned off
 * based on a Mixed Integer Linear Program
 * Changes are applied only if the model is valid (M0 != 0)
 */
class MRMU_MILP : public MILP
{
public:
  // constructor, which builds the model
  MRMU_MILP (Topology* _T, long _timeout = -1, bool _apply = true);

private:
  // apply the solution of the program (if ptr_var is null, get_ptr_variables is called)
  virtual void updateFlows (const double* ptr_var, const double* ptr_constr = 0);

  int Ncol;
  vector<unsigned int> cumNumNeighbs, cumNumRadios;
  map<TRadio, set<TLinkItrs> > Lr;  // maps each radio to the set of links using it
};


  
/** Lagrangian Relaxation **/
/* TODO Test after conversion to Coinor-Clp */

class lag_constraint
{
public:
//   lag_constraint(REAL* _row, int* _colno, int len, int _type, REAL _rhs, REAL _mu)
//     :  row(vector<REAL>(_row,_row+len)),
//        colno(vector<int>(_colno,_colno+len)),
//        type(_type),
//        rhs(_rhs),
//        mu(_mu)
//     {}

  lag_constraint(double* _row, int* _colno, int len, double _lower, double _upper, double _mu)
    :  row(vector<double>(_row,_row+len)),
       colno(vector<int>(_colno,_colno+len)),
       lower(_lower),
       upper(_upper),
       mu(_mu)
    {}

  lag_constraint() {}

  vector<double> row;
  vector<int> colno;
  double lower, upper;
//   int type;
//   REAL rhs;
  double mu;
};


/* Attempts to solve a given (MI)LP by using Lagrangian Relaxation */
/* The defaul subgradient method is 1/k. To change the default, call set_sgm */
class Lag_Relax
{
public:
  /* Constructor. Input parameters:
   * - lp: an already built (MI)LP model 
   * - relax_constr: the set of constraints to be relaxed
   * - mu_0: the initial value for the lagrangian multipliers
   * - max_iter: the maximum number of Lagrangian subproblems to solve
   */ 
//   Lag_Relax (lprec* lp, set<int> relax_constr, REAL mu_0, int max_iter, int max_noimprove = 5);
  Lag_Relax (ClpSimplex* model, set<int> relax_constr, double mu_0, int max_iter, int max_noimprove = 5);

  void set_sgm (sub_grad_method* sgm);

//   void set_solver (MILP_SOLVER ms) { m_solver = ms; }

  void set_breakValue (double bvalue) { m_breakValue = bvalue; m_hasBreakValue = true; }

  void solveLAG ();

  bool haveSolution () { return m_have_solution; }
  bool haveBound () { return m_have_bound; }

  double* getSolution () { return m_solution; }
  double getObjValue () { return m_obj_val; }
  double getBound () { return m_lag_bound; }

  /* destructor */
  ~Lag_Relax ();

private:

  /* update the objective function with the current values of the multipliers */
  void update_obj_func ();
  
  /* update the lagrangian multipliers */
  void update_multipliers (bool failed = false, bool noimprove = false);

  /* solves a single lagrangian subproblem */
  void solveSubProblem (bool &foundSolution, bool &foundOptSolution, double &objValue, double &bound);

//   lprec* m_lp;         // the linear program. NOTE that the program is modified by the constructor
  ClpSimplex* m_model;   // the linear program. NOTE that the program is modified by the constructor
  int m_curr_iter;     // number of completed iterations
  int m_max_iter;      // maximum number of Lagrangian subproblems to solve
  int m_max_noimprove;
  bool m_hasBreakValue;
  double m_breakValue;     // stop the search if a feasible solution with a value better than this one is found
  vector<double> m_obj_func;                 // original objective function
  vector<lag_constraint> m_relax_constr;   // relaxed constraints
  bool m_have_solution;   // true if a feasible solution for the original problem has been found
  double* m_solution;       // best feasible solution found for the original problem (values of the variables)
  double m_obj_val;         // value of the objective function (of the original problem) for the best solution found
  bool m_have_bound;      // true if at least one lagrangian subproblem has been solved
  double m_lag_bound;       // best upper (or lower) bound found by solving the lagrangian subproblems
  
  sub_grad_method* m_sgm; // pointer to the method used to update the multipliers
//   MILP_SOLVER m_solver;

  // used to update lagrangian multipliers
  const double* m_curr_solution;    // solution of the current lagrangian subproblem
  double m_curr_lag;                  // objective function value of the solution to the current lagrangian subproblem
  vector<double> m_Ax_b;              // vector Ax-b, where x is the solution of the current lagrangian subproblem
  int m_noimprove;                  // number of consecutive iterations without improvement in the upper bound
  
  friend class sub_grad_method;
};

/** update multipliers functions **/

class sub_grad_method
{
public:
  sub_grad_method (Lag_Relax* _lag) : lag(_lag) {}
  
  virtual double theta (bool repair) = 0;
  virtual void set_lag (Lag_Relax* _lag) { lag = _lag; }
  virtual ~sub_grad_method () {}

protected:
  /* protected methods make private members of lag available to inherited classes */
  double m_curr_iter () { return lag->m_curr_iter; }
  double m_curr_lag () { return lag->m_curr_lag; }
  const vector<double>& m_Ax_b () { return lag->m_Ax_b; }
  bool m_have_solution () { return lag->m_have_solution; }
  double m_obj_val () { return lag->m_obj_val; }

private:
  Lag_Relax* lag;
};

class inverseK : public sub_grad_method
{
public:
  inverseK (Lag_Relax* _lag = 0) : sub_grad_method(_lag) {}

  virtual double theta (bool repair) { return 1. / (m_curr_iter()+1); }
};

class newton : public sub_grad_method
{
public:
  newton(double _lambda, double _ub, Lag_Relax* _lag = 0) : sub_grad_method(_lag), m_lambda(_lambda), m_ub(_ub) {}

  virtual double theta (bool repair);

private:
  double m_lambda;
  double m_ub;
};

#endif
