#include "CAalgos.h"
#include "WMutils.h"

#include "lp.h"
#include "milp.h"
#include <sstream>
#include <queue>
#include <map>
#include <algorithm>
#include <ext/functional>
#include <numeric>
#include <cassert>
#include <memory>
#include <ClpSimplex.hpp>
#include <CoinBuild.hpp>



/** Minimum Power-Channel Assignment and Routing Algorithm **/

namespace MP_CARA_NS
{

  /* compute the power consumption of a single node */
  float calcNodePowerConsumption(Topology* T, int n)
  {
    vector<TPower> power;
    
    T->calcAllNodesPowerConsumption(power);

    return power.at(n).power;
  }


  /** Priority function objects **/
  
  /* Abstract base class. T is the type of element in the queue */
  template <class T>
  class PrioFuncObj
  {
  public:
    /* returns the priority of an element */
    virtual double priority (const T& x) = 0;
    
    bool operator()(const T& p, const T& q) { return priority(p) < priority(q); }
  };


  /* Priority based on the sum of the utots */
  class SumUtotPrio : public PrioFuncObj<pair<int, set<TLinkItrs> > >
  {
  public:
    SumUtotPrio (Topology* _T, double _u0) : T(_T), u0(_u0) {}
    double priority (const pair<int, set<TLinkItrs> >& p);

  private:
    Topology* T;
    double u0;
  };

  /* The priority of a node is the sum of the costs of the links incident on it.
   * The cost of a link is its f/c ratio multiplied by the number of links
   * with Utot >= u0 having that link in their collision domain */
  double SumUtotPrio::priority (const pair<int, set<TLinkItrs> >& p)
  {
    double ret = 0.;
 
    for (int x=0; x<T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++) 
	if (y->UTot() >= u0)
	  for (set<TLinkItrs>::iterator pp=p.second.begin(); pp!=p.second.end(); pp++)
	    if (pp->IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)))
	      ret += pp->second->Flow() / pp->second->Rate();

    return ret;
  }

  
  /* Node suffer priority */
  class NodeSufferPrio : public PrioFuncObj<pair<int, set<TLinkItrs> > >
  {
  public:
    NodeSufferPrio (vector<TPower>* _consumption, vector<float>* _constraint) : consumption(_consumption), constraint(_constraint) {}
    double priority (const pair<int, set<TLinkItrs> >& x);

  private:
    float NodeSuffer (int n) { float ret = max(consumption->at(n).power/constraint->at(n), 1.f); for (unsigned int i=0; i<consumption->at(n).tau.size(); i++) ret *= max(consumption->at(n).tau[i], 1.f); return ret; }

    vector<TPower>* consumption;
    vector<float>* constraint;
  };

  /* compute the "suffering" of a node, which measures the extent to which tau_tx + tau_rx exceeds 1
   * (for each radio) plus the extent to which the power consumption exceeds the power constraint */
  double NodeSufferPrio::priority (const pair<int, set<TLinkItrs> >& x)
  {
    double x_prio = 0.;
    
    for (set<TLinkItrs>::iterator l=x.second.begin(); l!=x.second.end(); l++)
      if (l->first->Id() == x.first)
	x_prio += (l->second->Flow() * NodeSuffer(l->second->Id()));

    return x_prio;
  }


  /* Sum flow priority */
  class SumFlowPrio : public PrioFuncObj<pair<TRadio, set<TLinkItrs> > >
  {
  public:
    SumFlowPrio (double F = 0.) : maxF(F) {}
    double priority (const pair<TRadio, set<TLinkItrs> >& x);

  private:
    double maxF;
  };
  
  /* computes the sum of the flows of all the links in L */
  double SumFlowPrio::priority(const pair<TRadio, set< TLinkItrs > >& x)
  {
    int numRadiosOn = 0;

    vector<double> F(x.second.size());

    transform(x.second.begin(), x.second.end(), F.begin(), 
	      __gnu_cxx::compose1(mem_fun_ref(&TGraphNeighbor::Flow),
				  __gnu_cxx::compose1(mem_fun_ref (&TGraphNeighborIterator::operator*),
					              __gnu_cxx::select2nd<pair<TGraphNodeIterator,TGraphNeighborIterator> >() )  ) );
// 	      __gnu_cxx::compose1(mem_fun(&TGraphNeighbor::Flow), mem_fun_ref(&TLinkItrs::getNeighb_ptr) )  );

    // compute the number of active radios on the same node as the considered radio
    if (!x.second.empty())
    {
      MeshNode* info = NULL;

      if (x.second.begin()->first->Id() == x.first.first)
	info = x.second.begin()->first->Info();

      if (x.second.begin()->second->Id() == x.first.first)
	info = x.second.begin()->second->Info();

      if (info)
	numRadiosOn = info->numAvailRadios() - info->countRadiosWithChannel(0);
      else
	cerr << endl << "*** WARNING: SumFlowPrio: The provided link is not incident on the node owning the given radio ***" << endl << endl; 
    }

    return accumulate(F.begin(), F.end(), 0.) + maxF * (numRadiosOn-1);
  }


  /* Radio suffer priority */
  class RadioSufferPrio : public PrioFuncObj<pair<TRadio, set<TLinkItrs> > >
  {
  public:
    RadioSufferPrio (vector<TPower>* _consumption, vector<float>* _constraint) : consumption(_consumption), constraint(_constraint) {}
    double priority (const pair<TRadio, set<TLinkItrs> >& x);

  private:
    vector<TPower>* consumption;
    vector<float>* constraint;
  };

  /* compute the "suffering" of a radio, which measures the extent to which tau_tx + tau_rx exceeds 1
   * for that radio plus the extent to which the power consumption exceeds the power constraint */
  double RadioSufferPrio::priority (const pair<TRadio, set<TLinkItrs> >& x)
  {
    return 1.f / (max(consumption->at(x.first.first).power/constraint->at(x.first.first), 1.f) * max(consumption->at(x.first.first).tau.at(x.first.second), 1.f));
  }

}  /* MP_CARA_NS namespace */


using namespace MP_CARA_NS;


void MPCARA::Configure (const list<pair<string,string> >& params)
{
  // default values
  m_npc = NULL;
  m_seed = RndVar::getSeed();
  m_reset = false;
  m_u0 = 0.5;
  m_skipchannelcheck = false;
  m_old = false;
  m_nodepc = NULL;

  string initalgo;
  list<pair<string,string> > initparams;

  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    if (!m_npc)
      RP.read ("npc", m_npc);
    RP.read ("seed", m_seed);
    RP.read ("reset", m_reset);
    RP.read ("init_algo", initalgo);
    RP.read ("old", m_old);
    RP.read ("skip_channel_check", m_skipchannelcheck);
    RP.read ("u0", m_u0);

    if (RP.notFound())
    {
      // if the attribute name starts with "i_", it is an attribute of the init algo
      if (p->first.compare(0, 2, "i_") == 0)
      {
	initparams.push_back(pair<string,string>(p->first.substr(2),p->second));
      }
      else
	cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the MPCARA algorithm" << endl << endl;
    }
  }

  m_initalgo = CAalgo::Create (initalgo, initparams);
}


void MPCARA::SetName ()
{
  stringstream ss;
  string tmp;

  ss << "MP";

  if (m_reset)
    ss << "R";

  if (m_initalgo)
    ss << m_initalgo->name();

  if (m_old)
    ss << "1";

  if (m_npc)
    tmp = m_npc->getName();

  replace(tmp.begin(), tmp.end(), ' ', '_');

  ss << "_" << tmp << "_" << m_seed << "_" << m_u0;

  ss >> m_name;
}


double MPCARA::Run (Topology* T)
{
  // terminate if it has not been specified how to generate node power constraints
  if (!m_npc)
  {
    cout << endl << "ERROR: The random variable needed to determine node power constraints has not been defined. Exiting." << endl << endl;
    return -1;
  }

  // generate the power constraints
  MotherOfAll urng(m_seed);            // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used

  m_nodepc = new vector<float>;
  for (unsigned int i=0; i<T->GetNumNodes(); i++)
    m_nodepc->push_back(m_npc->rv->Next());

  if (m_reset)
    T->ResetChannels();

  if (m_initalgo)
    m_initalgo->Run(T);

  if (T->isThereLinkWithoutChannel())
  {
    cout << endl << ( m_skipchannelcheck ? "WARNING" : "ERROR" )
    << ": Looks like some links in the topology have not been assigned a channel."
    << endl << "This might cause problems to the algorithm." 
    << endl << "Maybe you did not configure an init algorithm in the config file." << endl
    << ( m_skipchannelcheck
         ? "skip_channel_check is set to true in the config file, thus proceeding anyway."
         : "If you want to proceed anyway, please set skip_channel_check to true in the config file." )
    << endl << endl;

    if (!m_skipchannelcheck)
      return -1;
  }

  return doRun (T);
}


/* Given a network topology, attempts to minimize the whole power consumption
   while keeping the maximum Utot below u0
   It should be always true that if a link u->v exists then a link v->u exists
   (not necessarily using the same channel)
   Returns 0 if a feasible solution has been found, -1 otherwise
*/
double MPCARA::doRun(Topology *T)
{
  int r;
  double maxUtot=numeric_limits<double>::max();
  
  cout << "Starting MP_CARA" << endl << endl;

  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands;
  
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: MP_CARA: Checking E2E flows before beginning FAILED" << endl << endl;
  cout << "Demands:" << endl;
  for(map<pair<int,int>,double>::iterator m=demands.begin(); m!=demands.end(); m++)
    cout << "d[" << m->first.first << "," << m->first.second << "] = " << m->second << endl;

  /* ensure that there is a power constraint value for each node */
  if (m_nodepc->size() != T->GetNumNodes())
  {
    cerr << endl << "WARNING: MP_CARA: The size of nodePowerConstr does not match the number of nodes" << endl << endl;
//     nodePowerConstr->resize(T->node.size(), T->pm.StandbyPowerConsumption);
    m_nodepc->resize(T->GetNumNodes(), T->Node(0)->Info()->GetStandbyPowerConsumption());
  }

  /* find the set of nodes that must be in sleep mode due to power constraints */
  /* and put them in sleep mode (remove their links and turn off their radios) */
  set<int> mustBeOff;
  TGraphNeighborIterator nbi;
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
  {
    float tmp = T->Node(u)->Info()->GetBasePowerConsumption();

    for (int ri = 0; ri < T->Node(u)->Info()->numAvailRadios(); ri++)
      tmp += T->Node(u)->Info()->Radio(ri)->Rho_Idle();

//     if (T->pm.BasePowerConsumption + T->pm.rho_idle * T->node[u].info->numAvailRadios() > nodePowerConstr->at(u))
    if (tmp > m_nodepc->at(u))
    {
      // if the node is a source or destination of a demand, there is no feasible solution
      for(map<pair<int,int>,double>::iterator m=demands.begin(); m!=demands.end(); m++)
	if (m->first.first == u || m->first.second == u)
	{
	  cout << endl << "Node " << u << ", which is the src/dst of a demand, has a too strict power constraint."
	       << endl << "A feasible solution cannot be found" << endl << endl;
	  return -1;
	}

      mustBeOff.insert(u);
      // set all the channels to 0
      for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	SetChannel (T->Node(u)->Info()->Radio(i), 0);
      // erase all the links outgoing node u
//       T->node[u].neighbor.clear();
      while (T->Node(u)->GetNumNeighbors())
	EraseNeighbor (T->Node(u), T->Node(u)->NeighborBegin());

      // erase all the links incoming node u
      for (unsigned int v=0; v<T->GetNumNodes(); v++)
	if (T->Node(v)->GetNumNeighbors())
	  while ( (nbi = T->Node(v)->FindNeighbor(u)) != T->Node(v)->NeighborEnd() )
	    EraseNeighbor (T->Node(v), nbi);
    }
  }

  /* FIRST STAGE: rearrange e2e flows to minimize the maximum Utot */
  MinMaxUtot mmu (T, &demands, &maxUtot, EmptyVector, false, m_nodepc);
  if (mmu.lp_solve())
  {
    // MinMaxUtot returns a non-zero value if the optimal solution has not been found
    // Hence, we are here if the program is not feasible
    // Run MinMaxUtot again without the power constraints
    MinMaxUtot mmu2 (T, &demands, &maxUtot);
    mmu2.lp_solve();
    // set maxUtot to infinity again
    maxUtot=numeric_limits<double>::max();
  }

  cout << "Back to MP_CARA after MinMaxUtot" << endl;

  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: MP_CARA: Checking E2E flows after MinMaxUtot FAILED" << endl << endl;


  /* FIRST STAGE: attempt to turn on radios to minimize Utot */


  /* compute a map that associates each node having radios that can be turned on with the set of its links */
  /* nodes belonging to mustBeOff are excluded */
  map<int, set<TLinkItrs> > nodeLinks;

  for (int u=0; u<T->GetNumNodes(); u++)
    if (mustBeOff.find(u) == mustBeOff.end())
      for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
      {
	if (T->Node(u)->Info()->countRadiosWithChannel(0))
	  nodeLinks[u].insert(T->LinkItrs(T->Node(u),v));
	if (v->Info()->countRadiosWithChannel(0))
	  nodeLinks[v->Id()].insert(T->LinkItrs(T->Node(u),v));
      }

  vector<TPower> consumption;

  while (!nodeLinks.empty() && maxUtot > m_u0) {

    map<int,set<TLinkItrs> >::iterator r0;

    if (maxUtot < numeric_limits<double>::max())
      r0 = max_element(nodeLinks.begin(), nodeLinks.end(), (m_old ? SumUtotPrio(T,m_u0) : SumUtotPrio(T,maxUtot)));
    else
    {
      T->calcAllNodesPowerConsumption(consumption);
      r0 = max_element(nodeLinks.begin(), nodeLinks.end(), NodeSufferPrio(&consumption, m_nodepc));
    }

    /* if the priority of the extracted node is null, it means that none of its links
     * is in the collision domain of a link with Utot > u0, hence turning on its radio is useless
     * Exit the loop because all the other nodes will have a zero priority 
     */
//     if (!calcPriority(T,u0,*r0))
//       break;

    /** BEGIN NEW FIRST STAGE **/

    map<TLinkItrs, pair<int,int> > lp_var;         // link entering or leaving the extracted node, pair<its position in the LP, count>

    if (!m_old)
      for (int u=0; u<T->GetNumNodes(); u++)
	for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
	  if (u == r0->first || v->Id() == r0->first)
	  {
// 	    lp_var[TLinkItrs(u,v)] = pair<int,int>(lp_var.size(),0);
	    lp_var[T->LinkItrs(T->Node(u),v)] = pair<int,int>(lp_var.size(),0);

	    for (int x=0; x<T->GetNumNodes(); x++)
	      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
		if (T->LinkItrs(T->Node(u),v).IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)))
		  lp_var[T->LinkItrs(T->Node(u),v)].second++;
	  }

    /** END NEW FIRST STAGE **/


    map<unsigned int, double> S;            // pairs channel, maximum Utot
    map<unsigned int, double>::iterator maxS;
    TGraphNeighborIterator v;
    map<unsigned int,map<int,list<TGraphNeighbor> > > addedNeighbors;       // set of neighbors added for each channel
    map<int,set<int> > neighbOn;        // for each channel stores the neighbor where an additional radio has been turned on
    int ptr_radio;

    /* find the channels of the extracted node and avoid to put them in S */
    set<unsigned int> r0Channels = T->Node(r0->first)->Info()->findChannels();

    /* for each available channel not in use by the extracted node */
    for (unsigned int c=1; c<=WMS.numAvailChannels(); c++)
      if (r0Channels.find(c) == r0Channels.end())
      {
	/* find all the links that can be added on such channel */
	for (int u=0; u<T->GetNumNodes(); u++)
	  /* for each node u in the network (other than the extracted node) that can be a neighbor */
	  if ((u != r0->first) && (mustBeOff.find(u) == mustBeOff.end()) && (r = T->MaxAllowLinkRate(r0->first,u,WMS.defaultTxPowerLevel(),WMS.min_rate(),WMS.max_rate())) >= WMS.min_rate() )
	  {
	    /* if u has a radio on channel c  or an available radio*/
	    if (T->Node(u)->Info()->countRadiosWithChannel(c) || T->Node(u)->Info()->countRadiosWithChannel(0))
	    {
	      /* a link on channel c can be established between the extracted node and u,*/
	      addedNeighbors[c][r0->first].push_front(TGraphNeighbor(T->Node(u)->Info(),WMS.availTXrates(r),WMS.defaultTxPowerLevel()));
	      /* a link on channel c can be established between u and the extracted node */
	      addedNeighbors[c][u].push_front(TGraphNeighbor(T->Node(r0->first)->Info(),WMS.availTXrates(r),WMS.defaultTxPowerLevel()));

	      if (!T->Node(u)->Info()->countRadiosWithChannel(c))
		neighbOn[c].insert(u);
	    }
	  }

	int numAddedLinks = 0;
	for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	  numAddedLinks += m->second.size();

	if (!numAddedLinks)
	  continue;          // no new links can be established on this channel

	cout << endl << "trying to turn on a radio of node " << r0->first << " on channel " << c << endl;

	if (m_old)
	{
	  double meanrate = 0.;
	  
	  S[c] = 0.;
	  
	  /* add the neighbors computed above for the selected channel to the topology */
	  int num = 0;
	  for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	    for (TGraphNeighborIterator n=m->second.begin(); n!=m->second.end(); n++, num++)
	    {
	      v = AddNeighbor (T->Node(m->first),&*n);
	      T->AssignChannel2Link(T->LinkItrs(T->Node(m->first),v), c);

	      S[c] = max(S[c], T->computeUtot(T->LinkItrs(T->Node(m->first),v)));
	      meanrate += v->Rate();
	    }
	  meanrate = meanrate / num;

	  /* estimate the amount of flow that can be routed on the new links */
	  if (S[c])
	    S[c] = (m_u0-S[c])*meanrate;
  // 	else
  // 	  S[c] = 100000.0;

	  cout << "  S[c] = " << S[c] << endl;

	// Optimal solution (very slow)
// 	double temp_utot=maxUtot;
//	MinMaxUtot(T, &demands, temp_utot, EmptyVector, false, nodePowerConstr);
//	S[c] = 1./temp_utot;
//	
//	cout << "  S[c] = " << temp_utot << endl;

	  /* remove the added links */
	  for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	    for (TGraphNeighborIterator n=m->second.begin(); n!=m->second.end(); n++)
	    {
	      v = T->Node(m->first)->FindNeighbor(n->Id(), c);

	      if (v != T->Node(m->first)->NeighborEnd())
		EraseNeighbor (T->Node(m->first), v);
	    }

	  /* remove channel c (if it was assigned to) from the channels of the extracted node */
// 	  ptr_radio = find_if(T->node[r0->first].info->radio, T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios,
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	  ptr_radio = T->Node(r0->first)->Info()->findRadioWithChannel(c);

// 	  if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
	  if (ptr_radio >= 0)
	    SetChannel (T->Node(r0->first)->Info()->Radio(ptr_radio), 0);

	  for (set<int>::iterator n=neighbOn[c].begin(); n!=neighbOn[c].end(); n++) {
	    /* remove channel c (if it was assigned to) from the channels of the selected neighbors */
// 	    ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
// 		    __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	    ptr_radio = T->Node(*n)->Info()->findRadioWithChannel(c);
	    
// 	    if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	    if (ptr_radio >= 0)
	      SetChannel (T->Node(*n)->Info()->Radio(ptr_radio), 0);
	  }
	}
	else
	{
	  /** BEGIN NEW FIRST STAGE **/

	  map<TLinkItrs, pair<int,int> > lp_var_c(lp_var);            // link entering or leaving the extracted node, pair<its position in the LP, count>

	  /* add the new links to lp_var_c */
	  for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	    for (TGraphNeighborIterator n=m->second.begin(); n!=m->second.end(); n++)
	    {
// 	      lp_var_c[TLinkItrs(T->node.begin()+m->first, n)] = pair<int,int>(lp_var_c.size(),0);
	      lp_var_c[T->LinkItrs(T->Node(m->first), n)] = pair<int,int>(lp_var_c.size(),0);

	      for (int x=0; x<T->GetNumNodes(); x++)
		for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
		{
		  if (y->RadioChannel() == c && T->LinkItrs(T->Node(m->first), n).PotInterferesWith(T->LinkItrs(T->Node(x),y)))
		    lp_var_c[T->LinkItrs(T->Node(m->first), n)].second++;
		  // update the Utot of the newly added link
		  if (y->RadioChannel() == c && T->LinkItrs(T->Node(x),y).PotInterferesWith(T->LinkItrs(T->Node(m->first), n)))
		    SetUtot (n, n->UTot() + y->Flow() / y->Rate());
		}
	    }

	  /* prepare the LP*/
	  ClpSimplex model;

	  model.resize(0, lp_var_c.size());

	  /* set bounds on variables */
	  for (map<TLinkItrs,pair<int,int> >::iterator m=lp_var_c.begin(); m!=lp_var_c.end(); m++)
	    if (m->first.second->Radio())
	      model.setColumnBounds(m->second.first, -m->first.second->Flow(), 0.);
	    else
	      model.setColumnBounds(m->second.first, 0., m->first.second->Rate());

	  int *colno = new int[lp_var_c.size()], *colno1 = new int[lp_var_c.size()];
	  double *coeff = new double[lp_var_c.size()], *coeff1 = new double[lp_var_c.size()];

  // 	for (int i=1; i<=lp_var_c.size(); i++)
  // 	  coeff[i] = 0.;               // reset the elements of coeff

	  for (map<TLinkItrs,pair<int,int> >::iterator l=lp_var_c.begin(); l!=lp_var_c.end(); l++)
	    coeff[l->second.first] = l->second.second / l->first.second->Rate();

	  for (map<TLinkItrs,pair<int,int> >::iterator l=lp_var_c.begin(); l!=lp_var_c.end(); l++)
	    coeff[l->second.first] += numAddedLinks / l->first.second->Rate();

	  for (unsigned int i=0; i<lp_var_c.size(); i++)
	    model.setObjectiveCoefficient(i, coeff[i]);

	  /* set the object direction to minimize */
	  model.setOptimizationDirection(1);

	  CoinBuild buildObject;

	  int k=0, k1=0;

	  /* constraints to balance the flow moved from old links to new links */
	  for (map<TLinkItrs,pair<int,int> >::iterator l=lp_var_c.begin(); l!=lp_var_c.end(); l++)
	    // outgoing links from the extracted node
	    if (l->first.first->Id() == r0->first)
	    {
	      colno[k] = l->second.first;
	      coeff[k++] = 1.;
	    }
	    else
	    {
	      colno1[k1] = l->second.first;
	      coeff1[k1++] = 1.;
	    }
	  buildObject.addRow(k, colno, coeff, 0., 0.);
	  buildObject.addRow(k1, colno1, coeff1, 0., 0.);

	  model.addRows(buildObject);

	  assert (!model.initialSolve());      // this program should always be feasible

	  S[c] = model.objectiveValue();

	  /*  Before quitting, clear memory*/
	  delete colno;
	  delete colno1;
	  delete coeff;
	  delete coeff1;

	  cout << "  M = " << S[c] << "  **" << endl;
	  
	  /** END NEW FIRST STAGE **/
	}

      }

    if (addedNeighbors.empty())
    {
      nodeLinks.erase(r0);
      continue;
    }

    /* select the channel maximizing S */
    if (m_old)
      maxS = max_element(S.begin(), S.end(), compareSecond<unsigned int,double>());
    else
      maxS = min_element(S.begin(), S.end(), compareSecond<unsigned int,double>());

    cout << "the selected channel is " <<  maxS->first << endl;

    /* add the neighbors computed earlier for the selected channel to the topology before calling MinMaxUtot */
    for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[maxS->first].begin(); m!=addedNeighbors[maxS->first].end(); m++)
      for (TGraphNeighborIterator n=m->second.begin(); n!=m->second.end(); n++)
      {
	v = AddNeighbor(T->Node(m->first), &*n);

	T->AssignChannel2Link(T->LinkItrs(T->Node(m->first),v),maxS->first);

	SetUtot (v, T->computeUtot(T->LinkItrs(T->Node(m->first),v)));

	// do not add the new link if its Utot is greater than maxUtot (even with no flow on it)
	// this ensures that maxUtot cannot increase after turning on this radio
	if (v->UTot() > maxUtot)
	  EraseNeighbor (T->Node(m->first), v);
      }

    /* the extracted radio is always left on, unless no link is established with nodes already
       having the selected channel */
    bool extrRadioOn = true;
    
    MinMaxUtot mmu (T, &demands, &maxUtot, EmptyVector, false, m_nodepc);
    if (mmu.lp_solve())
    {
#if 0
      /* remove the added links involving nodes in neighbOn */
      set<pair<int,TGraphNeighbor> >::iterator tmp;

      for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); )
      {
	if (neighbOn[maxS->first].find(n->first) != neighbOn[maxS->first].end() || neighbOn[maxS->first].find(n->second.info->id) != neighbOn[maxS->first].end())
	{
	  /* find the added link and remove it */
	  v = T->find_neighbor(n->first, n->second.info->id, maxS->first);

	  if (v != T->node[n->first].neighbor.end())
	    T->node[n->first].neighbor.erase(v);
	  else
	    cerr << "** WARNING: an added link has not been found!" << endl << endl;

	  /* remove the link from addedNeighbors */
	  tmp = n++; 
	  addedNeighbors[maxS->first].erase(tmp);
	}
	else
	  n++;
      }

      /* remove channel c from the neighbors in neighbOn */
      for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++)
      {
	ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
		__gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
	
	if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	  ptr_radio->channel = 0;
      }

      /* if there are no links left in addedNeighbors, remove channel c from the extracted node, too */
      /* if a feasible solution exists but leaving the radio on the extracted node on violates the power constraint on that node,
         then remove channel c from the extracted node and the remaining added links */
      if (addedNeighbors[maxS->first].empty() || (maxUtot < numeric_limits<REAL>::max() && calcNodePowerConsumption(T,r0->first) > nodePowerConstr->at(r0->first)))
      {
	extrRadioOn = false;

	for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++)
	{
	  /* find the added link and remove it */
	  v = T->find_neighbor(n->first, n->second.info->id, maxS->first);

	  if (v != T->node[n->first].neighbor.end())
	    T->node[n->first].neighbor.erase(v);
	  else
	    cerr << "** WARNING: an added link has not been found!" << endl << endl;
	}

	ptr_radio = find_if(T->node[r0->first].info->radio, T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios,
		__gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
	
	if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
	  ptr_radio->channel = 0;

	/* remove the links from addedNeighbors */
	addedNeighbors[maxS->first].clear();
      }
#endif
      /* if a feasible solution has not been found yet and the radio has been left on, then */
      if (!addedNeighbors[maxS->first].empty() && maxUtot == numeric_limits<double>::max())
      {
	// run MinMaxUtot again without the power constraints
	MinMaxUtot mmu2 (T, &demands, &maxUtot);
	mmu2.lp_solve();
	// set maxUtot to infinity again
	maxUtot=numeric_limits<double>::max();
      }
    }
    else
      cout << "** Turning on the radio the max Utot decreased to " << maxUtot << " **" << endl << endl;

    /* add the new links to the nodeLinks map to correctly compute the priority  (the ones remaining
       in addedNeighbors, if any, in case the LP was not feasible) */
    for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[maxS->first].begin(); m!=addedNeighbors[maxS->first].end(); m++)
      for (TGraphNeighborIterator n=m->second.begin(); n!=m->second.end(); n++)
	if ( (v=T->Node(m->first)->FindNeighbor(n->Id(),maxS->first)) != T->Node(m->first)->NeighborEnd() )
	{
	  if ( nodeLinks.find(m->first) != nodeLinks.end() )
	    nodeLinks[m->first].insert(T->LinkItrs(T->Node(m->first), v));

	  if ( nodeLinks.find(n->Id()) != nodeLinks.end() )
	    nodeLinks[n->Id()].insert(T->LinkItrs(T->Node(m->first), v));
	}

    /* if the radio has been turned on and there are other radios on the extracted node that can be turned on,
     * leave the node in the queue, so that the other radios can be turned on */
    if ( !extrRadioOn || !(T->Node(r0->first)->Info()->countRadiosWithChannel(0)) )
      nodeLinks.erase(r0);

    /* remove from the queue the neighbors that no more have available radios */
    if (extrRadioOn)
      for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++)
	if (!T->Node(*n)->Info()->countRadiosWithChannel(0) && (nodeLinks.find(*n)!=nodeLinks.end()))
	  nodeLinks.erase(*n);
  }


  /* Ok, we tried our best to keep Utot <= u0 for every link 
     Now, let's try to switch some radio off, while not 
     introducing any other violation to the constraint Utot <= u0 */

  cout << endl << "** STARTING SECOND STAGE **" << endl << endl;


  map<TRadio, set<TLinkItrs> > Lr;  // maps each radio to the set of links using it
  int ri;

  /* compute Lr for all the radios */
  for (unsigned int i=0; i<T->GetNumNodes(); i++)
    for (TGraphNeighborIterator v=T->Node(i)->NeighborBegin(); v!=T->Node(i)->NeighborEnd(); v++)
    {
      ri = v->Info()->findRadioWithChannel(v->RadioChannel());
      /* insert the link in the sets of the two radios at the endpoints of the link */
      Lr[TRadio(i,distance(T->Node(i)->Info()->Radio(0),RadioIfaceIterator(v->Radio())))].insert(T->LinkItrs(T->Node(i),v));
      if (ri >= 0)
	Lr[TRadio(v->Id(), ri)].insert(T->LinkItrs(T->Node(i),v));
    }

  if (m_old)
  {
    map<TRadio, set<TLinkItrs> >::iterator minLr;
    double UtotOpt = maxUtot;
    double Ptot = numeric_limits<double>::max();

    if (maxUtot < numeric_limits<double>::max())
      Ptot = T->TotalPowerConsumption();    

    while (!Lr.empty()) {
      
      /* DEBUG */
  //     for (map<TRadio, set<TLinkItrs> >::iterator r=Lr.begin(); r!=Lr.end(); r++) {
  //       cout << "Link che usano la radio del nodo " << r->first.first << " settata sul canale " << T->node[r->first.first].info->radio[r->first.second].channel << endl;
  //       for (set<TLinkItrs>::iterator l=r->second.begin(); l!=r->second.end(); l++)
  // 	cout << l->first->info->id << "->" << l->second->info->id << "[" << l->second->radio->channel << "]" << endl;
  //       cout << endl;
  //     }

      /* extract the pair (node,radio) with the minimum flow */
      double minP = -1.;   // flow routed on the links using the extracted radio
      
      if (maxUtot < numeric_limits<double>::max())
	minLr = min_element(Lr.begin(), Lr.end(), SumFlowPrio());
      else
      {
	T->calcAllNodesPowerConsumption(consumption);
	minLr = min_element(Lr.begin(), Lr.end(), RadioSufferPrio(&consumption, m_nodepc));
      }

      SumFlowPrio sfp;
      minP = sfp.priority(*minLr);
      
      cout << "** Extracting node " << minLr->first.first << " radio " << minLr->first.second << "(channel " << T->Node(minLr->first.first)->Info()->Radio(minLr->first.second)->Channel() << ") flow " << minP << " **" << endl;

      if (UtotOpt < numeric_limits<double>::max())
	maxUtot = UtotOpt;// * 1.005;

      /* if TurnOffRadio is feasible, the radio can be turned off and all the links in Lr can be removed from the topology */
  //    if (!minP || TurnOffRadio(T, &demands, &(minLr->second), u0)) {
      TurnOffRadioPower tor (T, &demands, minLr->first, m_u0, &maxUtot, &Ptot, m_nodepc);
      if (!tor.lp_solve())
      {
	cout << "** EUREKA ** turning off on node " << minLr->first.first << " radio " << minLr->first.second << "(channel " << T->Node(minLr->first.first)->Info()->Radio(minLr->first.second)->Channel() << ")  **" << endl;
	/* for all the links in minLr */
	for (set<TLinkItrs>::iterator l=minLr->second.begin(); l!=minLr->second.end(); l++)
	{
	  if (l->second->Flow())
	    cerr << "WARNING: TurnOffRadio is removing a link with non-null flow!" << endl;
	  /* first, remove the link from the Lr set of the other radio (not the extracted one), if it is still in Lr */
	  ri = l->second->Info()->findRadioWithChannel(l->second->RadioChannel());
	  if (minLr->first.first == l->second->Id())
	  {
	    if (Lr.find(TRadio(l->first->Id(), distance(l->first->Info()->Radio(0),RadioIfaceIterator(l->second->Radio())))) != Lr.end())
	      Lr[TRadio(l->first->Id(), distance(l->first->Info()->Radio(0),RadioIfaceIterator(l->second->Radio())))].erase(*l);
	  }
	  else if (ri >= 0)
	  {
	    if (Lr.find(TRadio(l->second->Id(), ri)) != Lr.end())
	      Lr[TRadio(l->second->Id(), ri)].erase(*l);
	  }
	  /* then, remove the link from the topology*/
	  EraseNeighbor (T->Node(l->first->Id()), l->second);
	}
	/* switch off the radio (set channel to 0) */
	SetChannel (T->Node(minLr->first.first)->Info()->Radio(minLr->first.second), 0);
      }

      /* remove the whole Lr set of the extracted radio */
      Lr.erase(minLr);

      // T->printTopology();  // DEBUG
    }

    /* erase links with null flow */
    //T->eraseLinksWithNullFlow();
  }
  else
  {

    double maxF = 0.;
    for (map<pair<int,int>,double>::iterator d = demands.begin(); d != demands.end(); d++)
      maxF = max (maxF, d->second);

    vector<ArcEpsY> epsy;

    // initialize the ArcEpsY structure
    for (int u = 0; u < T->GetNumNodes(); u++)
      epsy.push_back(ArcEpsY(maxF/2, 1, T->Node(u)->Info()->numAvailRadios()));

    bool stop;
    set<TRadio> AdjustEpsRadioList;
    set<int> AdjustEpsNodeList;
    float alpha = 0.5;

    do
    {
      // reset y_i
      for (vector<ArcEpsY>::iterator aey = epsy.begin(); aey != epsy.end(); aey++)
      {
	aey->yN = 1;
	fill (aey->yR.begin(), aey->yR.end(), 1);
      }

      do
      {
	// solve MinPower_eps
	MinPower_eps mpe(T, &demands, max(m_u0,maxUtot), m_nodepc, epsy);
	mpe.lp_solve();

	stop = true;
	AdjustEpsRadioList.clear();
	AdjustEpsNodeList.clear();

	// update y_i while checking if something changed
	for (unsigned int u=0; u<T->GetNumNodes(); u++)
	  if (T->Node(u)->Info()->countRadiosWithChannel(0) != T->Node(u)->Info()->numAvailRadios())
	  {
	    double nodeflow = 0.;

	    /* for each active radio */
	    for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	      if (T->Node(u)->Info()->Radio(i)->Channel())
	      {
		double radioflow = 0.;
		if (Lr.find(TRadio(u,i)) != Lr.end())
		  for (set<TLinkItrs>::iterator l = Lr[TRadio(u,i)].begin(); l != Lr[TRadio(u,i)].end(); l++)
		    radioflow += l->second->Flow();

		int yi = (radioflow < epsy.at(u).epsR.at(i) ? 0 : 1);

		if (yi != epsy.at(u).yR.at(i))
		{
		  epsy.at(u).yR.at(i) = yi;
		  stop = false;
		}

		if (0. < radioflow && radioflow < epsy.at(u).epsR.at(i))
		  AdjustEpsRadioList.insert(TRadio(u,i));

		nodeflow += radioflow;
	      }

	    int yi = (nodeflow < epsy.at(u).epsN ? 0 : 1);

	    if (yi != epsy.at(u).yN)
	    {
	      epsy.at(u).yN = yi;
	      stop = false;
	    }
	    
	    if (0. < nodeflow && nodeflow < epsy.at(u).epsN)
	      AdjustEpsNodeList.insert(u);
	  }
      } while (!stop);

      for (set<TRadio>::iterator r = AdjustEpsRadioList.begin(); r != AdjustEpsRadioList.end(); r++)
	epsy.at(r->first).epsR.at(r->second) *= alpha;

      for (set<int>::iterator n = AdjustEpsNodeList.begin(); n != AdjustEpsNodeList.end(); n++)
	epsy.at(*n).epsN *= alpha;

      stop = (AdjustEpsNodeList.empty() && AdjustEpsRadioList.empty());

    } while (!stop);


    /* remove links with null flow and turn off radios */
    for (map<TRadio, set<TLinkItrs> >::iterator lr = Lr.begin(); lr != Lr.end(); lr++)
    {
      double sumflow = 0.;

      /* for each link incoming to or outgoing from a radio */
      for (set<TLinkItrs>::iterator l=lr->second.begin(); l!=lr->second.end(); l++)
      {
	/* if it carries no flow */
	if (!l->second->Flow())
	{
	  /* first, remove the link from the Lr set of the other radio, if it is still in Lr */
// 	  ch = SearchChannel(l->second->info, l->second->radio->channel);
	  ri = l->second->Info()->findRadioWithChannel(l->second->RadioChannel());
	  if (lr->first.first == l->second->Id())
	  {
	    if (Lr.find(TRadio(l->first->Id(), distance(l->first->Info()->Radio(0),RadioIfaceIterator(l->second->Radio())))) != Lr.end())
	      Lr[TRadio(l->first->Id(), distance(l->first->Info()->Radio(0),RadioIfaceIterator(l->second->Radio())))].erase(*l);
	  }
	  else if (ri >= 0)
	  {
	    if (Lr.find(TRadio(l->second->Id(), ri)) != Lr.end())
	      Lr[TRadio(l->second->Id(), ri)].erase(*l);
	  }
	  /* then, remove the link from the topology*/
	  EraseNeighbor (T->Node(l->first->Id()), l->second);
	}
	else
	  sumflow += l->second->Flow();
      }

      /* if sumflow is null, the radio can be turned off */
      if (!sumflow)
	SetChannel (T->Node(lr->first.first)->Info()->Radio(lr->first.second), 0);
    }

  }

  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: MP_CARA: Checking E2E flows at the end FAILED" << endl << endl;

  if (maxUtot < numeric_limits<double>::max())
    return 0;

  return -1;
}



void MPCARA_OPT::Configure (const list<pair<string,string> >& params)
{
  // default values
  m_npc = NULL;
  m_seed = RndVar::getSeed();
  m_u0 = 0.5;
  m_to1 = 300;
  m_to2 = 300;
  m_nodepc = NULL;

  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    if (!m_npc)
      RP.read ("npc", m_npc);
    RP.read ("seed", m_seed);
    RP.read ("u0", m_u0);
    RP.read ("to1", m_to1);
    RP.read ("to2", m_to2);

    if (RP.notFound())
      cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the MPCARA_OPT algorithm" << endl << endl;
  }
}


void MPCARA_OPT::SetName ()
{
  stringstream ss;
  string tmp;

  ss << "MPO_";

  tmp = m_npc->getName();

  replace(tmp.begin(), tmp.end(), ' ', '_');

  ss << "_" << tmp << "_" << m_seed << "_" << m_u0 << "_" << m_to1 << "_" << m_to2;

  ss >> m_name;
}

/*
 * Solves the minimum power CA and routing problem
 * Returns 0 if a feasible solution has been found, -1 otherwise
*/
double MPCARA_OPT::Run (Topology* T)
{
  MotherOfAll urng(m_seed);            // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used

  m_nodepc = new vector<float>;
  for (unsigned int i=0; i<T->GetNumNodes(); i++)
    m_nodepc->push_back(m_npc->rv->Next());

  double maxUtot;
  set<int> S;
  double mu_0 = 1.0;      // initial value of the lagrangian multipliers
  int max_iter = 1;
  // long timeout = 150;

  // parameters for the newton's method
  double lambda = 2.;
  double ub = 90;     // this should be the objective function value of a feasible solution
  
  /* clone the topology because MinMaxUtot_MILP removes links with null flow */
  //Topology* U = new Topology(*T);

  MinMaxUtot_MILP mmu(T, m_nodepc, m_to1, false);

  mmu.set_breakValue(m_u0);
  
  if (mmu.lp_solve())
    return -1;

//   S.insert(6);
// //  S.insert(4);
// 
//   if (!mmu.lag_solve (S, mu_0, max_iter, new newton(2.,2.)))
//   {
//     cerr << endl << "Error: MinMaxUtot_MILP did not find a feasible solution" << endl << endl;
//     return -1;
//   }
// 
   maxUtot = mmu.getObjValue();
//maxUtot=0.5;
  cout << endl << "Minimum maxUtot found by MinMaxUtot_MILP: " << maxUtot << endl << endl;

  MinPower_MILP mpm(T, max(maxUtot,m_u0), m_nodepc, m_to2);

  // 1-2-7 2-3 / (4-)6 for bound
//  S.clear();
  //S.insert(3);
//   S.insert(4);
//   S.insert(6);
//   S.insert(7);
  
//  if (mpm.lp_solve() == INFEASIBLE)
//  mpm.lag_solve (S, mu_0, max_iter, new newton(lambda,ub), CoinCbc);

  if (mpm.lp_solve())
    return -1;
  
  m_lbound = mpm.getBound();
  for (int u = 0; u < T->GetNumNodes(); u++)
    m_lbound += T->Node(u)->Info()->GetStandbyPowerConsumption();

  //T->printStats("MPCARA_OPT",true);
  return 0;
}

