// Copyright (C) 2005, 2006 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Eclipse Public License.
//
// $Id: hs071_nlp.cpp 1864 2010-12-22 19:21:02Z andreasw $
//
// Authors:  Carl Laird, Andreas Waechter     IBM    2005-08-16

#include "split_ratios.h"
#include <cassert>
#include <iterator>
#include <fstream>
#include <math.h>


using namespace Ipopt;

/* computes sum_{l\in Dcoll{l0}} fi_l^d / c(l) */
inline Number sumcap(const Number* x, vector<TLink> links, int l0, int d) {
  Number result = 0.;
  for (set<int>::iterator l=links[l0].Dcoll.begin(); l!=links[l0].Dcoll.end(); l++)
    if (links[*l].demIndex[d] != -1)
      result += x[links[*l].demIndex[d]] / links[*l].capacity();
  return result;
}

// constructor
SplitRatios::SplitRatios(Topology* _T, vector<TDemand> _demands, vector< set< pair< int, int > > > _Sd) :
  T(_T), demands(_demands), Sd(_Sd), ratios(vector<map<TLinkItrs,double> >(_demands.size()))
{
  assert(demands.size() == Sd.size());
  
  list<TGraphNeighbor>::iterator n;
  vector<TLink>::iterator p;
  map<pair<int,int>,double>::iterator m;
  int pos;

  /* compute demProd */
  demProd = 1.0;
  for (int d=0; d<demands.size(); d++)
    demProd *= demands[d].Fmax();
  
  /* populate the Vd structure */
  for (int d=0; d<demands.size(); d++) {
    set<int> s;
    for (set<pair<int,int> >::iterator l=Sd[d].begin(); l!=Sd[d].end(); l++) {
      s.insert(l->first);
      s.insert(l->second);
    }
    Vd.push_back(s);
  }
  
  /* populate the links and vars structures */
  
  /* for each demand d */
  for (int d=0; d<demands.size(); d++)
    /* for each pair of nodes in Sd */
    for (set<pair<int,int> >::iterator l=Sd[d].begin(); l!=Sd[d].end(); l++) {
      /* find all the links between them */
      n = T->Node(l->first)->NeighborBegin();
      do {
	if ((n = T->Node(l->first)->FindNeighbor(l->second, n)) != T->Node(l->first)->NeighborEnd()) {
	  /* determine if the link is already present in links */
	  if (!findLink(links,pos,l->first,l->second,n->RadioChannel())) {
	    /* if not, add a new TLink to links */
	    pos = links.size();
	    links.push_back(TLink(l->first,l->second,n->RadioChannel(),n->Rate()));
	    /* initialize all the elements of demIndex to -1 */
	    links[pos].demIndex = vector<int>(demands.size(),-1);
	  }
	  /* the element in demIndex corresponding to the current demand is set to the position in vars */
	  links[pos].demIndex[d] = vars.size();
	  vars.push_back(pair<int,int>(pos,d));
	  /* fill the corresponding element in the vector of the initial solution */
// 	  if ((m = n->f_e2e.find(pair<int,int>(demands[d].src(),demands[d].dst()))) != n->f_e2e.end())
	  if (n->IsDemandPresent (pair<int,int>(demands[d].src(),demands[d].dst())))
	    fi_0.push_back(n->F_e2e (pair<int,int>(demands[d].src(),demands[d].dst())) / demands[d].Fmax());
	  else
	    fi_0.push_back(0.0);
	  n++;
	}
      } while (n != T->Node(l->first)->NeighborEnd());
    }
  /* find the collision domain for each link */
  for (vector<TLink>::iterator l=links.begin(); l!=links.end(); l++)
    for (vector<TLink>::iterator l0=links.begin(); l0!=links.end(); l0++) {
      TLinkItrs e(T->LinkItrs(T->Node(l->fromId()), T->Node(l->fromId())->FindNeighbor(l->toId(),l->channel())));
      TLinkItrs e0(T->LinkItrs(T->Node(l0->fromId()), T->Node(l0->fromId())->FindNeighbor(l0->toId(),l0->channel())));
      if (e0.IsInTheCollDomainOf(e))
	l->Dcoll.insert(distance(links.begin(),l0));
    }
}

//destructor
SplitRatios::~SplitRatios()
{

}


// returns the size of the problem
bool SplitRatios::get_nlp_info(Index& n, Index& m, Index& nnz_jac_g,
                             Index& nnz_h_lag, IndexStyleEnum& index_style)
{
  /* the number of variables */
  n = vars.size();

  /* number of constraints */
  m = links.size();         // constraints on the link capacity
  for (int d=0; d<demands.size(); d++)
    m += Vd[d].size();      // constraints on the flow conservation

  /* number of nonzeros in the jacobian */
  nnz_jac_g = 3*vars.size();

  /* the hessian is symmetric hence we only need the lower left corner */
  nnz_h_lag = vars.size()*(vars.size()+1)/2;

  // use the C style indexing (0-based)
  index_style = TNLP::C_STYLE;

  return true;
}

// returns the variable bounds
bool SplitRatios::get_bounds_info(Index n, Number* x_l, Number* x_u,
                                Index m, Number* g_l, Number* g_u)
{
  // here, the n and m we gave IPOPT in get_nlp_info are passed back to us.
  // If desired, we could assert to make sure they are what we think they are.
  // assert(n == 4);
  // assert(m == 2);

  // the variables have lower bounds of 0
  for (Index i=0; i<n; i++) {
    x_l[i] = 0.0;
  }

  // the variables have upper bounds of 1
  for (Index i=0; i<n; i++) {
    x_u[i] = 1.0;
  }

  Index i=0;
  /* bounds of the flow conservation constraints */
  for (int d=0; d<demands.size(); d++) 
    for (set<int>::iterator u=Vd[d].begin(); u!=Vd[d].end(); u++, i++) {
      if (*u == demands[d].src())
	g_l[i] = g_u[i] = 1;
      else if (*u == demands[d].dst())
	g_l[i] = g_u[i] = -1;
      else
	g_l[i] = g_u[i] = 0;
    }
    
  /* bounds of the link capacity constraints */  
  for (vector<TLink>::iterator l=links.begin(); l!=links.end(); l++, i++) {
    g_l[i] = 0.0;
    g_u[i] = l->capacity();
  }
  return true;
}

// returns the initial point for the problem
bool SplitRatios::get_starting_point(Index n, bool init_x, Number* x,
                                   bool init_z, Number* z_L, Number* z_U,
                                   Index m, bool init_lambda,
                                   Number* lambda)
{
  // Here, we assume we only have starting values for x, if you code
  // your own NLP, you can provide starting values for the dual variables
  // if you wish
  assert(init_x == true);
  assert(init_z == false);
  assert(init_lambda == false);

  for (Index i=0; i < n; i++)
    x[i] = fi_0[i];

  return true;
}

// returns the value of the objective function
bool SplitRatios::eval_f(Index n, const Number* x, bool new_x, Number& obj_value)
{
  // assert(n == 4);
  Number prod, sum;

  obj_value = 0.0;
  for (int l0=0; l0<links.size(); l0++) {
    prod = 1;
    for (int d=0; d<demands.size(); d++) {
      if ((sum = sumcap(x, links, l0, d)) != 0)
	prod *= (exp(sum*demands[d].Fmax())-1)/sum;
      else
	prod *= demands[d].Fmax();
    }
    obj_value += prod;
  }
  
  obj_value = (obj_value / (demProd*links.size()) - 1) / (exp(1)-1);

  return true;
}

// return the gradient of the objective function grad_{x} f(x)
bool SplitRatios::eval_grad_f(Index n, const Number* x, bool new_x, Number* grad_f)
{
  // assert(n == 4);
  int li,di;
  Number prod, sum;

  for (Index i=0; i<n; i++) {
    grad_f[i] = 0.0;
    li = vars[i].first;
    di = vars[i].second;

    for (int l0=0; l0<links.size(); l0++)
      /* if li is in Dcoll(l0) */
      if (links[l0].Dcoll.find(li) != links[l0].Dcoll.end()) {
	prod = 1;
	for (int d=0; d<demands.size(); d++) 
	  if (d!=di) {
	    if ((sum = sumcap(x, links, l0, d)) != 0)
	      prod *= (exp(sum*demands[d].Fmax())-1)/sum;
	    else
	      prod *= demands[d].Fmax();
	  }
	if ((sum = sumcap(x, links, l0, di)) != 0)
	  prod *= (exp(sum*demands[di].Fmax())*(sum*demands[di].Fmax()-1)+1) / links[li].capacity() / pow(sum,2);
	else
	  prod *= pow(demands[di].Fmax(),2) / 2 / links[li].capacity();

	grad_f[i] += prod;
      }
  }

  return true;
}

// return the value of the constraints: g(x)
bool SplitRatios::eval_g(Index n, const Number* x, bool new_x, Index m, Number* g)
{
  // assert(n == 4);
  // assert(m == 2);
  
  int prev_d=0, cumV=0;
  
  for (Index i=0; i<m; i++)
    g[i] = 0.0;
  
  /* flow conservation constraints */
  for (Index i=0; i<vars.size(); i++) {
    /* if we are considering the next demand */
    if (vars[i].second > prev_d) {
      /* add the number of vertices of the previous demand to cumV */
      cumV += Vd[prev_d].size();
      prev_d = vars[i].second;
    }
    g[cumV+distance(Vd[vars[i].second].begin(),Vd[vars[i].second].find(links[vars[i].first].fromId()))] += x[i];
    g[cumV+distance(Vd[vars[i].second].begin(),Vd[vars[i].second].find(links[vars[i].first].toId()))] -= x[i];
  }
    
  cumV += Vd.back().size();  /* now cumV equals the number of flow conservation constraints */  
    
  /* link capacity constraints */  
  for (int l=0; l<links.size(); l++)
    for (vector<int>::iterator i=links[l].demIndex.begin(); i!=links[l].demIndex.end(); i++)
      if (*i != -1)
	g[cumV+l] += x[*i]*demands[vars[*i].second].Fmax();

  return true;
}

// return the structure or values of the jacobian
bool SplitRatios::eval_jac_g(Index n, const Number* x, bool new_x,
                           Index m, Index nele_jac, Index* iRow, Index *jCol,
                           Number* values)
{
  int prev_d=0, cumV=0;
  Index i = 0;

  if (values == NULL) {
    // return the structure of the jacobian

    /* flow conservation constraints */
    for (Index v=0; v<vars.size(); v++) {
      /* if we are considering the next demand */
      if (vars[v].second > prev_d) {
	/* add the number of vertices of the previous demand to cumV */
	cumV += Vd[prev_d].size();
	prev_d = vars[v].second;
      }
      iRow[i] = cumV + distance(Vd[vars[v].second].begin(),Vd[vars[v].second].find(links[vars[v].first].fromId()));
      jCol[i++] = v;
      
      iRow[i] = cumV + distance(Vd[vars[v].second].begin(),Vd[vars[v].second].find(links[vars[v].first].toId()));
      jCol[i++] = v;
    }
      
    cumV += Vd.back().size();  /* now cumV equals the number of flow conservation constraints */  
    
    /* link capacity constraints */  
    for (int l=0; l<links.size(); l++)
      for (vector<int>::iterator d=links[l].demIndex.begin(); d!=links[l].demIndex.end(); d++)
	if (*d != -1) {
	  iRow[i] = cumV + l;
	  jCol[i++] = *d;
	}
  }
  else {
    // return the values of the jacobian of the constraints

    /* flow conservation constraints */
    for (Index v=0; v<vars.size(); v++) {
      /* if we are considering the next demand */
      if (vars[v].second > prev_d) {
	/* add the number of vertices of the previous demand to cumV */
	cumV += Vd[prev_d].size();
	prev_d = vars[v].second;
      }
      values[i++] = 1;
      values[i++] = -1;
    }
      
    cumV += Vd.back().size();  /* now cumV equals the number of flow conservation constraints */  

    /* link capacity constraints */  
    for (int l=0; l<links.size(); l++)
      for (vector<int>::iterator d=links[l].demIndex.begin(); d!=links[l].demIndex.end(); d++)
	if (*d != -1)
	  values[i++] = demands[vars[*d].second].Fmax();
  }

  return true;
}

//return the structure or values of the hessian
bool SplitRatios::eval_h(Index n, const Number* x, bool new_x,
                       Number obj_factor, Index m, const Number* lambda,
                       bool new_lambda, Index nele_hess, Index* iRow,
                       Index* jCol, Number* values)
{
  if (values == NULL) {
    // return the structure. This is a symmetric matrix, fill the lower left
    // triangle only.

    // the hessian for this problem is actually dense
    Index idx=0;
    for (Index row = 0; row < n; row++) {
      for (Index col = 0; col <= row; col++) {
        iRow[idx] = row;
        jCol[idx] = col;
        idx++;
      }
    }

    assert(idx == nele_hess);
  }
  else {
    // return the values. This is a symmetric matrix, fill the lower left
    // triangle only

    // fill the objective portion
    int li, lj, di, dj;
    Number prod, sum;
    
    Index idx=0;
    for (Index i = 0; i < n; i++) {
      for (Index j = 0; j <= i; j++, idx++) {
	values[idx] = 0.0;
	li = vars[i].first;
	di = vars[i].second;
	lj = vars[j].first;
	dj = vars[j].second;

	for (int l0=0; l0<links.size(); l0++)
	  /* if li and lj are in Dcoll(l0) */
	  if ((links[l0].Dcoll.find(li) != links[l0].Dcoll.end()) && (links[l0].Dcoll.find(lj) != links[l0].Dcoll.end()) ) {
	    prod = 1;
	    for (int d=0; d<demands.size(); d++) 
	      if (d!=di && d!=dj) {
		if ((sum = sumcap(x, links, l0, d)) != 0)
		  prod *= (exp(sum*demands[d].Fmax())-1)/sum;
		else
		  prod *= demands[d].Fmax();
	      }
	    if (di!=dj) {
	      /* add the term related to di */
	      if ((sum = sumcap(x, links, l0, di)) != 0)
		prod *= (exp(sum*demands[di].Fmax())*(sum*demands[di].Fmax()-1)+1) / links[li].capacity() / pow(sum,2);
	      else
		prod *= pow(demands[di].Fmax(),2) / 2 / links[li].capacity();
	      /* add the term related to dj */
	      if ((sum = sumcap(x, links, l0, dj)) != 0)
		prod *= (exp(sum*demands[dj].Fmax())*(sum*demands[dj].Fmax()-1)+1) / links[lj].capacity() / pow(sum,2);
	      else
		prod *= pow(demands[dj].Fmax(),2) / 2 / links[lj].capacity();
	    }
	    else {      /* di = dj */
	      if ((sum = sumcap(x, links, l0, di)) != 0)
		prod *= (exp(sum*demands[di].Fmax())*(pow(sum*demands[di].Fmax()-1,2)+1)-2) / links[li].capacity() / links[lj].capacity() / pow(sum,3);
	      else
		prod *= pow(demands[di].Fmax(),3) / 3 / links[li].capacity() / links[lj].capacity();
	    }
	    values[idx] += prod;
	  }
	values[idx] *= obj_factor;  
      }
    }

    // the portion for the constraints is null
  }

  return true;
}

void SplitRatios::finalize_solution(SolverReturn status,
                                  Index n, const Number* x, const Number* z_L, const Number* z_U,
                                  Index m, const Number* g, const Number* lambda,
                                  Number obj_value,
				  const IpoptData* ip_data,
				  IpoptCalculatedQuantities* ip_cq)
{
  // here is where we would store the solution to variables, or write to a file, etc
  // so we could use the solution.

  // a flow below min_ratio is considered null
  Number min_ratio = 0.0001;

  // to compute the split ratios, we must first compute the total flow (of each demand) outgoing each node
  vector<map<int,Number> > nodeFlow(demands.size());
  Number ratio;
  
  // insert the vertices associated with each demand in the corresponding map
  for (int i=0; i<demands.size(); i++)
    for (set<int>::iterator u=Vd[i].begin(); u!=Vd[i].end(); u++)
      nodeFlow[i].insert(pair<int,Number>(*u,0.0));
    
  // first pass: compute the total flows on each node
  for (Index i=0; i<n; i++) 
    if (x[i] >= min_ratio)
      nodeFlow[vars[i].second][links[vars[i].first].fromId()] += x[i];

  // second pass: compute the split ratios
  for (Index i=0; i<n; i++) {
    int l = vars[i].first;
    
    TLinkItrs li(T->LinkItrs(T->Node(links[l].fromId()), T->Node(links[l].fromId())->FindNeighbor(links[l].toId(),links[l].channel())));

    ratio = x[i] >= min_ratio ? x[i]/nodeFlow[vars[i].second][links[l].fromId()] : 0.0;

    ratios[vars[i].second].insert(pair<TLinkItrs,double>(li, ratio));
  }

  object_value = obj_value;

  // For this example, we write the solution to the console
//  printf("\n\nSolution of the primal variables, x\n");
//  for (Index i=0; i<n; i++) {
//    printf("x[%d] = %e\n", i, x[i]);
//  }
//
//   printf("\n\nSolution of the bound multipliers, z_L and z_U\n");
//   for (Index i=0; i<n; i++) {
//     printf("z_L[%d] = %e\n", i, z_L[i]);
//   }
//   for (Index i=0; i<n; i++) {
//     printf("z_U[%d] = %e\n", i, z_U[i]);
//   }

//  printf("\n\nObjective value\n");
//  printf("f(x*) = %e\n", obj_value);

/*  printf("\nFinal value of the constraints:\n");
  for (Index i=0; i<m ;i++) {
    printf("g(%d) = %e\n", i, g[i]);
  }*/
}



/* Searches a link (fromId,toId,channel) in the structure links */
bool findLink(vector<TLink> links, int& pos, int _fromId, int _toId, int _channel) {
  vector<TLink>::iterator p = links.begin();
  bool trovato = false;
  while (!trovato && p!=links.end()) {
    if ((p->fromId()==_fromId) && (p->toId()==_toId) && (p->channel()==_channel)) {
      trovato = true;
      pos = distance(links.begin(),p);
    }
    else
      p++;
  }
  return trovato;
}

void SplitRatios::saveRatios(const char* fname)
{
  fstream f(fname, ofstream::out | ofstream::app);
  
  f << endl << "# Number of demands" << endl << ratios.size() << endl;
  
  for (int i=0; i<ratios.size(); i++) {
    f << demands[i].src() << " " << demands[i].dst() << " " << demands[i].Fmax() << " " << ratios[i].size() << endl;
    
    for (map<TLinkItrs,double>::iterator m=ratios[i].begin(); m!=ratios[i].end(); m++)
      f << m->first.first->Id() << " " << m->first.second->Id() << " " << m->first.second->RadioChannel() << " " << m->second << endl;
    f << endl;
  }
  
  f.close();
}


ostream& operator<<(ostream& os, TDemand& d) {
    os << "(s: " << d.Src << ",d: " << d.Dst << ",f: " << d.fmax << ")" << endl;
    return os;
}

ostream& operator<<(ostream& os, TLink& l) {
    os << "(u: " << l.FromId << ",v: " << l.ToId << ",ch: " << l.Channel << ",C: " << l.Capacity << ")" << endl;
    os << "demIndex: ";
    copy(l.demIndex.begin(), l.demIndex.end(), ostream_iterator<int>(os, " "));
    os << endl << "Dcoll: ";
    copy(l.Dcoll.begin(), l.Dcoll.end(), ostream_iterator<int>(os, " "));
    os << endl << endl;
    return os;
}

ostream& operator<<(ostream& os, SplitRatios& P) {
  os << "Demands:" << endl;
  for (vector<TDemand>::iterator d=P.demands.begin(); d!=P.demands.end(); d++)
    os << *d;
  os << "Links:" << endl;
  for(vector<TLink>::iterator l=P.links.begin(); l!=P.links.end(); l++)
    os << distance(P.links.begin(),l) << ": " << *l;
  os << "Vars:" << endl;
  for (int i=0; i<P.vars.size(); i++)
    os << i << ": l=" << P.vars[i].first << ", d=" << P.vars[i].second << ", fi_0 = " << P.fi_0[i] << endl;
  os << "Vertices:" << endl;
  for (int d=0; d<P.demands.size(); d++) {
    os << d << ": ";
    copy(P.Vd[d].begin(), P.Vd[d].end(), ostream_iterator<int>(os, " "));
    os << endl;
  }
  return os;
}
