// -*- C++ -*-

/* myparser.cc
 *
 * Copyright (C) 2002 The libxml++ development team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "parser.h"
#include "WMutils.h"
#include "WMtypes.h"

#include <glibmm/convert.h> //For Glib::ConvertError
#include <iostream>
#include <cassert>

using namespace std;


/** Topology parser **/

TopologyParser::TopologyParser (Topology* _T)
  : xmlpp::SaxParser(),
    defaultNradios (0),
    defaultTxGainDb (RadioIface::default_db_gain),
    defaultRxGainDb (RadioIface::default_db_gain),
    defaultMinAngle (RadioIface::default_min_angle),
    defaultMaxAngle (RadioIface::default_max_angle),
    T (_T)
{
}

TopologyParser::~TopologyParser()
{
}

void TopologyParser::on_start_document()
{
  cout << "Reading topology..." << endl << endl;
}

void TopologyParser::on_end_document()
{
  cout << "Finished reading topology" << endl << endl;
}

void TopologyParser::on_start_element(const Glib::ustring& name,
                                   const AttributeList& attributes)
{
  if (name == "Topology")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("nradios", defaultNradios);
      RP.read ("txgainDb", defaultTxGainDb);
      RP.read ("rxgainDb", defaultRxGainDb);
      RP.read ("minangle", defaultMinAngle);
      RP.read ("maxangle", defaultMaxAngle);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Topology element" << endl << endl;
    }
  }
  else if (name == "Node")
  {
    // a new node is found: set attributes to default (Topology) values
    nodeTxGainDb = defaultTxGainDb;
    nodeRxGainDb = defaultRxGainDb;
    nodeMinAngle = defaultMinAngle;
    nodeMaxAngle = defaultMaxAngle;

    unsigned int nradios = defaultNradios;
    float xpos = -1.;
    float ypos = -1.;
    TMeshRouter role = router;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("nradios", nradios);
      RP.read ("txgainDb", nodeTxGainDb);
      RP.read ("rxgainDb", nodeRxGainDb);
      RP.read ("minangle", nodeMinAngle);
      RP.read ("maxangle", nodeMaxAngle);
      RP.read ("x", xpos);
      RP.read ("y", ypos);

      if (RP.notFound())
      {
	if (iter->name == "role")
	{
	  if (iter->value == "R")
	    role = router;
	  else if (iter->value == "G")
	    role = gateway;
	  else if (iter->value == "A")
	    role = aggregation;
	  else
	    cerr << endl << "Ignoring invalid node role: " << iter->value << endl << endl;
	}
	else
	  cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Node element" << endl << endl;
      }
    }
    
    if (xpos < 0. || ypos < 0.)
    {
      cerr << endl << "Ignoring node with invalid coordinates" << endl << endl;
      return;
    }

    MeshNode* u = new MeshNode (T->node.size(), xpos, ypos, role, nradios);

    RadioIfaceIterator ri = u->Radio(0);
    for (int rr = 0; rr < u->GetNumRadios(); rr++, ri++)
    {
      ri->tx_gain_db = nodeTxGainDb;
      ri->rx_gain_db = nodeRxGainDb;
      ri->minangle = nodeMinAngle;
      ri->maxangle = nodeMaxAngle;
    }

    // insert the new node into the topology
    T->node.push_back(TGraphNode(u));

    if (role == aggregation) 
      T->aggregation_list.push_back(T->node.back().Id());
    else if (role == gateway)
      T->gateway_list.push_back(T->node.back().Id());
  }
  else if (name == "Radio")
  {
    // additional radio for the current node
    float TxGainDb = nodeTxGainDb;
    float RxGainDb = nodeRxGainDb;
    unsigned int minangle = nodeMinAngle;
    unsigned int maxangle = nodeMaxAngle;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("txgainDb", TxGainDb);
      RP.read ("rxgainDb", RxGainDb);
      RP.read ("minangle", minangle);
      RP.read ("maxangle", maxangle);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Radio element" << endl << endl;
    }
    
    // at least one node must have already been added to the topology
    assert ( T->node.size() );

    // add the radio to the current node
    T->node.back().Info()->radio.push_back(RadioIface(TxGainDb, RxGainDb, minangle, maxangle));
  }
  else
  {
    cerr << endl << "Ignoring invalid element (" << name << ")" << endl << endl;
  }
}

void TopologyParser::on_end_element(const Glib::ustring& /* name */)
{
//   std::cout << "on_end_element()" << std::endl;
}

void TopologyParser::on_characters(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_characters(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "TopologyParser::on_characters(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void TopologyParser::on_comment(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_comment(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "TopologyParser::on_comment(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void TopologyParser::on_warning(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_warning(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "TopologyParser::on_warning(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void TopologyParser::on_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "TopologyParser::on_error(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void TopologyParser::on_fatal_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_fatal_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "TopologyParser::on_characters(): Exception caught while converting value for std::cout: " << ex.what() << std::endl;
  }
}


