/* This file contains the implementation of the methods of the Topology class */

#include <fstream>
#include <iostream>
#include <cmath>
#include <cstring>
#include <set>
#include <queue>
#include <cassert>
#include <numeric>
#include <algorithm>
#include <ext/functional>
#include "WMtypes.h"
#include "WMutils.h"
#include "parser.h"
#include "subgraph.h"
#include "dag.h"
#include "DijkstraShortestPathAlg.h"
#include "YenTopKShortestPathsAlg.h"


// global variable that stores various parameters of a simulation
WiMeshSimulation WMS;


// definitions for static constexpr strings
constexpr char foRadioTxPowerLinear::name[];
constexpr char SINRbasedInterferenceModel::name[];
constexpr char kHopNeighbInterferenceModel::name[];
constexpr char LogDistancePropagationLossModel::name[];
constexpr char ThreeLogDistancePropagationLossModel::name[];


bool operator<(const TGraphNeighborIterator& x, const TGraphNeighborIterator& y) {
  return ( (x->Id() < y->Id()) || ((x->Id() == y->Id()) && (&(*x) < &(*y))) );
}

bool operator<(const TGraphNeighbor& x, const TGraphNeighbor& y) {
  return x.Id() < y.Id();
}

/*
 * This constructor makes use of the extended initializer list feature of C++11 (requires g++ >= 4.4)
 * 
 * TODO UPDATE The SINR thresholds are used by the SINR based interference model. They
 * are determined such that nodes can communicate
 *   at 54 Mbps when within 30m
 *   at 48 Mbps when within 32m
 *   at 36 Mbps when within 37m
 *   at 24 Mbps when within 45m
 *   at 18 Mbps when within 60m
 *   at 12 Mbps when within 69m
 *   at  9 Mbps when within 77m
 *   at  6 Mbps when within 90m
 * if transmitting at 15 dbm
 * 
 */
WiMeshSimulation::WiMeshSimulation ()
: m_numAvailChannels (12),

  m_availTXrates {6, 9, 12, 18, 24, 36, 48, 54},

  m_min_rate (0),

  m_max_rate (m_availTXrates.size()-1),

  m_SINRthresholds { {6, 54.471f}, {9, 54.919f}, {12, 55.372f}, {18, 56.292f}, 
                     {24, 57.711}, {36, 62.801f}, {48, 138.329f}, {54, 185.353f} },

  m_noise (4e-13),

  m_SINRthreshfactor (1.15),

  m_minTxPowerLevel (0.f),

  m_maxTxPowerLevel (20.f),

  m_numTxPowerLevels (21),             // 0, 1, 2, 3, ... 20 dBm

  m_defaultTxPowerLevel (15),

  m_interferenceModel (new SINRbasedInterferenceModel),

  m_propagationLossModel (new LogDistancePropagationLossModel),

  m_simulId ("xxx")
{}


string PropagationLossModel::Description () const
{
  stringstream ss;

  ss << Name() << "[";

  for (vector<pair<string,double> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ss << p->first << "=" << p->second;

    if (*p != params.back())
      ss << ", ";
  }

  ss << "]";

  return ss.str();
}

double LogDistancePropagationLossModel::Loss (double distance) const
{
  double m_exponent = params.at(0).second,
         m_referenceDistance = params.at(1).second,
         m_referenceLoss = params.at(2).second;

  if (distance <= m_referenceDistance)
  {
    return 0.0;
  }
  /**
   * The formula is:
   * rx = 10 * log (Pr0(tx)) - n * 10 * log (d/d0)
   *
   * Pr0: rx power at reference distance d0 (W)
   * d0: reference distance: 1.0 (m)
   * d: distance (m)
   * tx: tx power (dB)
   * rx: dB
   *
   * Which, in our case is:
   *
   * rx = rx0(tx) - 10 * n * log (d/d0)
   */
  double pathLossDb = 10 * m_exponent * std::log10 (distance / m_referenceDistance);
  double rxc = -m_referenceLoss - pathLossDb;
  return rxc;
}

double ThreeLogDistancePropagationLossModel::Loss (double distance) const
{
  double pathLossDb,
         m_distance0 = params.at(0).second,
         m_distance1 = params.at(1).second,
         m_distance2 = params.at(2).second,
         m_exponent0 = params.at(3).second,
         m_exponent1 = params.at(4).second,
         m_exponent2 = params.at(5).second,
         m_referenceLoss = params.at(6).second;

  if (distance < m_distance0)
    {
      pathLossDb = 0;
    }
  else if (distance < m_distance1)
    {
      pathLossDb = m_referenceLoss
        + 10 * m_exponent0 * std::log10 (distance / m_distance0);
    }
  else if (distance < m_distance2)
    {
      pathLossDb = m_referenceLoss
        + 10 * m_exponent0 * std::log10 (m_distance1 / m_distance0)
        + 10 * m_exponent1 * std::log10 (distance / m_distance1);
    }
  else
    {
      pathLossDb = m_referenceLoss
        + 10 * m_exponent0 * std::log10 (m_distance1 / m_distance0)
        + 10 * m_exponent1 * std::log10 (m_distance2 / m_distance1)
        + 10 * m_exponent2 * std::log10 (distance / m_distance2);
    }
  return -pathLossDb;
}

/* returns true if x->y potentially interferes with u->v according to the SINR based model */
bool SINRbasedInterferenceModel::PotInterfLinks (const TLinkItrs &xy, const TLinkItrs &uv) const
{
  Topology* T = xy.topology();

/*	if (xy == uv || FindInverseLink(xy) == uv)
	  return true;*/

  // x == u or y == v (links have the same transmitter or receiver)
  if (xy.first->Id() == uv.first->Id() || xy.second->Id() == uv.second->Id())
    return true;

  // x == v or y == u (the transmitter of a link is the receiver of the other)
  if (xy.first->Id() == uv.second->Id() || xy.second->Id() == uv.first->Id())
    return true;

//   float power1 = DbmToW(WMS.TxPowerLevel(uv.second->TxPower())),
// 	power2 = DbmToW(WMS.TxPowerLevel(xy.second->TxPower()));

  // x is close to u, so when x transmits u senses the channel busy and does not transmit
  // the received power at 90m when transmitting at 15dbm is taken as reference
// 	if (gain(T->dist(xy.first->Id(), uv.first->Id())) * power2 > gain(90) * 0.03162)
  float refpower = DbmToW (WMS.TxPowerLevel(WMS.defaultTxPowerLevel())
                           + 2 * RadioIface::default_db_gain
                           + WMS.PropagationModel()->Loss(90));

  if (T->ReceivedPower (xy, uv.first->Id()) > refpower)
    return true;

  // the SINR is affected by the distance between x and v (relative to the distance between u and v)
//   return gain(T->dist(uv.first->Id(),uv.second->Id())) * power1 / ( gain(T->dist(xy.first->Id(),uv.second->Id())) * power2 + WMS.noise() ) < WMS.SINRthresholds(uv.second->Rate());
  return (T->ReceivedPower(uv) / (T->ReceivedPower(xy, uv.second->Id()) + WMS.noise())) < WMS.SINRthresholds(uv.second->Rate());
}

string SINRbasedInterferenceModel::Description () const
{
  stringstream ss (name);

  return ss.str();
}

int kHopNeighbInterferenceModel::dist (Topology* T, int u, int v) const
{
  if (u == v)
    return 0;

  if (T->Node(u)->FindNeighbor(v) != T->Node(u)->NeighborEnd() || T->Node(v)->FindNeighbor(u) != T->Node(v)->NeighborEnd())
    return 1;

  // if k==2, no need to investigate further
  if (k == 2)
    return 2;

  // check if v is a 2-hop neighbor of u
  for (TGraphNeighborIterator x = T->Node(u)->NeighborBegin(); x != T->Node(u)->NeighborEnd(); x++)
    if (T->Node(x->Id())->FindNeighbor(v) != T->Node(x->Id())->NeighborEnd())
      return 2;

  // check if u is a 2-hop neighbor of v
  for (TGraphNeighborIterator x = T->Node(v)->NeighborBegin(); x != T->Node(v)->NeighborEnd(); x++)
    if (T->Node(x->Id())->FindNeighbor(u) != T->Node(x->Id())->NeighborEnd())
      return 2;

  return 3;
}

string kHopNeighbInterferenceModel::Description () const
{
  stringstream ss (name);

  ss << " [k=" << k << "]";

  return ss.str();
}

/* returns true if x->y potentially interferes with u->v according to the k-hop neighbor model */
bool kHopNeighbInterferenceModel::PotInterfLinks (const TLinkItrs &xy, const TLinkItrs &uv) const
{
  Topology* T = xy.topology();

  return ( dist (T, xy.first->Id(), uv.first->Id()) < k ||
           dist (T, xy.first->Id(), uv.second->Id()) < k ||
           dist (T, xy.second->Id(), uv.first->Id()) < k ||
           dist (T, xy.second->Id(), uv.second->Id()) < k );
}


float Topology::ReceivedPower (int u, int v, unsigned int power)
{
  RadioIfaceIterator it;
  float Gtx, Grx;

  // find the minimum tx gain of u's radios
  it = std::min_element(node[u].Info()->Radio(0), node[u].Info()->Radio(node[u].Info()->GetNumRadios()-1), 
    [](const RadioIface& x, const RadioIface& y) {  return x.TxGainDb() < y.TxGainDb(); });

  Gtx = it->TxGainDb();

  // find the minimum rx gain of v's radios
  it = std::min_element(node[v].Info()->Radio(0), node[v].Info()->Radio(node[v].Info()->GetNumRadios()-1), 
    [](const RadioIface& x, const RadioIface& y) {  return x.RxGainDb() < y.RxGainDb(); });

  Grx = it->RxGainDb();

  return DbmToW (WMS.TxPowerLevel(power) + Gtx + Grx + WMS.PropagationModel()->Loss(dist(u,v)));
}

float Topology::ReceivedPower (TLinkItrs l, int x)
{
  float Gtx, Grx;

  if (l.second->Radio())
  {
    Gtx = l.second->Radio()->TxGainDb();
  }
  else
  {
    // find the minimum tx gain of the sender's radios
    Gtx = std::min_element(l.first->Info()->Radio(0), l.first->Info()->Radio(l.first->Info()->GetNumRadios()-1), 
    [](const RadioIface& x, const RadioIface& y) {  return x.TxGainDb() < y.TxGainDb(); }) -> TxGainDb();
  }

  if (l.second->Radio() && x < 0)
  {
    int ri = l.second->Info()->findRadioWithChannel (l.second->RadioChannel());
    Grx = l.second->Info()->Radio(ri)->RxGainDb();
  }
  else
  {
    int dest = ( x < 0 ? l.second->Id() : x );
    // find the minimum rx gain of dest's radios
    Grx = std::min_element(node[dest].Info()->Radio(0), node[dest].Info()->Radio(node[dest].Info()->GetNumRadios()-1), 
    [](const RadioIface& x, const RadioIface& y) {  return x.RxGainDb() < y.RxGainDb(); }) -> RxGainDb();
  }

  double distance = ( x < 0 ? dist (l.first->Id(), l.second->Id()) : dist (l.first->Id(), x) );

  return DbmToW (WMS.TxPowerLevel(l.second->TxPower()) + Gtx + Grx + WMS.PropagationModel()->Loss(distance));
}

bool TLinkItrs::PotInterferesWith (const TLinkItrs &l) const
{
  return WMS.InterfModel()->PotInterfLinks (*this, l);
}



bool TLinkItrs::IsInTheCollDomainOf (const TLinkItrs &l) const
{
  return (this->second->RadioChannel() == l.second->RadioChannel()) && PotInterferesWith(l);
}


int MeshNode::countRadiosWithChannel(int c)
{ 
  return count_if(radio.begin(), radio.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::Channel) ) ); 
}

int MeshNode::findRadioWithChannel(int c)
{
  RadioIfaceIterator ri = find_if(radio.begin(), radio.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::Channel) ) );

  if (ri == radio.end())
    return -1;

  return distance (radio.begin(), ri);
}

set<unsigned int> MeshNode::findChannels ()
{
  set<unsigned int> chset;

  chset.clear();

  for (RadioIfaceIterator ri=radio.begin(); ri!=radio.end(); ri++)
    chset.insert(ri->Channel());

  chset.erase(0);

  return chset;
}


set<unsigned int> Topology::haveCommonChannels (int u, int v)
{
  set<unsigned int> S, uChannels, vChannels;

  uChannels = node[u].Info()->findChannels();
  vChannels = node[v].Info()->findChannels();
  
  S.clear();
  set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

  return S;
}

#if 0
/* getNodeInfo: get node info from file 'filename' 
   on error, returns a nonzero value
*/
int Topology::getNodeInfo(FILE *f)
{
	int id=0, nradios;
	TMeshRouter _role;
	MeshNode *u;
	TGraphNode gnode;
	float Xpos(), Ypos();
	char c;
	char msg[80]="Unexpected value or EOF while reading the ";
	
	// read the number of available channels
	//READ_PARAM(f,"%ui",&numAvailChannels,strcat(msg,"number of available channels"));

	//read nodes
	skipComment(f);
	while (fscanf(f,"%c",&c)!=EOF) 
	{
		switch (c)
		{
			case 'A':
				_role = aggregation;
				break;
			case 'G':
				_role = gateway;
				break;
			default:
				_role = router;
				break;
		}

		READ_PARAM(f,"%f",&Xpos(),strcat(msg,"x coordinate"));
		READ_PARAM(f,"%f",&Ypos(),strcat(msg,"y coordinate"));
		READ_PARAM(f,"%i",&nradios,strcat(msg,"number of radios"));
		u = new MeshNode(id++, Xpos(), Ypos(), _role, nradios);

		gnode.info = u;
		node.push_back(gnode);

		if (_role == aggregation) 
			aggregation_list.push_back(u->id); 
		else if (_role == gateway)
			gateway_list.push_back(u->id);

		skipComment(f);
	}
	return 0;
}


int Topology::getNodeInfo(const char *filename)
{
	FILE *f;

	if ((f = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "ERR: Cannot open file %s to read topology\n",filename);
		return -1;
	}
	int res = getNodeInfo(f);
	fclose(f);
	return res;
}
#endif

float Topology::dist(int x, int y)
{ 
  return sqrtf(powf(fabsf(node[x].Info()->Xpos() - node[y].Info()->Xpos()),2)+powf(fabsf(node[x].Info()->Ypos() - node[y].Info()->Ypos()),2));
}

void Topology::addNeighbor(int u, int v, float rate, unsigned int power)
{
  TGraphNeighbor *gneighbor;
  list<TGraphNeighbor>::const_iterator pos;

  gneighbor = new TGraphNeighbor(node[v].Info(), rate, power);

  /* insert gneighbor in the neighbor list  */
  node[u].neighbor.push_back(*gneighbor);
}

/* Builds an initial graph using the given (fixed) transmission power 
 * A link exists between two nodes if there is a rate between min_rate and max_rate
 * that allows to establish the link (in such a case, the highest rate is selected) */
void Topology::buildGi_FixedPowerMaxAllowRate(unsigned int power, int minrate, int maxrate)
{
  int i;

  for (int u=0; u<node.size(); u++)
    for (int v=0; v<node.size(); v++)
      if (v!=u)
      {
	i = MaxAllowLinkRate (u, v, power, minrate, maxrate);

	if (i>=minrate)
	  addNeighbor(u,v,WMS.availTXrates(i),power);
      }

  WMS.m_min_rate = minrate;
  WMS.m_max_rate = maxrate;
  WMS.m_defaultTxPowerLevel = power;
}


int Topology::MaxAllowLinkRate (int u, int v, unsigned int power, int minrate, int maxrate)
{
  int i = maxrate;
  while ( (i>=minrate) && ( ReceivedPower(u,v,power)/WMS.noise() < WMS.SINRthreshfactor() * WMS.SINRthresholds(WMS.availTXrates(i))) )
    i--;

  return i;
}


/* Checks whether two nodes u and v can be neighbors using some values for TX power and rate
   Starts from the highest TX rate and the lowest TX power, then increases the TX power
   If the SNR is still below the threshold, lowers the TX rate
   Thus, determines the highest possible capacity
   if c>=0 then it is possible to establish a link between u and v
*/
bool Topology::canBeNeighbors(int u, int v, int & c, int & p, int maxp)
{
  c = WMS.numTXrates()-1;           // start with the maximum TX rate
  do
  {
    p = 0;
    while ( (p<maxp) && (ReceivedPower(u,v,p)/WMS.noise() < WMS.SINRthreshfactor() * WMS.SINRthresholds(WMS.availTXrates(c))) )
      p++;

    if (p==maxp)
      c--;

  } while (p==maxp && c>=0);

  return (c>=0);
}

// Searches for new links that can be added to the network graph
void Topology::findToBePossiblyAddedLinks(int maxp)
{
	int c,p;
	for (int u=0; u<node.size(); u++)
		for (int v=0; v<node.size(); v++)
			// if v!=u and u and v are not already neighbors
			if ( (v!=u) &&  find_if(node[u].neighbor.begin(), node[u].neighbor.end(), 
			  __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), mem_fun_ref(&TGraphNeighbor::Id) )) == node[u].neighbor.end() )
			{
				canBeNeighbors(u,v,c,p,maxp);
				if (c>=0) {
					addNeighbor(u,v,WMS.availTXrates(c),p);
					cout << "findToBePossiblyAddedLinks(): Aggiunto link " << u << "->" << v << " c " << c 
					<< " rate " << WMS.availTXrates(c) << " power  " << WMS.TxPowerLevel(p)
					<< " SNR " << ReceivedPower (u,v,p) / WMS.noise()
					<< " threshold " << WMS.SINRthreshfactor() * WMS.SINRthresholds(WMS.availTXrates(c)) << endl;
				}
			}
}

/* constructor */
Topology::Topology (const char* filename)
 : throughput (0)
{
  TopologyParser parser (this);
  parser.set_substitute_entities(true);
  parser.parse_file(filename);

  topopath = string (filename);
  string::size_type pos2 = topopath.rfind(".");
  string::size_type pos1 = topopath.rfind("/");
  toponame = topopath.substr(pos1+1,pos2-pos1-1);
}


/* copy constructor*/
Topology::Topology(const Topology& Tin)
{
	list<TGraphNeighbor>::const_iterator pos;

	/* simple copy of the fields (node, aggregation_list and gateway_list) */
	*this = Tin;

	/* in addition, duplicate undirectedLink structures */
	for (int u=0; u<node.size(); u++)
		for (TGraphNeighborIterator v=node[u].neighbor.begin(); v!=node[u].neighbor.end(); v++) {
			v->info = node[v->Id()].info;
// 			v->radio = u->info->radio + (v->radio - Tin.node[u->Id()].info->radio);
			v->radio = &*node[u].info->Radio(distance(Tin.node[u].info->Radio(0),RadioIfaceIterator(v->Radio())));
		}
}


/* set the current TX range of each interface so as to just reach the farthest neighbor reachable on the channel that interface is set to 
void Topology::AdjustPower()
{
	set<unsigned int> uChannels, vChannels, CommonChannels;

	for (vector<TGraphNode>::const_iterator u=node.begin(); u!=node.end(); u++)
	{
		float maxdist[u->info->numAvailRadios()];
		for (unsigned int i=0; i<u->info->numAvailRadios(); maxdist[i++]=0.);

// OLD CODE
// 		for (vector<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
// 			if (v->uLink->distance > maxdist[v->radio - u->info->radio])
// 				maxdist[v->radio - u->info->radio] = v->uLink->distance;

		uChannels.clear();
		for (unsigned int r=0; r<u->info->numAvailRadios(); uChannels.insert(u->info->radio[r++].channel));

		for (vector<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
			vChannels.clear();
			CommonChannels.clear();
			for (unsigned int r=0; r<v->info->numAvailRadios(); vChannels.insert(v->info->radio[r++].channel));
			set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(CommonChannels, CommonChannels.begin()));
			CommonChannels.erase(0);
			assert(CommonChannels.size());
			unsigned int i;
			for (i=0; u->info->radio[i].channel != *CommonChannels.begin(); i++);
			assert(i<u->info->numAvailRadios());
			if (v->uLink->distance > maxdist[i])
				maxdist[i] = v->uLink->distance;
		}

//		for (unsigned int i=0; i<u->info->numAvailRadios(); i++)
//			u->info->radio[i].curTXrange = maxdist[i];
	}
}
*/

void Topology::ResetChannels()
{
  for (int u=0; u<node.size(); u++)
  {
    // set all channels to 0
    for (unsigned int i=0; i<node[u].info->numAvailRadios(); i++)
      node[u].info->Radio(i)->channel = 0;

    // remove parallel links
    TGraphNeighborIterator i;

    for (TGraphNeighborIterator v=node[u].neighbor.begin(); v!=node[u].neighbor.end(); v++)
    {
      // if no link to neighbor v has been already found
      if ((i = node[u].FindNeighbor(v->Id())) == v)
      {
	v->radio = NULL;
	v->Utot = 0.;
      }
      else
      {
	i->flow += v->flow;
	i->rate = (i->rate > v->rate ? i->rate : v->rate);
	i->txpower = (i->txpower > v->txpower ? i->txpower : v->txpower);

	for (map<pair<int,int>,double>::iterator f = v->f_e2e.begin(); f != v->f_e2e.end(); f++)
	  if (i->f_e2e.find(f->first) != i->f_e2e.end())
	    i->f_e2e[f->first] += f->second;
	  else
	    i->f_e2e[f->first] = f->second;
	  
	// remove this neighbor
	i = v--;
	node[u].neighbor.erase(i);
      }
    }
  }
}


/* 
 * If type = 2, swaps all the demands, in the sense that demands with less flow are assigned the flow of the demands with more flow.
 * Each demand is multiplied (type = 0 or 2) or set (type = 1) by a random value derived from the given distribution
 * Values for type: 0 (scale), 1 (set), 2 (swap and scale)
 */
void Topology::changeDemands(RndVar* RV, double seed, unsigned int type)
{
#define SETVALUE 0x1
#define SWAP     0x2
  static MotherOfAll urng(seed);         // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used

  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands, newDemands;

  if (!(checkE2Eflows(&demands)))
    cerr << endl << "WARNING: changeDemands: Checking E2E flows before beginning FAILED" << endl << endl;

  // copy the demands in a vector and sort it based on the amount of flow
  vector<pair<pair<int,int>,double> > V(demands.size());  
  copy(demands.begin(), demands.end(), V.begin());
  if (type & SWAP)
    sort(V.begin(), V.end(), compareSecond<pair<int,int>,double>());

  throughput = 0;
  vector<pair<pair<int,int>,double> >::reverse_iterator r = V.rbegin();

  for (vector<pair<pair<int,int>,double> >::iterator v=V.begin(); v!=V.end(); v++, r++)
  {
    throughput += (newDemands[v->first] = (type & SETVALUE ? 1. : (type & SWAP ? r->second : v->second)) * RV->rv->Next());
    cout << "Demands [" << v->first.first << "," << v->first.second << "] changes from " << v->second << " to " << newDemands[v->first] << endl; 
  }

  /* update the flows */
  for (vector<TGraphNode>::iterator u=node.begin(); u!= node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      for (map<pair<int,int>,double>::iterator f=v->f_e2e.begin(); f!=v->f_e2e.end(); f++)
      {
	v->flow += f->second / demands[f->first] * (newDemands[f->first] - demands[f->first]);
	f->second += f->second / demands[f->first] * (newDemands[f->first] - demands[f->first]);
      }

  if (!(checkE2Eflows(NULL)))
    cerr << endl << "WARNING: changeDemands: Checking E2E flows at the end FAILED" << endl << endl;
}


list<TLinkItrs>* Topology::changeFlowOnRandomPath(int maxLength, bool increase, double percent) {
  MotherOfAll urng(RndVar::getSeed());         // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used
  Uniform U;
  /* the path length is uniformly distributed between 1 and maxLength */
  SumRandom lenRV( (maxLength-1) * U + 1 );
  SumRandom srcRV( (node.size() - 1) * U );
  Uniform nh;
  int next, i = 0;
  double fmin = numeric_limits<double>::max();
  
  /* vector of nodes along the path (used to avoid loops) */
  vector<int> Pnode;
  
  /* the source of the path is chosen randomly */
  Pnode.push_back( (int)floor(srcRV.Next()) );
  
  list<TLinkItrs>* p = new list<TLinkItrs>;
  
  double len = lenRV.Next();
//  list<TGraphNeighbor>::iterator n = node[].
  
  while (p->size() < len) {
    next = (int)floor(nh.Next() * (node[Pnode.back()].neighbor.size() - 1));
    
    list<TGraphNeighbor>::iterator l = node[Pnode.back()].neighbor.begin();
    for (int n=0; n<next; n++)
      l++;
    
    if (find(Pnode.begin(), Pnode.end(), l->Id()) == Pnode.end()) {
      p->push_back(LinkItrs(node.begin() + Pnode.back(), l));

      Pnode.push_back(l->Id());
      
      fmin = min(fmin, p->back().second->flow);
    }
    
    if (i++ > 2 * len)
      break;
  }
  
//  fmin = fmin * (U.Next() * percent) * (increase ? 1 : -1);
  fmin = fmin * (U.Next() * percent) * 3;
  
  for (list<TLinkItrs>::iterator l=p->begin(); l!=p->end(); l++) {
    l->second->flow += fmin;
    
    /* update the Utot of all the affected links */
    for (vector<TGraphNode>::iterator x=node.begin(); x!=node.end(); x++)
      for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	// if u->v belongs to the collision domain of x->y
	if ( (l->PotInterferesWith(LinkItrs(x,y))) && (l->second->radio->channel == y->radio->channel) )
	  y->Utot += fmin/l->second->rate;

    cerr << l->first->Id() << "->" << l->second->Id() << " ";
  }
  cerr << endl;
  
  return p;
}

/*  Erases links with null flow provided that 
 *  either there is a parallel link on another channel
 *  or there are only inverse links with null flow
 */
void Topology::eraseLinksWithNullFlow() {
  list<TGraphNeighbor>::iterator next;
  /* for all the links in the network */
  for (vector<TGraphNode>::iterator x=node.begin(); x!=node.end(); x++)
    for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); ) {
      /* with null flow */
      if (!y->flow) {
	/* if there is a parallel link on another channel */
	if (find_if(x->neighbor.begin(), x->neighbor.end(), 
		__gnu_cxx::compose2(logical_and<bool>(),
		__gnu_cxx::compose1(bind2nd(equal_to<int>(),y->Id()), mem_fun_ref(&TGraphNeighbor::Id) ),
		__gnu_cxx::compose1(bind2nd(not_equal_to<unsigned int>(),y->radio->channel), mem_fun_ref(&TGraphNeighbor::RadioChannel) )) )
		!= x->neighbor.end() ) {
	  /* erase x->y */
	  next = y;
	  next++;
	  cout << "Erasing " << x->Id() << "->" << y->Id() << "[" << y->radio->channel <<"] with flow = " << y->flow << endl;
	  x->neighbor.erase(y);
	  y = next;
	}
	else 
	  /* there are no parallel links */
	  /* if there is no inverse link with non-null flow */
	  if (find_if(node[y->Id()].neighbor.begin(), node[y->Id()].neighbor.end(), 
		  __gnu_cxx::compose2(logical_and<bool>(),
		  __gnu_cxx::compose1(bind2nd(equal_to<int>(),x->Id()), mem_fun_ref(&TGraphNeighbor::Id) ),
		  __gnu_cxx::compose1(bind2nd(not_equal_to<double>(),0.), mem_fun_ref(&TGraphNeighbor::Flow) )) )
		  == node[y->Id()].neighbor.end() ) {
	    /* erase x->y */
	    next = y;
	    next++;
	    cout << "Erasing " << x->Id() << "->" << y->Id() << "[" << y->radio->channel <<"] with flow = " << y->flow << endl;
	    x->neighbor.erase(y);
	    y = next;
	  }
	  else
	    y++;
	}
      else
	y++;
    }
}

bool Topology::isThereLinkWithoutChannel ()
{
  for (TGraphNodeIterator u = node.begin(); u != node.end(); u++)
    for (TGraphNeighborIterator v = u->neighbor.begin(); v != u->neighbor.end(); v++)
      if (!v->RadioChannel())
	return true;

  return false;
}

/* Given a TLinkItrs u->v returns the TLinkItrs v->u */
TLinkItrs Topology::FindInverseLink(TLinkItrs l, unsigned int channel)
{
	vector<TGraphNode>::iterator node1 = find_if(node.begin(), node.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(),l.second->Id()), mem_fun_ref(&TGraphNode::Id)  ));

//cout << "FindInverseLink " << l.first->Id() << "->" << l.second->Id() << endl;

	assert( node1 != node.end() );

	TGraphNeighborIterator node2;

	if (!channel)
		node2 = find_if(node1->neighbor.begin(), node1->neighbor.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(),l.first->Id()), mem_fun_ref(&TGraphNeighbor::Id) ));
	else
		node2 = find_if(node1->neighbor.begin(), node1->neighbor.end(), 
				__gnu_cxx::compose2(logical_and<bool>(),
				__gnu_cxx::compose1(bind2nd(equal_to<int>(),l.first->Id()), mem_fun_ref(&TGraphNeighbor::Id) ),
				__gnu_cxx::compose1(bind2nd(equal_to<unsigned int>(),channel), mem_fun_ref(&TGraphNeighbor::RadioChannel) )) );

// 	if (!channel)
// 		assert( node2 != node1->neighbor.end() );

/*	TLinkItrs* res = NULL;
	if ( node2 != node1->neighbor.end() )
		res = new TLinkItrs(node1,node2);
	return res;*/
	
	return LinkItrs(node1,node2);
}



/* Assign channel c to the (directed) link l
   PRE: for both end nodes, either a radio assigned to c already exists or an available radio exists */
void Topology::AssignChannel2Link(TLinkItrs l, unsigned int c)
{
// 	pair<bool, unsigned int> foundCh;
	int ri;

// 	foundCh = SearchChannel(l.first->info, c);
	ri = l.first->info->findRadioWithChannel(c);
	
// 	if (!foundCh.first)
	if (ri < 0)
	{
// 		foundCh = SearchChannel(l.first->info, 0);
		ri = l.first->info->findRadioWithChannel(0);
// 		assert( foundCh.first );
		assert( ri >= 0 );
// 		l.first->info->radio[foundCh.second].channel = c;
		l.first->info->Radio(ri)->channel = c;
	}
// 	l.second->radio = l.first->info->radio + foundCh.second;
	l.second->radio = &*l.first->info->Radio(ri);

// 	foundCh = SearchChannel(l.second->info, c);
	ri = l.second->info->findRadioWithChannel(c);
// 	if (!foundCh.first)
	if (ri < 0)
	{
// 		foundCh = SearchChannel(l.second->info, 0);
		ri = l.second->info->findRadioWithChannel(0);
// 		assert( foundCh.first );
		assert ( ri >= 0 );
// 		l.second->info->radio[foundCh.second].channel = c;
		l.second->info->Radio(ri)->channel = c;
	}

/*	TLinkItrs InvLink = FindInverseLink(l);
	InvLink.second->radio = l.second->info->radio + foundCh.second;*/
}


TGraphNeighborIterator TGraphNode::FindNeighbor (int v, unsigned int c)
{
  if (c)
    return find_if(neighbor.begin(), neighbor.end(), 
                   __gnu_cxx::compose2(logical_and<bool>(),
                   __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), mem_fun_ref(&TGraphNeighbor::Id) ),
                   __gnu_cxx::compose1(bind2nd(equal_to<unsigned int>(),c), mem_fun_ref(&TGraphNeighbor::RadioChannel) )) );

  return find_if(neighbor.begin(), neighbor.end(),
                 __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), mem_fun_ref(&TGraphNeighbor::Id) ));
}

TGraphNeighborIterator TGraphNode::FindNeighbor (int v, TGraphNeighborIterator n, unsigned int c)
{
  if (c)
    return find_if(n, neighbor.end(), 
                   __gnu_cxx::compose2(logical_and<bool>(),
                   __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), mem_fun_ref(&TGraphNeighbor::Id) ),
                   __gnu_cxx::compose1(bind2nd(equal_to<unsigned int>(),c), mem_fun_ref(&TGraphNeighbor::RadioChannel) )) );

  return find_if(n, neighbor.end(),
                 __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), mem_fun_ref(&TGraphNeighbor::Id) ));
}

// list<TGraphNeighbor>::iterator Topology::find_neighbor(int u, int v)
// {
//   return find_if(node[u].neighbor.begin(), node[u].neighbor.end(),
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), ptr_fun<TGraphNeighbor,int>(getNodeID) ));
//   
// }
// 
// list<TGraphNeighbor>::iterator Topology::find_neighbor(int u, int v, list<TGraphNeighbor>::iterator n)
// {
//   return find_if(n, node[u].neighbor.end(),
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), ptr_fun<TGraphNeighbor,int>(getNodeID) ));
//   
// }
// 
// list<TGraphNeighbor>::iterator Topology::find_neighbor(int u, int v, unsigned int c)
// {
//   return find_if(node[u].neighbor.begin(), node[u].neighbor.end(), 
// 		__gnu_cxx::compose2(logical_and<bool>(),
// 		__gnu_cxx::compose1(bind2nd(equal_to<int>(),v), mem_fun_ref(&TGraphNeighbor::Id) ),
// 		__gnu_cxx::compose1(bind2nd(equal_to<unsigned int>(),c), mem_fun_ref(&TGraphNeighbor::RadioChannel) )) );
// }

/* Find all the links that potentially interfere with l */
vector<TLinkItrs> Topology::computeNeighborhood(TLinkItrs l)
{
	vector<TLinkItrs> neighborhood;

	// consider all the links i<->j
	for (vector<TGraphNode>::iterator i=node.begin(); i!=node.end(); i++)
		for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
			// if i<->j is in the neighborhood of the extracted link
			if (LinkItrs(i,j).PotInterferesWith(l))
			// add i<->j in neighborhood
				neighborhood.push_back(LinkItrs(i,j));

	return neighborhood;
}



/* computes the set of channels used in the topology */
void Topology::getChannelSet(set<unsigned int> &channelSet)
{
	channelSet.clear();
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (v->radio)
				channelSet.insert(v->radio->channel);
}


/* computes the Utot of the collision domain of a link */
double Topology::computeUtot(TLinkItrs l, bool verbose)
{
  double Utot = 0.;

  if (verbose)
    cout << endl << "Collision domain = ";

  // consider all the links i->j
  for (vector<TGraphNode>::iterator i=node.begin(); i!=node.end(); i++)
    for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
      // if i->j potentially interferes with l and has the same channel
      if ((LinkItrs(i,j).PotInterferesWith(l)) && (j->radio) && (j->radio->channel==l.second->radio->channel))
      {
	// add the flow of i->j
	Utot += j->flow / j->rate;

	if (verbose)
	  cout << i->Id() << "->" << j->Id() << " ";
      }

  if (verbose)
    cout << endl;

  return Utot;
}


bool Topology::checkUtot (int verbose, vector<pair<TLinkItrs,double> >** vec)
{
  bool ok = true;
  double Utot;

  if (vec)
    *vec = new vector<pair<TLinkItrs,double> >;

  for (vector<TGraphNode>::iterator u = node.begin(); u != node.end(); u++)
    for (list<TGraphNeighbor>::iterator v = u->neighbor.begin(); v != u->neighbor.end(); v++)
      if (v->radio)
      {
	if (verbose>=2 || (verbose==1 && fabs(v->Utot - Utot)>0.0001))
	  cout << "Link " << u->Id() << "->" << v->Id() << "[" << v->radio->channel << "] ";

	Utot = computeUtot(LinkItrs(u,v), verbose==3);

	if (vec)
	  (*vec)->push_back (pair<TLinkItrs,double>(LinkItrs(u,v),Utot));

	if (verbose>=2 || (verbose==1 && fabs(v->Utot - Utot)>0.0001))
	  cout << "Utot = " << Utot;

	if (verbose >= 1 && fabs(v->Utot - Utot)>0.0001)
	  cout << "  instead the Utot field contains " << v->Utot;

	if (verbose>=2 || (verbose==1 && fabs(v->Utot - Utot)>0.0001))
	  cout << endl;

	if (fabs(v->Utot - Utot)>0.0001)
	  ok = false;
      }

  return ok;
}


/* returns the maximum among the utilization of the collision domains of all the links */
pair<TLinkItrs,double> Topology::getUtotMax()
{
	vector< pair<TLinkItrs,double> >* UtotVect;
	
	checkUtot(0, &UtotVect);
	
	pair<TLinkItrs,double> p = *max_element(UtotVect->begin(), UtotVect->end(), compareSecond<TLinkItrs,double>());

	delete UtotVect;

	return p;
}


/* scales all the precomputed flow rates and the total utilizations */
void Topology::scaleFlowsByUtotMax()
{
	double UtotMax = getUtotMax().second;
	
	if (UtotMax > 1.)
	{
		// consider all the links i<->j
		for (vector<TGraphNode>::iterator i=node.begin(); i!=node.end(); i++)
			for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
			{
				j->flow /= UtotMax;
				j->Utot /= UtotMax;
			}
	}
}


/* Update the Utot for every link involved (except l itself) in case l is assigned channel c and rate r.
 * Must be invoked before making the change 
 */
void Topology::updateUtot (TLinkItrs l, unsigned int c, float r)
{
	/* the new Utot for the other links is obtained from the current Utot value */
	// for every directed link x->y
	for (vector<TGraphNode>::iterator x = node.begin(); x != node.end(); x++)
		for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
			// such that the potential collision domain of x->y includes l
			if ( (l!=LinkItrs(x,y)) && (y->radio) && (l.PotInterferesWith(LinkItrs(x,y))) )
			{
				/* by subtracting the F/C ratio of l if l escapes the collision domain of l0 */
				if (l.second->radio && (y->radio->channel == l.second->radio->channel) && (y->radio->channel != c))
					y->Utot -= l.second->flow / l.second->rate;
				/* by adding the F/C ratio of l if l joins the collision domain of l0 */
				else if ((!l.second->radio || y->radio->channel != l.second->radio->channel) && (y->radio->channel == c))
					y->Utot += l.second->flow / r;
			}
}



/* computes the degree of the network (i.e. average number of neighbors per node) */
double Topology::getDegree()
{
	double degree = 0.;
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		degree += u->neighbor.size();
	degree = degree / node.size();

	return degree;
}

class TNodeVisit {
  public:
	int predecessor; // -1 = NIL
	int color;	// 0 = white, 1 = gray, 2 = black
	int component;	// a number identifying each connected component (-1 = no component yet)

	/* constructor */
	TNodeVisit() : predecessor(-1), color(0), component(-1) {}
};

void DFS_Visit(Topology* T, vector<TNodeVisit> & nodes, unsigned int u, int compId) {
	nodes[u].color = 1;
	nodes[u].component = compId;
	for (list<TGraphNeighbor>::iterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
		if (!nodes[v->Id()].color) {
			nodes[v->Id()].predecessor = u;
			DFS_Visit(T, nodes, v->Id(), compId);
		}
	nodes[u].color = 2;
}

/* Performs a depth-first-visit and returns the number of connected components */
int DFS(Topology* T, vector<TNodeVisit> & nodes) {
	int compId = 0;
	for (unsigned int i=0; i<nodes.size(); i++)
		if (!nodes[i].color) {
			DFS_Visit(T, nodes, i, compId);
			compId++;
		}
	return compId;
}

/* Given two connected components in a graph, attempts to merge them by adding a link between
   a pair of nodes belonging to the two distinct components
   If the merge succeeds, nodes in the component comp2 are assigned to comp1 and 
   the function returns true 
*/
bool connectComponents(Topology* T, vector<TNodeVisit> & nodes, int comp1, int comp2) {
	int c, p, imin, imax;
// 	vector<TGraphNode>::iterator u = T->node.begin();
	int u = 0;

	vector<int> nodes_comp1, nodes_comp2;

	for (unsigned int i=0; i<nodes.size(); i++)
		if (nodes[i].component == comp1)
			nodes_comp1.push_back(i);
	// cout << "Componente " << comp1 << ": "; for (vector<int>::iterator i=nodes_comp1.begin(); i!=nodes_comp1.end(); i++) cout << *i << " "; cout << endl;
	for (unsigned int i=0; i<nodes.size(); i++)
		if (nodes[i].component == comp2)
			nodes_comp2.push_back(i);
	// cout << "Componente " << comp2 << ": "; for (vector<int>::iterator i=nodes_comp2.begin(); i!=nodes_comp2.end(); i++) cout << *i << " "; cout << endl;
	
	for (vector<int>::iterator i1=nodes_comp1.begin(); i1!=nodes_comp1.end(); i1++)
		for (vector<int>::iterator i2=nodes_comp2.begin(); i2!=nodes_comp2.end(); i2++)
		{
			T->canBeNeighbors(u + *i1, u + *i2, c, p);
			if (c>=0)
			{
				imin = min (*i1, *i2);
				imax = max (*i1, *i2);
				T->addNeighbor(u + imin,u + imax,WMS.availTXrates(c),p);
				T->addNeighbor(u + imax,u + imin,WMS.availTXrates(c),p);

				// cout << "connectComponents: Aggiungo link " << *i1 << "->" << *i2 << " per il merge di componenti " << comp1 << " e " << comp2 << endl;
				
				for (vector<int>::iterator i=nodes_comp2.begin(); i!=nodes_comp2.end(); i++)
					nodes[*i].component = comp1;
				return true;
			}
		}
	return false;
}

/* If connect is false, just checks whether the graph is connected
   Otherwise, if there are multiple connected components, it attempts to add links
   in order to get a connected graph
*/

bool Topology::isConnected(bool connect)
{
  vector<TNodeVisit> nodes(node.size());

  int numComp = DFS(this, nodes);
  cout << "Number of connected components in the graph: " << numComp << endl;

  if (!connect || numComp == 1)
    return (numComp == 1);

  cout << endl << "WARNING: The graph is not connected. Trying to make it connected by adding links using increased transmission power... ";

  set<int> comp_set, erased;
  for (int i=1; i<numComp; i++)
    comp_set.insert(i);

  do
  {
    erased.clear();
    for (set<int>::iterator comp2=comp_set.begin(); comp2!=comp_set.end(); comp2++)
      if (connectComponents(this, nodes, 0, *comp2)) 
	erased.insert(*comp2);

    for (set<int>::iterator i=erased.begin(); i!=erased.end(); i++)
      comp_set.erase(*i);
  } while (!comp_set.empty() && !erased.empty());

  if (comp_set.empty())
    cout << "DONE" << endl << endl;

  return (comp_set.empty());
}


/* Applies Dijkstra's algorithm to the topology
 * Returns a vector containing the distance of each node to the source
 * Links for which prune is true are not considered
 * The link cost function is computed by the cost function object
 */
vector<float> Topology::Dijkstra (int source, bool isSink, linkPruneFunc* prune, linkCostFunc* cost)
{
  SubGraph* my_graph = new SubGraph (this, isSink);

  my_graph->pruneEdges (prune);
  my_graph->setLinkCosts (cost);

  DijkstraShortestPathAlg dijkstra (my_graph);

  dijkstra.get_shortest_path(my_graph->get_vertex(source), NULL);

  vector<float> distance (node.size());

  for (int i = 0; i < node.size(); i++)
    distance.at(i) = dijkstra.get_start_distance_at (my_graph->get_vertex(i));

  return distance;
}


/* Applies Dijkstra algorithm to topology T
 * Returns the shortest feasible path between source and dest
 * TODO correctly handle parallel links between nodes
 */
vector<TLinkItrs>* Topology::ShortestPath(int source, int dest, linkPruneFunc* prune, linkCostFunc* cost)
{
  SubGraph* my_graph = new SubGraph (this);

  my_graph->pruneEdges (prune);
  my_graph->setLinkCosts (cost);

  DijkstraShortestPathAlg dijkstra (my_graph);
  
  BasePath* path = dijkstra.get_shortest_path(my_graph->get_vertex(source), my_graph->get_vertex(dest));

  vector<TLinkItrs>* result = new vector<TLinkItrs>;

  for (int i = 1; i < path->length(); i++)
  {
    TGraphNeighborIterator v = node[path->GetVertex(i-1)->getID()].FindNeighbor (path->GetVertex(i)->getID());

    assert (v != node[path->GetVertex(i-1)->getID()].NeighborEnd());

    result->push_back (LinkItrs(Node(path->GetVertex(i-1)->getID()),v));
  }

  delete path;

  return result;
}

/*
 * Find the first (at most k) shortest paths with length at most alpha times the length
 * of the shortest path
 * Returns the set of links on those paths
 */
set<pair<int,int> > Topology::find_path(pair<int,int> d, float alpha, int k)
{
  set<pair<int,int> > fksp;
  int count = 1;

  SubGraph my_graph(this);
  
  YenTopKShortestPathsAlg yenAlg(my_graph, my_graph.get_vertex(d.first), my_graph.get_vertex(d.second));

  BasePath* shortest = yenAlg.get_shortest_path(my_graph.get_vertex(d.first), my_graph.get_vertex(d.second));
  // shortest->PrintOut(cout);
  
  while(yenAlg.has_next()) {
    BasePath* path = yenAlg.next();
    
    if (path->length() > alpha*shortest->length() || count > k)
      break;
    
    // path->PrintOut(cout);

    for (int ver=0; ver<path->length()-1; ver++)
      fksp.insert(pair<int,int>(path->GetVertex(ver)->getID(),path->GetVertex(ver+1)->getID()));
    
    count++;
  }
  
  return fksp;
}

set<pair<int,int> > Topology::find_links_for_demand (pair<int,int> d)
{
  set<pair<int,int> > links;

  for (TGraphNodeIterator u = node.begin(); u != node.end(); u++)
    for (TGraphNeighborIterator v = u->neighbor.begin(); v != u->neighbor.end(); v++)
      if (v->IsDemandPresent (d))
	links.insert (pair<int,int>(u->Id(),v->Id()));

  return links;
}

/* Helper function to add a BasePath to a DAG */
static void addPathToDAG (DAG& dag, BasePath* path) {
  set<pair<int,int> > tmp;

  for (int ver=1; ver < path->length(); ver++)
    tmp.insert(pair<int,int>(path->GetVertex(ver-1)->getID(),path->GetVertex(ver)->getID()));

  dag.addLinks(tmp);
}


set<pair<int,int> > Topology::FindLooplessResilientDAG(int s, int d, float alpha, set<pair<int,int> > (Topology::*pmi) (pair<int,int>))
{
  SubGraph my_graph(this);

  BasePath* path;
  bool found;
  int Lsu, Lud;
  
  // create a DAG
  DAG dag(node.size());

  dag.addLinks ((*this.*pmi) (pair<int,int>(s,d)));

  // run a DFS to sort the nodes, find predecessors
  dag.DFS(s);
  
  // dag.Print(cout);

  int minlen = dag.getMaxDistToDest(s);

  // a node is "done" when it is protected against single link/node failures
  vector<bool> done(node.size(), false);
  
  /*
   * nodes are protected starting from the penultimate node in the topological ordering
   */ 
  
  list<int>::const_reverse_iterator u = ++dag.sorted_rbegin();
  set<int>::const_iterator v;
  
  while (u != dag.sorted_rend())
  {

#ifdef DEBUG_RDAS
    cout << endl << "Analyzing node " << *u << endl;
    cout << "num adj: " << dag.getNumAdj(*u) <<  endl;
#endif
    
    // if u is done, continue with its predecessor
    if (done[*u])
    {
      u++;
      continue;
    }
    
    assert(dag.getNumAdj(*u)>0);
    
    // we check whether u is able to reach d despite the failure of each of its neighbors.
    // As soon as this is not true, an alternative path is sought and added to the dag; then,
    // the while loop starts again from the penultimate node
    // for (set<int>::const_iterator v=dag.adj_begin(*u); v!=dag.adj_end(*u); v++) {
    if (dag.getNumAdj(*u) == 1)
    {
      // copy the current dag and prune the neighbor
      // DAG pruned_dag(dag);

      v = dag.adj_begin(*u);
      
/*      if (*v == d)
	pruned_dag.removeLink(*u,d);
      else
	pruned_dag.removeNode(*v);*/
      
      // check whether a path still exists in pruned_dag between u and d
      // this can be done by running DFS and checking whether d has a valid predecessor
      // pruned_dag.DFS(*u);
	
      // if a path is not found
      // if (pruned_dag.getPred(d) == pruned_dag.getNumNodes()) {
	// we try to find an alternative path in the graph
	SubGraph pruned_graph(my_graph);
	
	if (*v == d)
	{
	  pruned_graph.removeEdge(*u,d);
#ifdef DEBUG_RDAS
	  cout << "prune link " << *u << "->" << d << endl;
#endif
	}
	else
	{
	  pruned_graph.removeVertex(*v);
#ifdef DEBUG_RDAS
	  cout << "prune node " << *v << endl;
#endif
	}

	YenTopKShortestPathsAlg yen(pruned_graph, pruned_graph.get_vertex(*u), pruned_graph.get_vertex(d));

//	cout << (yen.has_next() ? "true" : "false") << endl;
	
	found = false;
	Lsu = dag.getMaxDistFromSource(*u);
	Lud = 0;

//	cout << "Lsu = " << Lsu << endl;

	while(yen.has_next() && !found && (Lsu + Lud <= alpha * minlen) )
	{
		path = yen.next();
		
#ifdef DEBUG_RDAS
		cout << "Trying path... ";
		path->PrintOut(cout);
#endif
		
		// try to add path to the dag and check for loops and maximum length
		DAG augm_dag(dag);

		addPathToDAG(augm_dag, path);
		
		augm_dag.DFS(s);

		if (augm_dag.isAcyclic() && augm_dag.getMaxDistToDest(s) <= alpha * minlen) {
		  found = true;
#ifdef DEBUG_RDAS
		  cout << "OK" << endl;
#endif
		}
#ifdef DEBUG_RDAS
		else if (!augm_dag.isAcyclic())
		  cout << "Loop in the DAG" << endl;
		else
		  cout << "Too long!" << endl;
#endif

		Lud = path->length() - 1;
#ifdef DEBUG_RDAS
		cout << "Lud = " << Lud << endl;
#endif
	}

	if (found)
	{
	  done[*u] = true;

	  // add the path to the DAG
	  addPathToDAG(dag, path);
	  
	  // sort the nodes in the new dag
	  dag.DFS(s);
	  
	  // restart from the penultimate node
	  u = ++dag.sorted_rbegin();
	  
//	  dag.Print(cout);
	  
	  continue;
	}
      //}
    }

    done[*u] = true;
    u++;
    
    // dag.Print(cout);
  }
  
  // dagLinks contains all the links of the dag
  set<pair<int,int> > dagLinks;
  
  for (list<int>::const_iterator p=dag.sorted_begin(); p!=dag.sorted_end(); p++)
    for (set<int>::const_iterator q=dag.adj_begin(*p); q!=dag.adj_end(*p); q++)
      dagLinks.insert(pair<int,int>(*p,*q));
  
  return dagLinks;
}

void Topology::AddNullFlowOnRDASlinks (float alpha)
{
  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands;

  if (!(checkE2Eflows(&demands)))
    cerr << endl << "WARNING: AddNullFlowOnRDASlinks: Checking E2E flows FAILED" << endl << endl;

  // for each demand
  for (map<pair<int,int>,double>::iterator d = demands.begin(); d != demands.end(); d++)
  {
    // run RDAS to compute a DAG that "protects" all the nodes carrying the flow of the current demand
    set<pair<int,int> > links = FindLooplessResilientDAG (d->first.first, d->first.second, alpha, &Topology::find_links_for_demand);

    // for each link of such DAG
    for (set<pair<int,int> >::iterator l = links.begin(); l != links.end(); l++)
    {
      TGraphNeighborIterator v = node.at(l->first).FindNeighbor (l->second);

      if (v == node.at(l->first).NeighborEnd() )
      {
	cerr << endl << "WARNING: AddNullFlowOnRDASlinks: " << l->second << " not found among the neighbors of " << l->first << endl << endl;
	continue;
      }

      // add a null flow if the link does not carry flow of the current demand
      if (!v->IsDemandPresent (d->first))
	v->f_e2e[d->first] = 0.;
    }
  }
}

/* copies the flow value in the availBW field */
void Topology::InitRouting()
{
	// for every link
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->availBW = v->flow;
}


/*
   Route flows along the shortest path
   AvailBW is used as link capacity
   Returns the cross-section goodput
 */
double Topology::RouteFlows(int NumFlows, RndVar *Source, RndVar *Dest, RndVar *Bwdr, double seed)
{
	MotherOfAll urng(seed);              // declare uniform random number generator
	Random::Set(urng);                   // set urng as generator to be used

	vector<TLinkItrs>* path;
	int src, dst, Accepted = 0;
	double bw, CSgoodput = 0.;

	for (int flowcounter=0; flowcounter<NumFlows; flowcounter++)
	{
		src = aggregation_list[(int)floor(Source->rv->Next())];
		dst = gateway_list[(int)floor(Dest->rv->Next())];
		bw = (double)Bwdr->rv->Next();
		cout << "Flow " << flowcounter << ": src=" << src << " dst=" << dst << " bw=" << bw << "\t Path: ";

		linkPruneFunc* prune = new linkPrune_Bwd (bw);
		path = ShortestPath(src, dst, prune);

		if (!path->empty())
		{
			Accepted ++;
			CSgoodput += bw;

			/* decrease availBW along the path and print the path */
			// vector<int>::iterator cur_node = path.begin();
			// vector<TGraphNeighbor>::iterator nexthop;

			cout << path->begin()->first->Id() << " ";

			for (vector<TLinkItrs>::iterator l=path->begin(); l!=path->end(); l++) {
				l->second->availBW -= 1.0;
				cout << l->second->Id() << " ";
			}

/*			while (cur_node+1 != path.end() )
			{
				nexthop = find_if(node[*cur_node].neighbor.begin(), node[*cur_node].neighbor.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(),*(cur_node+1)), ptr_fun<TGraphNeighbor,int>(getNodeID) ));
		
				assert( nexthop != node[*cur_node].neighbor.end() );
		
				nexthop->availBW -= bw;
		
				cout << *++cur_node << " ";
			}*/
		}
		else
			cout << "Not found";
		cout << endl;

		delete prune;
		delete path;
	}
	cout << endl << "Accepted: " << Accepted << " of " << NumFlows << endl << "Cross-section Goodput: " << CSgoodput << endl << endl;
	return CSgoodput;
}

int Topology::countRadiosOff() {
  int i = 0;
  for (vector<TGraphNode>::iterator n=node.begin(); n!=node.end(); n++)
    i += n->info->countRadiosWithChannel(0);
  return i;
}

int Topology::countNodesOff() {
  int i=0;
  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
    if (u->info->countRadiosWithChannel(0) == u->info->numAvailRadios())
      i++;
  return i;
}

int Topology::countNumRadios() {
  int i = 0;
  for (vector<TGraphNode>::iterator n=node.begin(); n!=node.end(); n++)
    i += n->info->numAvailRadios();
  return i;
}


int Topology::countRadioChanged (Topology* Tnew, vector<multiset<unsigned int> >& oldChSet, vector<multiset<unsigned int> >& newChSet)
{
  multiset<unsigned int> commonCh;
  int changes = 0;
  
  // the old and new topologies must have the same number of nodes
  assert (node.size() == Tnew->node.size());
  
  oldChSet.resize(node.size());
  newChSet.resize(node.size());

  for (int u = 0; u < (int)node.size(); u++)
  {
    // this node must have the same number of radios in both topologies
    assert (Tnew->node[u].info->numAvailRadios() == node[u].info->numAvailRadios());
    
    oldChSet[u].clear();
    newChSet[u].clear();
    commonCh.clear();

    for (unsigned int j=0; j<node[u].info->numAvailRadios(); j++)
    {
      oldChSet[u].insert (node[u].info->Radio(j)->channel);
      newChSet[u].insert (Tnew->node[u].info->Radio(j)->channel);
    }

    set_intersection (oldChSet[u].begin(), oldChSet[u].end(), newChSet[u].begin(), newChSet[u].end(), inserter(commonCh, commonCh.begin()));
    
    for (multiset<unsigned int>::iterator c = commonCh.begin(); c != commonCh.end(); c++)
    {
      oldChSet[u].erase (*c);
      newChSet[u].erase (*c);
    }
    
    changes += newChSet[u].size();
  }

  return changes;
}


/* Computes the power consumption of all the nodes and the time usage (tau) for each of their radios.
  * Results are stored in the vector consumption (which can also be empty when passed) */  
void Topology::calcAllNodesPowerConsumption(vector<TPower>& consumption)
{
  consumption.clear();
  
  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
//     consumption.push_back(TPower(u->info->numAvailRadios(), pm.BasePowerConsumption));
    consumption.push_back(TPower(u->info->numAvailRadios(), u->info->GetBasePowerConsumption()));

  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
    {
//       consumption[u->Id()].power += v->flow / v->rate * (pm.rho_tx - pm.rho_idle);
      consumption[u->Id()].power += v->flow / v->rate * ((*v->radio->rho_tx)(DbmToW(WMS.TxPowerLevel(v->txpower))) - v->radio->rho_idle);

//       consumption[u->Id()].tau.at(v->radio - u->info->radio) += v->flow / v->rate;
      consumption[u->Id()].tau.at(v->radio - &*u->info->Radio(0)) += v->flow / v->rate;

//       pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
      int ri = v->info->findRadioWithChannel(v->radio->channel);
//       if (ch.first)
      if (ri >= 0)
      {
// 	consumption[v->Id()].power += v->flow / v->rate * (pm.rho_rx - pm.rho_idle);
	consumption[v->Id()].power += v->flow / v->rate * (v->info->Radio(ri)->Rho_Rx() - v->info->Radio(ri)->Rho_Idle());

// 	consumption[v->Id()].tau.at(ch.second) += v->flow / v->rate;
	consumption[v->Id()].tau.at(ri) += v->flow / v->rate;
      }
    }

  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
  {
    if (u->info->countRadiosWithChannel(0) == u->info->numAvailRadios())
//       consumption.at(u->Id()).power = pm.StandbyPowerConsumption;
      consumption.at(u->Id()).power = u->info->GetStandbyPowerConsumption();
    else
//       consumption.at(u->Id()).power += u->info->countRadiosWithChannel(0)*pm.rho_sleep + (u->info->numAvailRadios()-u->info->countRadiosWithChannel(0))*pm.rho_idle;
//       for (RadioIfaceIterator ri = u->info->radio.begin(); ri != u->info->radio.end(); ri++)
      for (int ri = 0; ri < u->info->GetNumRadios(); ri++)
// 	consumption.at(u->Id()).power += u->info->countRadiosWithChannel(0)*pm.rho_sleep + (u->info->numAvailRadios()-u->info->countRadiosWithChannel(0))*pm.rho_idle;
	consumption.at(u->Id()).power += (u->info->Radio(ri)->channel ? u->info->Radio(ri)->Rho_Idle() : u->info->Radio(ri)->Rho_Sleep());
  }
}


/* checks whether the power constraints on all the nodes and the time constraints (tau) on all the radios are satisfied.
 * If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
 * can be passed so that calcAllNodesPowerConsumption is not called again
 */
bool Topology::checkPowerConstraints (vector<float>* nodePowerConstr, vector<TPower>* consumption)
{
  bool computed = false;

  if (!consumption)
  {
    consumption = new vector<TPower>;
    computed = true;
    calcAllNodesPowerConsumption(*consumption);
  }

  for (unsigned int u=0; u<node.size(); u++)
  {
    if (consumption->at(u).power > nodePowerConstr->at(u))
    {
      if (computed)
	delete consumption;
      return false;
    }

    for (unsigned int r=0; r<node[u].info->numAvailRadios(); r++)
      if (consumption->at(u).tau.at(r) > 1.f)
      {
	if (computed)
	  delete consumption;
	return false;
      }
  }

  if (computed)
    delete consumption;

  return true;
}


/* computes the total power consumption of the whole network.
 * If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
 * can be passed so that calcAllNodesPowerConsumption is not called again
 */
float Topology::TotalPowerConsumption (vector<TPower>* consumption)
{
  bool computed = false;

  if (!consumption)
  {
    consumption = new vector<TPower>;
    computed = true;
    calcAllNodesPowerConsumption(*consumption);
  }

  float Ptot = 0.f;
  for (vector<TPower>::iterator i=consumption->begin(); i!=consumption->end(); i++)
    Ptot += i->power;
  
  if (computed)
    delete consumption;

  return Ptot;
}


void Topology::printTopology(ostream& f)
/* file f must be already open for writing*/
{
  streamsize prec = f.precision();
  // use fixed (as opposed to scientific) floating-point notation
  f.setf (ios::fixed, ios::floatfield);
  f.precision (2);
  f << "Number of available channels: " << WMS.numAvailChannels() << endl;
  f << "Node id\t type\t x_pos\t y_pos\t #radios #radio_off\tchannels\t" << endl;
  for (vector<TGraphNode>::const_iterator i=node.begin(); i!=node.end(); i++)
  {
    f << i->Id() << "\t " << MeshTypes[i->info->Role()] << "\t " << i->info->Xpos()
      << "\t " << i->info->Ypos()<< "\t " <<  i->info->numAvailRadios()
      << "\t " << i->info->countRadiosWithChannel(0) << "\t\t";
    for (unsigned int j=0; j<i->info->numAvailRadios(); j++)
      f << i->info->Radio(j)->channel << " ";
    f << endl;
  }

  f << endl << "Number of radios off: " << countRadiosOff() << endl;

  f << endl << "Total number of radios: " << countNumRadios() << endl;

  f << endl << "Aggregation devices" << endl;
  for (vector<int>::const_iterator i=aggregation_list.begin(); i!=aggregation_list.end(); i++)
    f << *i << " ";

  f << endl << "Gateways" << endl;
  for (vector<int>::const_iterator i=gateway_list.begin(); i!=gateway_list.end(); i++)
    f << *i << " ";

  f << endl << endl << "Interference model: " << WMS.InterfModel()->Description() << endl;
  f << endl << "Propagation model: " << WMS.PropagationModel()->Description() << endl;

  f << endl << endl << "Node and radio power consumption (Watts)" << endl;
  f.precision(3);
  for (TGraphNodeIterator u = node.begin(); u != node.end(); u++)
  {
    f << "Node " << u->Id() << " [bpc=" << u->Info()->GetBasePowerConsumption() << ", spc=" << u->Info()->GetStandbyPowerConsumption() << "]" << endl;
    for (int i = 0; i < u->Info()->GetNumRadios(); i++)
      cout << "  Radio(" << i << ") [tx=" << u->Info()->Radio(i)->Rho_Tx()->GetNameWithParams()
           << ", rx=" << u->Info()->Radio(i)->Rho_Rx() << ", idle=" << u->Info()->Radio(i)->Rho_Idle()
	   << ", sleep=" << u->Info()->Radio(i)->Rho_Sleep() << "]" << endl;
  }

  f << endl << endl << "Neighbors" << endl;
  f.precision(4);
  for (vector<TGraphNode>::const_iterator i=node.begin(); i!=node.end(); i++)
  {
    f << i->Id() << ":";
    for (list<TGraphNeighbor>::const_iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
    {
      f << "\t" << j->Id() << "(d=" << dist(i->Id(),j->Id())
        << ", f=" << j->flow << ", C=" << j->rate << ", P=" << WMS.TxPowerLevel(j->txpower) << " dbm" ;
      if (j->radio != NULL)
	f << ", c=" << j->radio->channel;
      f << ") {";
    for (map<pair<int,int>,double>::const_iterator q=j->f_e2e.begin(); q!=j->f_e2e.end(); q++)
      f << "f[" << q->first.first << "," << q->first.second << "] = " << q->second << " ";
    f << "}" << endl;
    }
  }
  f.precision(prec);
  f.unsetf (ios::floatfield);
}


void Topology::printStats(const char* name, bool verbose)
{
	set<unsigned int> channelSet;
	double Umax, Umin, Umean, Ustdev;
	vector<pair<TLinkItrs,double> >* m;

	cout << "*** " << name << " ***  " << endl;

	if (verbose)
	{
		getChannelSet(channelSet);
		for (set<unsigned int>::iterator i=channelSet.begin(); i!=channelSet.end(); i++)
		cout << *i << "  ";
		cout << endl << "Average node degree " << getDegree() << endl;
		
		printTopology();
	}

	// verbosity 2 => print the Utot of all the links
	checkUtot(2, &m);

	vector<double> u(m->size());

	transform(m->begin(), m->end(), u.begin(), __gnu_cxx::select2nd<pair<TLinkItrs,double> >());

	Umax = *max_element(u.begin(), u.end());
	Umin = *min_element(u.begin(), u.end());
	Umean = accumulate(u.begin(), u.end(), 0.) / (double)u.size();

	/* compute the standard deviation */
	transform(u.begin(), u.end(), u.begin(), bind1st(minus<double>(),Umean));
	transform(u.begin(), u.end(), u.begin(), u.begin(), multiplies<double>());

	Ustdev = sqrt(accumulate(u.begin(), u.end(), 0.) / (double)u.size());

	cout << "Maximum Utot = " << Umax << endl;
	cout << "Average Utot = " << Umean << endl;
	cout << endl;

	/* the maximum Utot only is sent to the standard error */
	cerr << name << "\t" << Umean - Ustdev << "\t" << Umin << '\t' << Umax << '\t' << Umean + Ustdev << '\t' << countRadiosOff() << endl;

	delete m;
}


/* returns the distance between two nodes */
// float Topology::dist(int x, int y)
// {
// 	return sqrtf(pow(fabsf(node[x].info->Xpos() - node[y].info->Xpos()),2)+pow(fabsf(node[x].info->Ypos() - node[y].info->Ypos()),2));
// }


/* creates an octave file named draw_<toponame>.m to draw the topology */
void Topology::drawTopology()
{
	const float ratio = 0.05;
	float xmin = node[0].info->Xpos(),
	      xmax = 0.,
	      ymin = node[0].info->Ypos(),
	      ymax = 0.,
	      radius, dist, Dx, Dy;

	/* compute the minimum and maximum coordinates */
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
	{
		if (u->info->Xpos() < xmin)
			xmin = u->info->Xpos();
		else if (u->info->Xpos() > xmax)
			xmax = u->info->Xpos();

		if (u->info->Ypos() < ymin)
			ymin = u->info->Ypos();
		else if (u->info->Ypos() > ymax)
			ymax = u->info->Ypos();
	}

	radius = ratio * (xmax - xmin);

	char fname[100];
	sprintf(fname,"draw_%s.m",toponame.c_str());

	ofstream f;
	f.open(fname);

	f << "printf(" << '"' << "Legenda:\\n\\n 54Mbps -> green   (thick) \\n 48Mbps -> cyan    (thick) \\n 36Mbps -> magenta (thick) \\n 24Mbps -> blue    (thick)\\n" << '"' << ")" << endl;
	f << "printf(" << '"' << " 18Mbps -> magenta \\n 12Mbps -> blue \\n  9Mbps -> cyan \\n  6Mbps -> green\\n\\n" << '"' << ")" << endl;

	/* clear the current figure and hold on */
	//f << "#! /usr/bin/octave -qf" << endl << endl;
	f << "clf" << endl << "hold on" << endl;

	/* draw a labelled circle for each node */
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		f << "x = " << u->info->Xpos() - radius << ":0.1:" << u->info->Xpos() + radius << ";" << endl
		  << "plot(x, " << u->info->Ypos() << " + sqrt( " << radius*radius << " - (x - " << u->info->Xpos() << ").^2), \"" << 3-u->info->Role() << "\" )" << endl
		  << "plot(x, " << u->info->Ypos() << " - sqrt( " << radius*radius << " - (x - " << u->info->Xpos() << ").^2), \"" << 3-u->info->Role() << "\" )" << endl
		  << "text(" << u->info->Xpos() << ", " << u->info->Ypos() << ", \"" << u->Id() << "\")" << endl;

	/* draw links */
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->Xpos() <= v->info->Xpos())          /* draw one undirected link between two connected nodes */
			{
				/* distance between the centers of the nodes */
				dist = sqrt(pow(u->info->Xpos() - v->info->Xpos(),2) + pow(u->info->Ypos() - v->info->Ypos(),2));
				if (dist <= 2 * radius)            /* nodes intersect with each other */
					continue;
				Dx = radius * fabs(u->info->Xpos() - v->info->Xpos()) / dist;
				Dy = radius * fabs(u->info->Ypos() - v->info->Ypos()) / dist;

				if (v->info->Xpos() - u->info->Xpos() > 0.1 + 2 * Dx)  /* there is at least one element in the vector x (non-vertical lines) */
					f << "x = " << u->info->Xpos() + Dx << ":0.1:" << v->info->Xpos() - Dx << ";" << endl
					<< "plot(x, " << u->info->Ypos() << " + " << (v->info->Ypos() - u->info->Ypos()) / (v->info->Xpos() - u->info->Xpos())
					<< " * (x - " << u->info->Xpos() << ")";
				else
				{
					if (u->info->Ypos() < v->info->Ypos())
						f << "x = " << u->info->Ypos() + Dy << ":0.1:" << v->info->Ypos() - Dy << ";" << endl;
					else
						f << "x = " << v->info->Ypos() + Dy << ":0.1:" << u->info->Ypos() - Dy << ";" << endl;
					f << "plot(" << u->info->Xpos() << " + " << (v->info->Xpos() - u->info->Xpos()) / (v->info->Ypos() - u->info->Ypos())
					<< " * (x - " << u->info->Ypos() << "), x";
				}
				/* set line color and width */
				f << "," << '"' << (v->rate < 20 ? ((int)v->rate) % 5 + 1 : ((int)v->rate) % 11 % 9 + 1 ) << '"' << ", " << '"' <<
				     "linewidth" << '"' << ", " << (v->rate < 20 ? 1 : 3) << " )" << endl;
			}

	f << "hold off" << endl;

	f.close();
}

#if 0
int createTopology(Topology **T)
{
	char f_topology[60],    /* name of the file containing the topology */
		msg[80]="Unexpected value or EOF while reading ",
		msg_power[30],
		msg_rate[30],
		buildGiFuncStr[30],
		filenameLinks[256];

	unsigned int NumChannels, power;
	int min_rate, max_rate;
	
	/* read the number of available channels */
	READ_PARAM(stdin,"%ud",&NumChannels,strcat(msg,"the number of channels"));
	
	/* read topology file name */
	READ_PARAM(stdin,"%59s",f_topology,strcat(msg,"topology file name"));
	
// 	*T = new Topology(NumChannels);
// 	
// 	(*T)->getNodeInfo(f_topology);

	*T = new Topology (f_topology);

	(*T)->numAvailChannels = NumChannels;

	string s(f_topology);
	strcpy((*T)->topopath, s.c_str());
	string::size_type pos2 = s.rfind(".");
	string::size_type pos1 = s.rfind("/");
	strcpy((*T)->toponame, s.substr(pos1+1,pos2-pos1-1).c_str());

	/* read buildGi function */
	READ_PARAM(stdin,"%29s",buildGiFuncStr,strcat(msg,"buildGiFunc name"));
	
	sprintf(msg_power,"the power level (0-%ld expected)",NUMTXPWR-1);
	sprintf(msg_rate,"the rate index (0-%ld expected)",NUMRATES-1);

	if (!strcmp(buildGiFuncStr,"FixedPowerMaxAllowRate"))
	{
		/* read the power value (dbm) */
		READ_PARAM(stdin,"%ud",&power,strcat(msg,msg_power));
		if ( (power < 0) || (power > NUMTXPWR-1) )
			PRINT_ERROR(strcat(msg,msg_power));
		/* read the min rate index */
		READ_PARAM(stdin,"%d",&min_rate,strcat(msg,msg_rate));
		if ( (min_rate < 0) || (min_rate > (int)(NUMRATES-1)) )
			PRINT_ERROR(strcat(msg,msg_rate));
		/* read the max rate index */
		READ_PARAM(stdin,"%d",&max_rate,strcat(msg,msg_rate));
		if ( (max_rate < 0) || (max_rate > (int)(NUMRATES-1)) )
			PRINT_ERROR(strcat(msg,msg_rate));
		(*T)->buildGi_FixedPowerMaxAllowRate(availTxPwr[power], min_rate, max_rate);
		(*T)->maxP = power;
		(*T)->min_rate = min_rate;
		(*T)->max_rate = max_rate;
	} else if (!strcmp(buildGiFuncStr,"FromFile"))
	{
		READ_PARAM(stdin,"%255s",filenameLinks,strcat(msg,"Filename links missing"));
		cout<<"building topo from file\n";
		(*T)->buildGi_FromFile(filenameLinks);
	}  
/*	else if (!strcmp(buildGiFuncStr,"FixedPowerMaxRate"))
	{
		// read the power value (dbm)
		READ_PARAM(stdin,"%ud",&power,strcat(msg,msg_power));
		if ( (power < 0) || (power > NUMTXPWR-1) )
			PRINT_ERROR(strcat(msg,msg_power));
		(*T)->buildGi_FixedPowerMaxRate(availTxPwr[power]);
		(*T)->maxP = power;
	}*/
/*	else if (!strcmp(buildGiFuncStr,"StepTXrate"))
		buildGiFunc = StepTXrate;
	else // default
		buildGiFunc = OnOffTXrate;
	*/
	return 1;
}
#endif

/* The caller must declare a map<pair<int,int>,double> and
 *  pass its address. The map will be filled in with the amount
 *  of flow for each demand
 * Otherwise, pass ret=NULL */
bool Topology::checkE2Eflows(map<pair<int,int>,double>* ret)
{
  vector<map<pair<int,int>,double > > in(node.size()), out(node.size()), diff(node.size());
  
  if (ret)
    ret->clear();

  throughput = 0.;

  for (vector<TGraphNode>::const_iterator u=node.begin(); u!=node.end(); u++)
    for (list<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      for (map<pair<int,int>,double>::const_iterator f=v->f_e2e.begin(); f!=v->f_e2e.end(); f++) {
	out[u->Id()][f->first] += f->second;
	in[v->Id()][f->first] += f->second;
	diff[u->Id()][f->first] += f->second;
	diff[v->Id()][f->first] -= f->second;
      }
    }

#if 0
  /*uncomment for detailed DEBUG */
  for (int i=0; i<node.size(); i++) {
    cout << "Nodo " << i << ": in: ";
    for (map<pair<int,int>,double>::const_iterator f=in[i].begin(); f!=in[i].end(); f++)
      cout << "[" << f->first.first << "," << f->first.second << "] = " << f->second << "  ";
    cout << endl << "Nodo " << i << ": out: ";
    for (map<pair<int,int>,double>::const_iterator f=out[i].begin(); f!=out[i].end(); f++)
      cout << "[" << f->first.first << "," << f->first.second << "] = " << f->second << "  ";
    cout << endl << "Nodo " << i << ": diff: ";
    for (map<pair<int,int>,double>::const_iterator f=diff[i].begin(); f!=diff[i].end(); f++)
      cout << "[" << f->first.first << "," << f->first.second << "] = " << f->second << "  ";
    cout << endl << endl;
  }
  cout << endl << endl;
#endif

  bool ok = true;
  
  cout << endl;
 
  for (int i=0; i<(int)node.size(); i++) {
    /* if node i is an edge node (aggregated or gateway) */
    if ( IsAggregationNode(i) || IsGatewayNode(i) )
    {
      for (map<pair<int,int>,double>::const_iterator f=in[i].begin(); f!=in[i].end(); f++) {
      /* if the in entry has node i as source node, it must be null  */
	if ( f->first.first == i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": entering non-null amount (" << f->second << ") of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
	  ok = false;
	}
      /* if the in entry has node i as destination node, print the amount of flow entering node i */
	else if ( f->first.second == i )
	  cout << "Node " << i << ": entering " << f->second << " of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
      }
      for (map<pair<int,int>,double>::const_iterator f=out[i].begin(); f!=out[i].end(); f++) {
      /* if the out entry has node i as destination node, it must be null  */
	if ( f->first.second== i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": leaving non-null amount (" << f->second << ") of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
	  ok = false;
	}
      /* if the out entry has node i as source node, print the amount of flow leaving node i */
	else if ( f->first.first == i ) {
	  cout << "Node " << i << ": leaving " << f->second << " of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
	  if (ret)
	    ret->insert(*f);
	  throughput += f->second;
	}
      }
      for (map<pair<int,int>,double>::const_iterator f=diff[i].begin(); f!=diff[i].end(); f++) {
      /* if the diff entry hasn't node i as source or destination node, the diff must be null */
	if ( f->first.first != i  && f->first.second != i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": the amount of e2e flow [" << f->first.first << "," << f->first.second << "] leaving the node does not equal the amount of flow entering" << endl;
	  ok = false;
	}
      }
    }
    else {
      /* node i is an intermediate node */
      for (map<pair<int,int>,double>::const_iterator f=diff[i].begin(); f!=diff[i].end(); f++)
      /* if the diff entry hasn't node i as source or destination node, the diff must be null */
	if ( f->first.first != i  && f->first.second != i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": the amount of e2e flow [" << f->first.first << "," << f->first.second << "] leaving the node does not equal the amount of flow entering" << endl;
	  ok = false;
	}
      }
  }

return ok;
}

#if 0
void Topology::saveConfig(const char* fname, PFR_method* pfr, bool commented)
{
  fstream f(fname, fstream::out);
  char str[3];
  
  if (commented)
    strcpy (str,"# ");
  else
    strcpy (str, " ");
  
  f << str << numAvailChannels << endl << str << topopath << endl;
  
  f << str << "FixedPowerMaxAllowRate " << maxP << " " << min_rate << " " << max_rate << endl;
  
  f << str << pfr->getStr() << endl;

//   f << str << "Power Consumption model: " << pm.name << endl;

  f.close();
}
#endif


void SINRbasedInterferenceModel::SerializeTo (ofstream& ofs)
{
  int slen = strlen(name);
  ofs.write ((const char*)&slen, sizeof(int));
  ofs.write (name, slen);
}

void kHopNeighbInterferenceModel::SerializeTo (ofstream& ofs)
{
  int slen = strlen(name);
  ofs.write ((const char*)&slen, sizeof(int));
  ofs.write (name, slen);
  ofs.write ((const char*)&k, sizeof(int));
}

InterferenceModel* InterferenceModel::DeserializeFrom (ifstream& ifs)
{
  int slen;

  ifs.read ((char*)&slen, sizeof(int));

  char tmp[slen+1];

  ifs.read (tmp, slen);
  tmp[slen] = '\0';

  if (!strcmp(tmp, SINRbasedInterferenceModel::name))
    return SINRbasedInterferenceModel::DeserializeFrom (ifs);

  if (!strcmp(tmp, kHopNeighbInterferenceModel::name))
    return kHopNeighbInterferenceModel::DeserializeFrom (ifs);

  cerr << endl << "** ERROR: Unknown interference model name! **" << endl << endl;
  return NULL;
}

SINRbasedInterferenceModel* SINRbasedInterferenceModel::DeserializeFrom (ifstream& ifs)
{
  return new SINRbasedInterferenceModel;
}

kHopNeighbInterferenceModel* kHopNeighbInterferenceModel::DeserializeFrom (ifstream& ifs)
{
  int _k;
  ifs.read ((char*)&_k, sizeof(int));
  return new kHopNeighbInterferenceModel (_k);
}

void LogDistancePropagationLossModel::SerializeTo (ofstream& ofs)
{
  int slen = strlen(name);
  ofs.write ((const char*)&slen, sizeof(int));
  ofs.write (name, slen);
  ofs.write ((const char*)&params.at(0).second, sizeof(double));
  ofs.write ((const char*)&params.at(1).second, sizeof(double));
  ofs.write ((const char*)&params.at(2).second, sizeof(double));
}

void ThreeLogDistancePropagationLossModel::SerializeTo (ofstream& ofs)
{
  int slen = strlen(name);
  ofs.write ((const char*)&slen, sizeof(int));
  ofs.write (name, slen);
  ofs.write ((const char*)&params.at(0).second, sizeof(double));
  ofs.write ((const char*)&params.at(1).second, sizeof(double));
  ofs.write ((const char*)&params.at(2).second, sizeof(double));
  ofs.write ((const char*)&params.at(3).second, sizeof(double));
  ofs.write ((const char*)&params.at(4).second, sizeof(double));
  ofs.write ((const char*)&params.at(5).second, sizeof(double));
  ofs.write ((const char*)&params.at(6).second, sizeof(double));
}

PropagationLossModel* PropagationLossModel::DeserializeFrom (ifstream& ifs)
{
  int slen;

  ifs.read ((char*)&slen, sizeof(int));

  char tmp[slen+1];

  ifs.read (tmp, slen);
  tmp[slen] = '\0';
cout << endl << "*********** " << tmp << endl;
  if (!strcmp(tmp, LogDistancePropagationLossModel::name))
    return LogDistancePropagationLossModel::DeserializeFrom (ifs);

  if (!strcmp(tmp, ThreeLogDistancePropagationLossModel::name))
    return ThreeLogDistancePropagationLossModel::DeserializeFrom (ifs);

  cerr << endl << "** ERROR: Unknown propagation loss model name! **" << endl << endl;
  return NULL;
}

LogDistancePropagationLossModel* LogDistancePropagationLossModel::DeserializeFrom (ifstream& ifs)
{
  double exponent, referenceDistance, referenceLoss;
  ifs.read ((char*)&exponent, sizeof(double));
  ifs.read ((char*)&referenceDistance, sizeof(double));
  ifs.read ((char*)&referenceLoss, sizeof(double));
  return new LogDistancePropagationLossModel (exponent, referenceDistance, referenceLoss);
}

ThreeLogDistancePropagationLossModel* ThreeLogDistancePropagationLossModel::DeserializeFrom (ifstream& ifs)
{
  double distance0, distance1, distance2, exponent0, exponent1, exponent2, referenceLoss;
  ifs.read ((char*)&distance0, sizeof(double));
  ifs.read ((char*)&distance1, sizeof(double));
  ifs.read ((char*)&distance2, sizeof(double));
  ifs.read ((char*)&exponent0, sizeof(double));
  ifs.read ((char*)&exponent1, sizeof(double));
  ifs.read ((char*)&exponent2, sizeof(double));
  ifs.read ((char*)&referenceLoss, sizeof(double));
  return new ThreeLogDistancePropagationLossModel (distance0, distance1, distance2,
                                                   exponent0, exponent1, exponent2, referenceLoss);
}

void WiMeshSimulation::SerializeTo (ofstream& ofs)
{
  ofs.write ((const char*)&m_numAvailChannels, sizeof(unsigned int));
  ofs.write ((const char*)&m_defaultTxPowerLevel, sizeof(unsigned int));
  ofs.write ((const char*)&m_min_rate, sizeof(int));
  ofs.write ((const char*)&m_max_rate, sizeof(int));
  ofs.write ((const char*)&m_SINRthreshfactor, sizeof(float));
  m_interferenceModel->SerializeTo (ofs);
  m_propagationLossModel->SerializeTo (ofs);

  int slen = m_simulId.length();
  ofs.write ((const char*)&slen, sizeof(int));
  ofs.write (m_simulId.c_str(), slen);
}

void WiMeshSimulation::DeserializeFrom (ifstream& ifs)
{
  ifs.read ((char*)&m_numAvailChannels, sizeof(unsigned int));
  ifs.read ((char*)&m_defaultTxPowerLevel, sizeof(unsigned int));
  ifs.read ((char*)&m_min_rate, sizeof(int));
  ifs.read ((char*)&m_max_rate, sizeof(int));
  ifs.read ((char*)&m_SINRthreshfactor, sizeof(float));
  m_interferenceModel = InterferenceModel::DeserializeFrom (ifs);
  m_propagationLossModel = PropagationLossModel::DeserializeFrom (ifs);

  int slen;
  ifs.read ((char*)&slen, sizeof(int));
  char tmp[slen+1];

  ifs.read (tmp, slen);
  tmp[slen] = '\0';

  m_simulId = string(tmp);
}

void Topology::SerializeTo (const char *filename)
{
  ofstream ofs (filename, ofstream::binary);
  int tmp, n;

  WMS.SerializeTo (ofs);

  // write member variables
  ofs.write ((const char*)&throughput, sizeof(float));

  tmp = toponame.length();
  ofs.write ((const char*)&tmp, sizeof(int));
  ofs.write (toponame.c_str(), tmp);
  
  tmp = topopath.length();
  ofs.write ((const char*)&tmp, sizeof(int));
  ofs.write (topopath.c_str(), tmp);

  // write aggregation list
  tmp = aggregation_list.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
    ofs.write ((const char*)&aggregation_list[i], sizeof(int));
  
  // write gateway list
  tmp = gateway_list.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
    ofs.write ((const char*)&gateway_list[i], sizeof(int));
  
  // write node info (MeshNode objects)
  tmp = node.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  
  for (int i=0; i<tmp; i++)
    node[i].info->SerializeTo (ofs);
    
  for (int i=0; i<tmp; i++)
  {
    n = node[i].neighbor.size();
    ofs.write ((const char*)&n, sizeof(int));
    for (list<TGraphNeighbor>::iterator j=node[i].neighbor.begin(); j!=node[i].neighbor.end(); j++)
      j->SerializeTo (ofs);
  }
    
  ofs.close ();
}

Topology* Topology::DeserializeFrom (const char *filename)
{
  int tmp, elem, n;
  ifstream ifs (filename, ifstream::binary);

  if (!ifs)
  {
    cerr << endl << "*** ERROR: Cannot find the file storing the serialized topology (" << filename << ") ***" << endl << endl;
    return NULL;
  }

  WMS.DeserializeFrom (ifs);

  Topology* T = new Topology;

  // read member variables
  ifs.read ((char*)&T->throughput, sizeof(float));

  ifs.read ((char*)&tmp, sizeof(int));
  char name_tmp[tmp+1];
  ifs.read (name_tmp, tmp);
  name_tmp[tmp] = '\0';
  T->toponame = string (name_tmp);

  ifs.read ((char*)&tmp, sizeof(int));
  char path_tmp[tmp+1];
  ifs.read (path_tmp, tmp);
  path_tmp[tmp] = '\0';
  T->topopath = string (path_tmp);

  // read aggregation list
  ifs.read ((char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&elem, sizeof(int));
    T->aggregation_list.push_back (elem);
  }

  // read gateway list
  ifs.read ((char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&elem, sizeof(int));
    T->gateway_list.push_back (elem);
  }

  // resize the node vector
  ifs.read ((char*)&tmp, sizeof(int));
  T->node.clear();

  TGraphNode gnode;
  for (int i=0; i<tmp; i++)
  {
    gnode.info = MeshNode::DeserializeFrom (ifs);
    T->node.push_back (gnode);
  }

  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&n, sizeof(int));
    for (int j=0; j<n; j++)
      T->node[i].neighbor.push_back (*TGraphNeighbor::DeserializeFrom (ifs, T, i));
  }

  ifs.close ();

  return T;
}


void MeshNode::SerializeTo (ofstream& ofs)
{
  ofs.write ((const char*)&id, sizeof(int));
  ofs.write ((const char*)&xpos, sizeof(float));
  ofs.write ((const char*)&ypos, sizeof(float));
  ofs.write ((const char*)&role, sizeof(int));
  ofs.write ((const char*)&BasePowerConsumption, sizeof(float));
  ofs.write ((const char*)&StandbyPowerConsumption, sizeof(float));

  unsigned int numAvailRadios = radio.size();
  ofs.write ((const char*)&numAvailRadios, sizeof(unsigned int));
  
  for (unsigned int i=0; i<numAvailRadios; i++)
    radio[i].SerializeTo (ofs);
}

MeshNode* MeshNode::DeserializeFrom (ifstream& ifs)
{
  int _id;
  TMeshRouter _role;
  float _xpos, _ypos, bpc, spc;
  unsigned int _numAvailRadios;
  
  ifs.read ((char*)&_id, sizeof(int));
  ifs.read ((char*)&_xpos, sizeof(float));
  ifs.read ((char*)&_ypos, sizeof(float));
  ifs.read ((char*)&_role, sizeof(int));
  ifs.read ((char*)&bpc, sizeof(float));
  ifs.read ((char*)&spc, sizeof(float));
  ifs.read ((char*)&_numAvailRadios, sizeof(unsigned int));
  
  MeshNode* m = new MeshNode (_id, _xpos, _ypos, _role, _numAvailRadios, bpc, spc);
  
  for (unsigned int i=0; i<_numAvailRadios; i++)
    m->radio[i].DeserializeFrom (ifs);
  
  return m;
}

void foRadioTxPowerLinear::SerializeTo (ofstream& ofs)
{
  int slen = strlen(name);
  ofs.write ((const char*)&slen, sizeof(int));
  ofs.write (name, slen);
  
  ofs.write ((const char*)&a, sizeof(float));
  ofs.write ((const char*)&b, sizeof(float));
}

foRadioTxPower* foRadioTxPower::DeserializeFrom (ifstream& ifs)
{
  int slen;

  ifs.read ((char*)&slen, sizeof(int));
  
  char tmp[slen+1];
  
  ifs.read (tmp, slen);
  tmp[slen] = '\0';
  
  if (!strcmp(tmp, foRadioTxPowerLinear::name))
    return foRadioTxPowerLinear::DeserializeFrom (ifs);

  cerr << endl << "** ERROR: Unknown RadioTxPower function object name! **" << endl << endl;
  return NULL;
}

foRadioTxPowerLinear* foRadioTxPowerLinear::DeserializeFrom (ifstream& ifs)
{
  float a, b;
  
  ifs.read ((char*)&a, sizeof(float));
  ifs.read ((char*)&b, sizeof(float));
  
  return new foRadioTxPowerLinear (a, b);
}

void RadioIface::SerializeTo (ofstream& ofs)
{
  ofs.write ((const char*)&channel, sizeof(unsigned int));
  ofs.write ((const char*)&tx_gain_db, sizeof(float));
  ofs.write ((const char*)&rx_gain_db, sizeof(float));
  ofs.write ((const char*)&minangle, sizeof(unsigned int));
  ofs.write ((const char*)&maxangle, sizeof(unsigned int));
  rho_tx->SerializeTo (ofs);
  ofs.write ((const char*)&rho_rx, sizeof(float));
  ofs.write ((const char*)&rho_idle, sizeof(float));
  ofs.write ((const char*)&rho_sleep, sizeof(float));
}

void RadioIface::DeserializeFrom (ifstream& ifs)
{
  ifs.read ((char*)&channel, sizeof(unsigned int));
  ifs.read ((char*)&tx_gain_db, sizeof(float));
  ifs.read ((char*)&rx_gain_db, sizeof(float));
  ifs.read ((char*)&minangle, sizeof(unsigned int));
  ifs.read ((char*)&maxangle, sizeof(unsigned int));
  rho_tx = foRadioTxPower::DeserializeFrom (ifs);
  ifs.read ((char*)&rho_rx, sizeof(float));
  ifs.read ((char*)&rho_idle, sizeof(float));
  ifs.read ((char*)&rho_sleep, sizeof(float));
}

void TGraphNeighbor::SerializeTo (ofstream& ofs)
{
  int tmp;
  int _id = Id();
  
  // write the id of the neighbor
  ofs.write ((const char*)&_id, sizeof(int));

  if (!radio)
    cerr << "ERROR: Trying to serialize a TGraphNeighbor node with NULL radio pointer" << endl << endl;

  assert (radio != 0);

  int _channel = radio->Channel();
  ofs.write ((const char*)&_channel, sizeof(unsigned int));

  ofs.write ((const char*)&availBW, sizeof(double));
  ofs.write ((const char*)&flow, sizeof(double));
  ofs.write ((const char*)&Utot, sizeof(double));
  ofs.write ((const char*)&rate, sizeof(float));
  ofs.write ((const char*)&txpower, sizeof(float));

  tmp = f_e2e.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  
  for (map<pair<int,int>,double>::iterator m=f_e2e.begin(); m!=f_e2e.end(); m++)
  {
    ofs.write ((const char*)&m->first.first, sizeof(int));
    ofs.write ((const char*)&m->first.second, sizeof(int));
    ofs.write ((const char*)&m->second, sizeof(double));
  }
}

TGraphNeighbor* TGraphNeighbor::DeserializeFrom (ifstream& ifs, Topology* T, int me)
{
  int tmp, s, d;
  unsigned int ch;
  double dem;
  TGraphNeighbor* n;

  // read the id of the neighbor;
  ifs.read ((char*)&tmp, sizeof(int));

  n = new TGraphNeighbor (T->Node(tmp)->Info(), 0, 0);

  ifs.read ((char*)&ch, sizeof(unsigned int));
  
//   for (tmp=0; tmp<(int)T->node[me].info->numAvailRadios() && T->node[me].info->radio[tmp].channel != ch; tmp++);
  int ri = T->Node(me)->Info()->findRadioWithChannel (ch);

//   assert (tmp < (int)T->node[me].info->numAvailRadios());
  assert ( ri >= 0 );

//   n->radio = T->node[me].info->radio + tmp;
  n->radio = &(*T->Node(me)->Info()->Radio(ri));

  ifs.read ((char*)&n->availBW, sizeof(double));
  ifs.read ((char*)&n->flow, sizeof(double));
  ifs.read ((char*)&n->Utot, sizeof(double));
  ifs.read ((char*)&n->rate, sizeof(float));
  ifs.read ((char*)&n->txpower, sizeof(unsigned int));

  ifs.read ((char*)&tmp, sizeof(int));
  
  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&s, sizeof(int));
    ifs.read ((char*)&d, sizeof(int));
    ifs.read ((char*)&dem, sizeof(double));
    
    n->f_e2e[pair<int,int>(s,d)] = dem;
  }
  
  return n;
}



int Topology::buildGi_FromFile(const char * filename)
{
	int rate;
	int idu,idv;
	char c;
	string s;

	// read the number of available channels
	//READ_PARAM(f,"%ui",&numAvailChannels,strcat(msg,"number of available channels"));

	ifstream ifs;

	ifs.open (filename);
	
	if (!ifs)
	{
	  cerr << endl << "ERROR: Cannot open file " << filename << " to read topology links" << endl << endl;
	  return -1;
	}

// 	skipComment(f);
	while (ifs >> c)
	{
		switch (c)
		{
			case 'L':
			  readValue (ifs, idu, "the id of the first endpoint");
			  readValue (ifs, idv, "the id of the second endpoint");
			  readValue (ifs, rate, "the rate");
		
			  //cout << "Name "<<nameu<<endl;
			  cout << "Node "<< idu << " and " << idv <<" are set to be neighbors and the rate is "<< rate<<".\n";
			  addNeighbor(idu,idv,rate,WMS.numTxPowerLevels()-1);
			  //addNeighbor(v,u,rate,availTxPwr[15]);
			  break;

			case '#':
			  getline (ifs, s);
			  break;

			default:
			  break;
		}
// 		skipComment(f);
	}

	return 0;
}

