#include "ns3/log.h"
#include "ns3/ipv4-address.h"
#include "ns3/nstime.h"
#include "ns3/inet-socket-address.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/uinteger.h"
#include "ns3/string.h"
#include "udp-trace-client.h"
#include <fstream>

namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("UdpTraceBasedClient");
NS_OBJECT_ENSURE_REGISTERED (UdpTraceBasedClient);

TypeId
UdpTraceBasedClient::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::UdpTraceBasedClient")
    .SetParent<Application> ()
    .AddConstructor<UdpTraceBasedClient> ()
    .AddAttribute ("RemoteAddress",
                   "The destination Ipv4Address of the outbound packets",
                   Ipv4AddressValue (),
                   MakeIpv4AddressAccessor (&UdpTraceBasedClient::m_peerAddress),
                   MakeIpv4AddressChecker ())
    .AddAttribute ("RemotePort",
                   "The destination port of the outbound packets",
                   UintegerValue (100),
                   MakeUintegerAccessor (&UdpTraceBasedClient::m_peerPort),
                   MakeUintegerChecker<uint16_t> ())
    .AddAttribute ("MaxPacketSize",
                   "The maximum size of a packet.",
                   UintegerValue (1450),
                   MakeUintegerAccessor (&UdpTraceBasedClient::m_maxPacketSize),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("TraceFilename",
                   "Name of file to load a trace from. By default, does nothing.",
                   StringValue (""),
                   MakeStringAccessor (&UdpTraceBasedClient::SetTraceFile),
                   MakeStringChecker ())

  ;
  return tid;
}

UdpTraceBasedClient::UdpTraceBasedClient ()
{
  NS_LOG_FUNCTION (this);
  m_sent = 0;
  m_socket = 0;
  m_sendEvent = EventId ();
  m_maxPacketSize = 1450;
}

UdpTraceBasedClient::UdpTraceBasedClient (Ipv4Address ip, uint16_t port,
                                char *traceFile)
{
  NS_LOG_FUNCTION (this);
  m_sent = 0;
  m_socket = 0;
  m_sendEvent = EventId ();
  m_peerAddress = ip;
  m_peerPort = port;
  m_currentEntry = 0;
  m_maxPacketSize = 1400;
  if (traceFile != NULL)
    {
      SetTraceFile (traceFile);
    }
}

UdpTraceBasedClient::~UdpTraceBasedClient ()
{
  NS_LOG_FUNCTION (this);
  m_entries.clear ();
}

void
UdpTraceBasedClient::SetRemote (Ipv4Address ip, uint16_t port)
{
  m_entries.clear ();
  m_peerAddress = ip;
  m_peerPort = port;
}

void
UdpTraceBasedClient::SetTraceFile (std::string traceFile)
{
  NS_LOG_FUNCTION (this << traceFile);
  double time, prevTime = 0;
  uint16_t size;
  TraceEntry entry;
  std::ifstream ifTraceFile;
  ifTraceFile.open (traceFile.c_str (), std::ifstream::in);
  m_entries.clear ();
  if (!ifTraceFile.good ())
    {
      NS_LOG_WARN ("Unable to open the trace file");
    }
  while (ifTraceFile.good ())
    {
      ifTraceFile >> time >> size;

      if (ifTraceFile.eof())
	break;
      
      entry.timeToSend = time - prevTime;
      prevTime = time;

      entry.packetSize = size;
      m_entries.push_back (entry);
    }
  ifTraceFile.close ();
  m_currentEntry = 0;
}

void
UdpTraceBasedClient::SetMaxPacketSize (uint16_t maxPacketSize)
{
  m_maxPacketSize = maxPacketSize;
}


uint16_t UdpTraceBasedClient::GetMaxPacketSize (void)
{
  return m_maxPacketSize;
}


void
UdpTraceBasedClient::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  Application::DoDispose ();
}

void
UdpTraceBasedClient::StartApplication (void)
{
  NS_LOG_FUNCTION (this);

  if (m_socket == 0)
    {
      TypeId tid = TypeId::LookupByName ("ns3::UdpSocketFactory");
      m_socket = Socket::CreateSocket (GetNode (), tid);
      m_socket->Bind ();
      m_socket->Connect (InetSocketAddress (m_peerAddress, m_peerPort));
    }
  m_socket->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > ());
//  m_sendEvent = Simulator::Schedule (Seconds (0.0), &UdpTraceBasedClient::Send, this);
  m_sendEvent = Simulator::Schedule (Seconds (m_entries[0].timeToSend), &UdpTraceBasedClient::Send, this);
}

void
UdpTraceBasedClient::StopApplication ()
{
  NS_LOG_FUNCTION (this);
  Simulator::Cancel (m_sendEvent);
}

void
UdpTraceBasedClient::SendPacket (uint32_t size)
{
  NS_LOG_FUNCTION (this << size);
  Ptr<Packet> p;
  uint32_t packetSize = (size > m_maxPacketSize ? m_maxPacketSize : size);
  p = Create<Packet> (packetSize);

  if ((m_socket->Send (p)) >= 0)
    {
      ++m_sent;
      NS_LOG_INFO ("Sent " << size << " bytes to "
                           << m_peerAddress);
    }
  else
    {
      NS_LOG_INFO ("Error while sending " << size << " bytes to "
                                          << m_peerAddress);
    }
}

void
UdpTraceBasedClient::Send (void)
{
  NS_LOG_FUNCTION (this);

  NS_ASSERT (m_sendEvent.IsExpired ());

  struct TraceEntry *entry = &m_entries[m_currentEntry];
  do
    {
      SendPacket (entry->packetSize);

      m_currentEntry++;
      m_currentEntry %= m_entries.size ();
      entry = &m_entries[m_currentEntry];
    }
  while (entry->timeToSend == 0);
  m_sendEvent = Simulator::Schedule (Seconds (entry->timeToSend), &UdpTraceBasedClient::Send, this);
}

/** Helper **/

UdpTraceBasedClientHelper::UdpTraceBasedClientHelper ()
{
}

UdpTraceBasedClientHelper::UdpTraceBasedClientHelper (Ipv4Address address, uint16_t port, std::string filename)
{
  m_factory.SetTypeId (UdpTraceBasedClient::GetTypeId ());
  SetAttribute ("RemoteAddress", Ipv4AddressValue (address));
  SetAttribute ("RemotePort", UintegerValue (port));
  SetAttribute ("TraceFilename", StringValue (filename));
}

void
UdpTraceBasedClientHelper::SetAttribute (std::string name, const AttributeValue &value)
{
  m_factory.Set (name, value);
}

ApplicationContainer
UdpTraceBasedClientHelper::Install (NodeContainer c)
{
  ApplicationContainer apps;
  for (NodeContainer::Iterator i = c.Begin (); i != c.End (); ++i)
    {
      Ptr<Node> node = *i;
      Ptr<UdpTraceBasedClient> client = m_factory.Create<UdpTraceBasedClient> ();
      node->AddApplication (client);
      apps.Add (client);
    }
  return apps;
}

} // Namespace ns3
