/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 Network Security Lab, University of Washington, Seattle.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Sidharth Nabar <snabar@uw.edu>, He Wu <mdzz@u.washington.edu>
 *          Stefano Avallone <stavallo@unina.it>
 */

#include "ns3/log.h"
#include "ns3/double.h"
#include "ns3/simulator.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/pointer.h"
#include "ns3/energy-source.h"
#include "ns3/wifi-net-device.h"
#include "wimesh-node-energy-model.h"

NS_LOG_COMPONENT_DEFINE ("WiMeshNodeEnergyModel");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (WiMeshNodeEnergyModel);

TypeId
WiMeshNodeEnergyModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::WiMeshNodeEnergyModel")
    .SetParent<DeviceEnergyModel> ()
    .AddConstructor<WiMeshNodeEnergyModel> ()
    .AddAttribute ("ActiveCurrentA",
                   "The active node current in Ampere.",
                   DoubleValue (0.70),
                   MakeDoubleAccessor (&WiMeshNodeEnergyModel::SetActiveCurrentA,
                                       &WiMeshNodeEnergyModel::GetActiveCurrentA),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("StandbyCurrentA",
                   "The standby node current in Ampere.",
                   DoubleValue (0.10),
                   MakeDoubleAccessor (&WiMeshNodeEnergyModel::SetStandbyCurrentA,
                                       &WiMeshNodeEnergyModel::GetStandbyCurrentA),
                   MakeDoubleChecker<double> ())
    .AddTraceSource ("TotalEnergyConsumption",
                     "Total energy consumption of the radio device.",
                     MakeTraceSourceAccessor (&WiMeshNodeEnergyModel::m_totalEnergyConsumption))
  ; 
  return tid;
}

WiMeshNodeEnergyModel::WiMeshNodeEnergyModel ()
{
  NS_LOG_FUNCTION (this);
  m_standbyTime = Seconds (0.0);
  m_currentState = WiMeshNodeEnergyModel::ACTIVE;  // initially ACTIVE
  m_lastUpdateTime = Seconds (0.0);
  m_nPendingChangeState = 0;
  m_isSupersededChangeState = false;
  m_source = NULL;
}

WiMeshNodeEnergyModel::~WiMeshNodeEnergyModel ()
{
  NS_LOG_FUNCTION (this);
  for (std::vector<WiMeshNodeEnergyModelPhyListener*>::iterator i = m_listeners.begin(); i != m_listeners.end(); ++i)
    delete (*i);
}

void
WiMeshNodeEnergyModel::SetEnergySource (Ptr<EnergySource> source)
{
  NS_LOG_FUNCTION (this << source);
  NS_ASSERT (source != NULL);
  m_source = source;
}

void
WiMeshNodeEnergyModel::AddNetDevice (Ptr<NetDevice> device)
{
  WiMeshNodeEnergyModelPhyListener* listener = new WiMeshNodeEnergyModelPhyListener;
  // set callback for WifiPhy listener
  listener->SetChangeStateCallback (MakeCallback (&DeviceEnergyModel::ChangeState, this));

  Ptr<WifiNetDevice> wnd = DynamicCast<WifiNetDevice> (device);
  wnd->GetPhy ()->RegisterListener (listener);

  m_listeners.push_back (listener);
  m_devices.Add (device);
}

double
WiMeshNodeEnergyModel::GetTotalEnergyConsumption (void) const
{
  NS_LOG_FUNCTION (this);
  return m_totalEnergyConsumption;
}

double
WiMeshNodeEnergyModel::GetStandbyTime (void) const
{
  NS_LOG_FUNCTION (this);
  return m_standbyTime.GetSeconds ();
}

double
WiMeshNodeEnergyModel::GetActiveCurrentA (void) const
{
  NS_LOG_FUNCTION (this);
  return m_activeCurrentA;
}

void
WiMeshNodeEnergyModel::SetActiveCurrentA (double activeCurrentA)
{
  NS_LOG_FUNCTION (this << activeCurrentA);
  m_activeCurrentA = activeCurrentA;
}

double
WiMeshNodeEnergyModel::GetStandbyCurrentA (void) const
{
  NS_LOG_FUNCTION (this);
  return m_standbyCurrentA;
}

void
WiMeshNodeEnergyModel::SetStandbyCurrentA (double standbyCurrentA)
{
  NS_LOG_FUNCTION (this << standbyCurrentA);
  m_standbyCurrentA = standbyCurrentA;
}

void
WiMeshNodeEnergyModel::ChangeState (int newState)
{
  NS_LOG_FUNCTION (this << newState);

  Time duration = Simulator::Now () - m_lastUpdateTime;
  NS_ASSERT (duration.GetNanoSeconds () >= 0); // check if duration is valid

  // energy to decrease = current * voltage * time
  double energyToDecrease = 0.0;
  double supplyVoltage = m_source->GetSupplyVoltage ();
  switch (m_currentState)
    {
    case WiMeshNodeEnergyModel::ACTIVE:
      energyToDecrease = duration.GetSeconds () * m_activeCurrentA * supplyVoltage;
      break;
    case WiMeshNodeEnergyModel::STANDBY:
      energyToDecrease = duration.GetSeconds () * m_standbyCurrentA * supplyVoltage;
      break;
    default:
      NS_FATAL_ERROR ("WiMeshNodeEnergyModel:Undefined node state: " << m_currentState);
    }

  // update total energy consumption
  m_totalEnergyConsumption += energyToDecrease;

  // update the time spent in standby mode
  if (m_currentState == WiMeshNodeEnergyModel::STANDBY)
    m_standbyTime += duration;

  // update last update time stamp
  m_lastUpdateTime = Simulator::Now ();

  m_nPendingChangeState++;

  // notify energy source
  m_source->UpdateEnergySource ();

  // in case the energy source is found to be depleted during the last update, a callback might be
  // invoked that might cause a change in the Wifi PHY state (e.g., the PHY is put into SLEEP mode).
  // This in turn causes a new call to this member function, with the consequence that the previous
  // instance is resumed after the termination of the new instance. In particular, the state set
  // by the previous instance is erroneously the final state stored in m_currentState. The check below
  // ensures that previous instances do not change m_currentState.

  if (!m_isSupersededChangeState)
    {
      // update current state
      WiMeshNodeEnergyModel::State state = WiMeshNodeEnergyModel::STANDBY;
      for (NetDeviceContainer::Iterator i = m_devices.Begin(); i != m_devices.End(); ++i)
	if (!DynamicCast<WifiNetDevice>(*i)->GetPhy()->IsStateSleep())
	  state = WiMeshNodeEnergyModel::ACTIVE;

      if ( (m_currentState == WiMeshNodeEnergyModel::ACTIVE && state == WiMeshNodeEnergyModel::STANDBY) ||
	   (m_currentState == WiMeshNodeEnergyModel::STANDBY && state == WiMeshNodeEnergyModel::ACTIVE) )
        {
	  SetWiMeshNodeState (state);

	  // some debug message
	  NS_LOG_DEBUG ("WiMeshNodeEnergyModel:Total energy consumption is " <<
			m_totalEnergyConsumption << "J");
	}
    }

  m_isSupersededChangeState = (m_nPendingChangeState > 1);

  m_nPendingChangeState--;
}

/*
 * Private functions start here.
 */

void
WiMeshNodeEnergyModel::DoDispose (void)
{
  NS_LOG_FUNCTION (this);
  m_source = NULL;
}

double
WiMeshNodeEnergyModel::DoGetCurrentA (void) const
{
  NS_LOG_FUNCTION (this);
  switch (m_currentState)
    {
    case WiMeshNodeEnergyModel::ACTIVE:
      return m_activeCurrentA;
    case WiMeshNodeEnergyModel::STANDBY:
      return m_standbyCurrentA;
    default:
      NS_FATAL_ERROR ("WiMeshNodeEnergyModel:Undefined node state:" << m_currentState);
    }
}

void
WiMeshNodeEnergyModel::SetWiMeshNodeState (const WiMeshNodeEnergyModel::State state)
{
  NS_LOG_FUNCTION (this << state);
  m_currentState = state;
  std::string stateName;
  switch (state)
    {
    case WiMeshNodeEnergyModel::ACTIVE:
      stateName = "ACTIVE";
      break;
    case WiMeshNodeEnergyModel::STANDBY:
      stateName = "STANDBY";
      break;
    }
  NS_LOG_DEBUG ("WiMeshNodeEnergyModel:Switching to state: " << stateName <<
                " at time = " << Simulator::Now ());
}

// -------------------------------------------------------------------------- //

WiMeshNodeEnergyModelPhyListener::WiMeshNodeEnergyModelPhyListener ()
{
  NS_LOG_FUNCTION (this);
  m_changeStateCallback.Nullify ();
}

WiMeshNodeEnergyModelPhyListener::~WiMeshNodeEnergyModelPhyListener ()
{
  NS_LOG_FUNCTION (this);
}

void
WiMeshNodeEnergyModelPhyListener::SetChangeStateCallback (DeviceEnergyModel::ChangeStateCallback callback)
{
  NS_LOG_FUNCTION (this << &callback);
  NS_ASSERT (!callback.IsNull ());
  m_changeStateCallback = callback;
}

void
WiMeshNodeEnergyModelPhyListener::NotifySleep (void)
{
  NS_LOG_FUNCTION (this);
  if (m_changeStateCallback.IsNull ())
    {
      NS_FATAL_ERROR ("WiMeshNodeEnergyModelPhyListener:Change state callback not set!");
    }

  m_changeStateCallback (WifiPhy::SLEEP);
}

void
WiMeshNodeEnergyModelPhyListener::NotifyWakeup(void)
{
  NS_LOG_FUNCTION (this);
  if (m_changeStateCallback.IsNull ())
    {
      NS_FATAL_ERROR ("WiMeshNodeEnergyModelPhyListener:Change state callback not set!");
    }
  m_changeStateCallback (WifiPhy::IDLE);
}

} // namespace ns3
